/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam;

import java.awt.geom.Point2D;
import java.util.List;
import java.util.Locale;

import rubeam.model.Element;
import rubeam.model.Node;
import rubeam.model.Structure;
import rubeam.model.loads.AbstractElementLoad;
import rubeam.model.loads.AbstractNodeLoad;
import rubeam.model.loads.ElementBend;
import rubeam.model.loads.ElementConLoadAng;
import rubeam.model.loads.ElementLeapDir;
import rubeam.model.loads.ElementLeapNor;
import rubeam.model.loads.ElementMoment;
import rubeam.model.loads.ElementSecOrder;
import rubeam.model.loads.ElementUniDistribLoadNor;
import rubeam.model.loads.ElementUniDistribLoadSecDir;
import rubeam.model.loads.ElementUniDistribLoadSecNor;
import rubeam.model.loads.ElementUniDistribMomentSec;
import rubeam.model.loads.ElementUniPreStress;
import rubeam.model.loads.NodalLoadAng;
import rubeam.model.loads.NodalMoment;
import rubeam.model.sections.AbstractSection;
import rubeam.model.sections.DefaultSection;
import rubeam.model.sections.LinearGeneralSection;
import rubeam.model.sections.LinearIProfileSection;
import rubeam.model.sections.LinearRectangleSection;
import rubeam.model.sections.LinearSandwichSection;
import Jama.Matrix;

public class Util {

    public static final String NUMBER_STRING_FORMAT = "%.5f";
    public static final String HMTL_STYLE = 
        "<style>\n" 
        + "body\n" 
        + "{\n"
        + "    color: black;\n" 
        + "    background: white;\n"
        + "    font-family: Arial, sans-serif;\n" 
        + "    font-size: small\n"
        + "}\n";

    public static Matrix getBj(Element element, double xs) {

        Matrix Bj = new Matrix(2, 7);

        double hi = element.getSection().getHi();
        double hk = element.getSection().getHk();
        double si = element.getSection().getS();
        double areaI = element.getSection().getAi();

        // convergent?
        if (hk >= 2 * hi){
            throw new IllegalStateException("Convergence problem!");
            // what has to be done is to set ni = nk and nk==ni but must be done in element.java

        }
        // xs in element?
        if (xs > element.getLength()) {
            System.out.println("xs=" + xs + ", len=" + element.getLength());
            throw new IllegalStateException("xs to big!");
        }

        double delta = si * hi / areaI;
        double thetas = 1 / element.getLength() * (hk / hi - 1);
        double Ki = element.getNII() / element.getSection().getEI();
        double eta1 = (6 - 3 * delta) / (3 - 2 * delta) * thetas;
        double eta2 = 3 / (3 - 2 * delta) * thetas * thetas;
        double eta3 = delta / (3 - 2 * delta) * thetas * thetas * thetas;

        double[] bs = new double[7]; // [j-1+1]?
        double[] b = new double[7]; // [j-1+1]?

        // calculate bj and bsj from 2..j
        // java: form 1..j-1
        int s;
        for (int i = 1; i < 6; i++) { // for i
            bs[i] = getAj(xs, i); // aj function used faj.m
            b[i] = bs[i] * xs / (i + 1);
            double beta1 = b[i];
            double deltab = Math.abs(beta1);
            double beta2 = 0;
            double beta3 = 0;
            s = i + 1;

            // routine for each bj and bsj
            while (deltab > Math.abs(bs[i]) * Math.pow(10, -9)) {
                s = s + 1;
                double beta0 = beta2
                * Ki
                - ((beta3 * eta3 * (s - 4) + beta2 * eta2) * (s - 3) + beta1
                        * eta1) * (s - 2);
                bs[i] = bs[i] + beta0;
                deltab = deltab / 10 + Math.abs(beta0);
                double e = xs / s;
                beta3 = beta2 * e;
                beta2 = beta1 * e;
                beta1 = beta0 * e;
                b[i] = b[i] + beta1;
            }

            // write into Bj and Bsj matrix
            Bj.set(0, i + 1, b[i]);
            Bj.set(1, i + 1, bs[i]);
        }

        // b0 und b'0
        Bj.set(0, 0, 1 + Ki * Bj.get(0, 2));
        Bj.set(1, 0, Ki * Bj.get(1, 2));
        // b1 und b'1
        Bj.set(0, 1, xs + Ki * Bj.get(0, 3));
        Bj.set(1, 1, 1 + Ki * Bj.get(1, 3));

        return Bj;
    }

    public static double getAj(double x, int j) {
        return Math.pow(x, j) / factorial(j);
    }

    public static long factorial(long n) {
        if (n < 0)
            throw new RuntimeException("Underflow error in factorial");
        else if (n > 20)
            throw new RuntimeException("Overflow error in factorial");
        else if (n == 0)
            return 1;
        else
            return n * factorial(n - 1);
    }

    public static Matrix getRotMatrix(Node ni, Node nk) {
        Point2D Pi = ni.getPoint();
        Point2D Pk = nk.getPoint();
        double len = Pi.distance(Pk);
        double a = Pk.getX() - Pi.getX();
        double g = Pk.getY() - Pi.getY();
        double cosa = a / len;
        double sina = g / len;
        double[][] _D = { { cosa, sina, 0, 0, 0, 0 },
                { -sina, cosa, 0, 0, 0, 0 }, { 0, 0, 1, 0, 0, 0 },
                { 0, 0, 0, 1, 0, 0 }, { 0, 0, 0, 0, cosa, -sina },
                { 0, 0, 0, 0, sina, cosa } };
        Matrix D = new Matrix(_D);
        return D;
    }

    public static Matrix getTransferL(List<AbstractElementLoad> elementLoads, double xs) {
        Matrix L = new Matrix(6, 1, 0.0);
        if (elementLoads != null) {
            for (AbstractElementLoad load : elementLoads) {
                L = L.plus(load.getLoadVectors(xs)[2]);
            }
            L = L.uminus();
            L.set(2, 0, -L.get(2, 0));
        }
        return L;
    }

    public static void addSubMatrix(Matrix mat, int[] rows, int[] cols, Matrix sub) {
        Matrix matSub = mat.getMatrix(rows, cols);
        matSub = matSub.plus(sub);
        mat.setMatrix(rows, cols, matSub);
    }

    public static String MatrixToString(Matrix mat) {
        StringBuffer mts = new StringBuffer();
        for (int i = 0; i < mat.getRowDimension(); ++ i) {
            for (int j = 0; j < mat.getColumnDimension(); ++j)
                mts.append(String.format(Locale.ENGLISH, "%12.5f", mat.get(i, j))).append("\t");
            mts.append("\n");
        }
        return mts.toString();
    }

    public static String structureToString(Structure structure) {
        Structure struct = Structure.getStructure();
        List < Node > nodes = struct.readNodes();
        List < Element > elements = struct.readElements();
        String structStr = "ruBeam 1.0 beta, Copyright 2009 E. Bombasaro, Ch. Koch\n";
        structStr += "\n\n";
        structStr += "Structure:\n----------\nNodes:\n"; 
        if (nodes.isEmpty()) {
            structStr += "No nodes available!\n";
        }
        else{
            structStr += "Name \t\t (x, y)\t\t dphi\t dy\t dx\n";
            for (Node n : nodes) {
                Point2D p = n.getPoint();
                String x = String.format(Locale.ENGLISH, "%.5f", p.getX());
                String y = String.format(Locale.ENGLISH, "%.5f", p.getY());
                structStr += n.getName() + "\t(" + x + ", " + y + ")\t";
                structStr += String.format(Locale.ENGLISH, "%5.4f", n.getContraints()[0])+ "\t";
                structStr += String.format(Locale.ENGLISH, "%5.4f", n.getContraints()[1])+ "\t";
                structStr += String.format(Locale.ENGLISH, "%5.4f", n.getContraints()[2]) + "\n";
                if (struct.getNodeLoads(n) != null) {
                    for (AbstractNodeLoad nl : struct.getNodeLoads(n)) {
                        if (nl instanceof NodalLoadAng) {
                            NodalLoadAng l = (NodalLoadAng) nl;
                            structStr += "\t\t Nodal Load angular: "
                                + l.getName() + ", P="    
                                + l.getP() + ", \u03b1="
                                + Math.toDegrees(l.getAngle()) + "\u00b0 \n";
                        }
                        else if (nl instanceof NodalMoment) {
                            NodalMoment l = (NodalMoment) nl;
                            structStr += "\t\t Nodal Moment: "
                                + l.getName() + ", M="    
                                + l.getM() + "\n";
                        }
                        else {
                            structStr += "\t\t No load assigned to node.\n";
                        }
                    }
                }
            }
        }
        // elements
        structStr += "----------\nElements:\n"; 
        if (elements.isEmpty()) {
            structStr += "No elements available!\n";
        }
        else{
            for (Element e : elements) {
                structStr += "Name \t Nodes (i, k)\t Section Name\t HingeI\tHingeK\n";
                structStr += e.getName() + "\t(" + e.getNodeI().getName() + ", "
                + e.getNodeK().getName() + ") \t\t" + e.getSection().getName() +"\t\t" + e.getHingeI() + "\t"+e.getHingeK()+ "\n";
                if (struct.getElementLoads(e) != null) {
                    for (AbstractElementLoad bl : struct.getElementLoads(e)) {

                        if (bl instanceof ElementBend) {
                            ElementBend l = (ElementBend) bl;
                            structStr += "\t\t Element point bend: "
                                + l.getName() + ", \u03c6E="
                                + l.getPhiE() + "??, li=" + l.getLi() +"m\n";
                        }
                        else if (bl instanceof ElementConLoadAng) {
                            ElementConLoadAng l = (ElementConLoadAng) bl;
                            structStr += "\t\t Element concentrated angular load : "
                                + l.getName() + ", F="
                                + l.getF() + "N, \u03b1=" + Math.toDegrees(l.getAngle()) + "\u00b0, li=" + l.getLi() +"m\n";
                        }
                        else if (bl instanceof ElementLeapDir) {
                            ElementLeapDir l = (ElementLeapDir) bl;
                            structStr += "\t\t Element point leap directional: "
                                + l.getName() + ", uE="
                                + l.getUE() + "m, li=" + l.getLi() +"m \n";
                        }
                        else if (bl instanceof ElementLeapNor) {
                            ElementLeapNor l = (ElementLeapNor) bl;
                            structStr += "\t\t Element point leap normal: "
                                + l.getName() + ", wE="
                                + l.getWE() + "m, li=" + l.getLi() +"m \n";
                        }
                        else if (bl instanceof ElementMoment) {
                            ElementMoment l = (ElementMoment) bl;
                            structStr += "\t\t Element point moment: "
                                + l.getName() + ", M="
                                + l.getM() + "Nm, li=" + l.getLi() +"m \n";
                        }
                        else if (bl instanceof ElementUniDistribLoadNor) {
                            ElementUniDistribLoadNor l = (ElementUniDistribLoadNor) bl;
                            structStr += "\t\t Element uniformly distributed normal load: "
                                + l.getName() + ", q="
                                + l.getQ0() + "N/m \n";
                        }
                        else if (bl instanceof ElementUniDistribLoadSecNor) {
                            ElementUniDistribLoadSecNor l = (ElementUniDistribLoadSecNor) bl;
                            structStr += "\t\t Element uniformly distributed sectioned normal load: "
                                + l.getName() + ", \u0394q="
                                + l.getDq() + "N/m, li=" + l.getLi() + "m, lj=" + l.getLj() + "m \n";
                        }
                        else if (bl instanceof ElementUniDistribLoadSecDir) {
                            ElementUniDistribLoadSecDir l = (ElementUniDistribLoadSecDir) bl;
                            structStr += "\t\t Element uniformly distributed sectioned directional load: "
                                + l.getName() + ", \u0394n="
                                + l.getDn() + "N/m, li=" + l.getLi() + "m, lj=" + l.getLj() + "m \n";
                        }
                        else if (bl instanceof ElementUniDistribMomentSec) {
                            ElementUniDistribMomentSec l = (ElementUniDistribMomentSec) bl;
                            structStr += "\t\t Element uniformly distributed sectioned moment: "
                                + l.getName() + ", \u0394m="
                                + l.getDm() + "Nm/m, li=" + l.getLi() + "m, lj=" + l.getLj() + "m \n";
                        }
                        else if (bl instanceof ElementUniPreStress) {
                            ElementUniPreStress l = (ElementUniPreStress) bl;
                            structStr += "\t\t Element uniformal prestress: "
                                + l.getName() + ", \u03b5E="
                                + l.getEE() + "??? \n";
                        }
                        //
                        else if (bl instanceof ElementSecOrder) {
                            structStr += "\t\t Secound order influence included: "
                                + bl.getName() + ", NII="+ e.getNII() + "N, w0=" 
                                + e.getW0() + "m, \u03c80=" + e.getPsi0() + "?? \n";
                        }
                        else {
                            structStr += "\t\t No load assinged to element.\n";
                        }
                    }
                }
            }
        }
        // sections
        structStr += "----------\nSections:\n"; 
        if (!(struct.readSections(LinearGeneralSection.class).isEmpty())){
            structStr += "General Section \nName \t Ai \t Ak \t Ii \t E \n";
        }
        for (AbstractSection sec : struct.readSections(LinearGeneralSection.class)) {
            LinearGeneralSection s = (LinearGeneralSection) sec;
            structStr += s.getName() + "\t"  + s.getAi() + "\t"+ s.getAk() + "\t"+ s.getIi() + "\t"+ s.getE()+ "\n";  
        }
        if (!(struct.readSections(LinearRectangleSection.class).isEmpty())){
            structStr += "Rectangular Section \nName \t hi \t hk \t b \t E \n";
        }
        for (AbstractSection sec : struct.readSections(LinearRectangleSection.class)) {
            LinearRectangleSection s = (LinearRectangleSection) sec;
            structStr += s.getName() + "\t" + s.getHi() + "\t"+ s.getHk() + "\t"+ s.getB() + "\t" + s.getE()+ "\n";  
        }
        if (!(struct.readSections(LinearIProfileSection.class).isEmpty())){
            structStr += "I Profil Section \nName \t hi \t hk \t s \t b \t t \t E \n";
        }
        for (AbstractSection sec : struct.readSections(LinearIProfileSection.class)) {
            LinearIProfileSection s = (LinearIProfileSection) sec;
            structStr += s.getName() + "\t" + s.getHi() + "\t"+ s.getHk() + "\t"+ s.getS() + "\t" + s.getB() + "\t"+ s.getT() + "\t"+ s.getE() + "\n";  
        }
        if (!(struct.readSections(LinearSandwichSection.class).isEmpty())){
            structStr += "Sandwich Section\nName \t hi \t hk \t b \t t \t E \n";
        }
        for (AbstractSection sec : struct.readSections(LinearSandwichSection.class)) {
            LinearSandwichSection s = (LinearSandwichSection) sec;          
            structStr += s.getName() + "\t" + s.getHi() + "\t"+ s.getHk() + "\t"+ s.getB() + "\t"+ s.getT() + "\t"+ s.getE() + "\n";  
        }
        return structStr;
    }

    public static String exportToString(Structure structure) {
        Structure struct = Structure.getStructure();
        List < Node > nodes = struct.readNodes();
        List < Element > elements = struct.readElements();
        List < AbstractSection > sections = struct.readSections();
        String exportPr = "// ruBeam 1.0 beta, Copyright 2009 E. Bombasaro, Ch. Koch\n\n";
        //
        String exportNode = "//Nodes:\n//-------------------------\n";
        String exportEle = "\n//Elements:\n//-------------------------\n";
        String exportConst = "\n//Constraints:\n//-------------------------\n";
        String exportSec = "\n//Sections\n//-------------------------\n";
        String exportAssignSec = "\n//Sections assignment:\n//-------------------------\n";
        String exportNodeLoad = "\n//Nodal Loads:\n//-------------------------\n";
        String exportEleLoad = "\n//Element Loads:\n//-------------------------\n";

        //nodes, constraints, nodal loads
        for (Node n : nodes) {
            Point2D p = n.getPoint();
            String x = String.format(Locale.ENGLISH, "%.10f", p.getX());
            String y = String.format(Locale.ENGLISH, "%.10f", p.getY());
            exportNode += "addnode "+ x + " "+ y + " "+ n.getName() + ";\n";

            if ( !Double.isNaN(n.getContraints()[0]) || !Double.isNaN(n.getContraints()[1]) || !Double.isNaN(n.getContraints()[2]) ){
                exportConst += "setnodeconsts "+ n.getName() + " ";
                exportConst += String.format(Locale.ENGLISH, "%.10f", n.getContraints()[0]) + " ";
                exportConst += String.format(Locale.ENGLISH, "%.10f", n.getContraints()[1]) + " ";
                exportConst += String.format(Locale.ENGLISH, "%.10f", n.getContraints()[2]) +  ";\n";
            }
            if (struct.getNodeLoads(n) != null) {
                for (AbstractNodeLoad nl : struct.getNodeLoads(n)) {
                    if (nl instanceof NodalLoadAng) {
                        NodalLoadAng l = (NodalLoadAng) nl;
                        exportNodeLoad += "addnodalloadang " + n.getName() + " " + l.getP() + " " + Math.toDegrees(l.getAngle()) + " " + l.getName() + ";\n";
                    }
                    if (nl instanceof NodalMoment) {
                        NodalMoment l = (NodalMoment) nl;
                        exportNodeLoad += "addnodalmoment " + n.getName() + " " + l.getM() + " " + l.getName() + ";\n";
                    }
                }
            }
        }

        //elements, element loads
        for (Element e : elements) {
            exportEle += "addelement "+ e.getNodeI().getName() + " "+ e.getNodeK().getName() + " "+ e.getName() + ";\n";
            if (e.getHingeI() == true)
            {exportEle += "setelementhingei "+ e.getName() + " "+ e.getHingeI() + ";\n";}
            if (e.getHingeK() == true)
            {exportEle += "setelementhingek "+ e.getName() + " "+ e.getHingeK() + ";\n";}

            if (!(e.getSection().getClass() == DefaultSection.class)){
                exportAssignSec += "assignsection " + e.getSection().getName() + " " + e.getName() + ";\n";
            }

            if (struct.getElementLoads(e) != null) {
                for (AbstractElementLoad el : struct.getElementLoads(e)) {
                    if (el instanceof ElementConLoadAng) {
                        ElementConLoadAng l = (ElementConLoadAng) el;
                        exportEleLoad += "addelementconload "+ e.getName() + " "+ l.getF() + " "+ Math.toDegrees(l.getAngle()) + " "+ l.getLi() + " "+ l.getName() + ";\n";
                    }
                    else if (el instanceof ElementUniDistribLoadNor) {
                        ElementUniDistribLoadNor l = (ElementUniDistribLoadNor) el;
                        exportEleLoad += "addelementunidisloadnor "+ e.getName() + " "+ l.getQ0() + " "+ l.getName() + ";\n";
                    }
                    else if (el instanceof ElementUniDistribLoadSecNor) {
                        ElementUniDistribLoadSecNor l = (ElementUniDistribLoadSecNor) el;
                        exportEleLoad += "addelementunidisloadsecnor "+ e.getName() + " "+ l.getDq() + " "+ l.getLi() + " "+ l.getLj() + " "+ l.getName() + ";\n"; 
                    }
                    else if (el instanceof ElementUniDistribLoadSecDir) {
                        ElementUniDistribLoadSecDir l = (ElementUniDistribLoadSecDir) el;
                        exportEleLoad += "addelementunidisloadsecdir "+ e.getName() + " "+ l.getDn() + " "+ l.getLi() + " "+ l.getLj() + " "+ l.getName() + ";\n";
                    }                    
                    else if (el instanceof ElementMoment) {
                        ElementMoment l = (ElementMoment) el;
                        exportEleLoad += "addelementmoment "+ e.getName() + " "+ l.getM() + " "+ l.getLi() + " "+ l.getName() + ";\n";
                    }
                    else if (el instanceof ElementUniDistribMomentSec) {
                        ElementUniDistribMomentSec l = (ElementUniDistribMomentSec) el;
                        exportEleLoad += "addelementunidismomentsec "+ e.getName() + " "+ l.getDm() + " "+ l.getLi() + " "+ l.getLj() + " "+ l.getName() + ";\n";
                    }
                    else if (el instanceof ElementUniPreStress) {
                        ElementUniPreStress l = (ElementUniPreStress) el;
                        exportEleLoad += "addelementprestress "+ e.getName() + " "+ l.getEE() + " " + l.getName() + ";\n";
                    }
                    else if (el instanceof ElementLeapNor) {
                        ElementLeapNor l = (ElementLeapNor) el;
                        exportEleLoad += "addelementleapnor "+ e.getName() + " "+ l.getWE() + " "+ l.getLi() + " "+ l.getName() + ";\n";
                    }                   
                    else if (el instanceof ElementLeapDir) {
                        ElementLeapDir l = (ElementLeapDir) el;
                        exportEleLoad += "addelementleapdir "+ e.getName() + " "+ l.getUE() + " "+ l.getLi() + " "+ l.getName() + ";\n";
                    }                   
                    if (el instanceof ElementBend) {
                        ElementBend l = (ElementBend) el;
                        exportEleLoad += "addelementbend "+ e.getName() + " "+ l.getPhiE() + " "+ l.getLi() + " "+ l.getName() + ";\n";

                    }
                }
            }
        }
        // sections
        for (AbstractSection sec : sections) {
            //if ()){
            if ((sec.getClass() == LinearGeneralSection.class)){
                exportSec += "addgensection "+ sec.getAi() + " " + sec.getAk() + " " + sec.getIi() + " " + sec.getE() + " "+ sec.getName() + ";\n";  
            }
            if ((sec.getClass() == LinearRectangleSection.class)){
                exportSec += "addrecsection "+ sec.getHi() + " " + sec.getHk() + " " + sec.getB() + " " + sec.getE() + " "+ sec.getName() + ";\n"; 
            }
            if ((sec.getClass() == LinearIProfileSection.class)){
                exportSec += "addiprofsection "+ sec.getHi() + " " + sec.getHk() + " " + sec.getS() + " " + sec.getB() + " " + sec.getT() + " " + sec.getE() + " "+ sec.getName() + ";\n"; 
            }
            if ((sec.getClass() == LinearSandwichSection.class)){
                exportSec += "addsandsection "+ sec.getHi() + " " + sec.getHk() + " " + sec.getB() + " " + sec.getT() + " " + sec.getE() + " "+ sec.getName() + ";\n"; 
            }
        }

        String exportAdd = "\n//Addendum:\n//-------------------------\n";
        exportAdd += "setvisualizesteps " + struct.getVisualizeSteps() + ";\n";

        return exportPr+exportNode+exportEle+exportConst+exportSec+exportAssignSec+exportNodeLoad+exportEleLoad+exportAdd;
    }

    public static String resultsToString(Structure structure) {
        Structure struct = Structure.getStructure();
        List < Node > nodes = struct.readNodes();
        List < Element > elements = struct.readElements();
        String result = "ruBeam 1.0 beta, Copyright 2009 E. Bombasaro, Ch. Koch\n";
        result += "\n\n";
        result += "Results for Structure:\n" + "----------\n";

        if (struct.isValid() == false) {
            result += "No results available, system has to be analysed in order to obtain results!";
        }
        else{
            result += "Nodal Condition Vectors\n\n";
            for (Node n : nodes) {
                String vec =  Util.MatrixToString(struct.getNodeResult(n));
                result += "\t" + n.getName() + "\n" + vec + "\n";
            }
            result += "\n Elements Condition Matrix:\n";
            for (Element e : elements) {
                result += "Global Element Condition Matrix\n";
                String matg =  Util.MatrixToString(struct.getElementResults(e, struct.getResultSteps())[0]);
                result += "\t" + e.getName() + "\n" + matg + "\n";
                result += "Local Element Condition Matrix\n";
                String matl =  Util.MatrixToString(struct.getElementResults(e, struct.getResultSteps())[1]);
                result += "\t" + e.getName() + "\n" + matl + "\n";
            }
        }
        return result;
    }

    public static String structureToHTML(Structure structure) {
        Structure struct = Structure.getStructure();
        List < Node > nodes = struct.readNodes();
        List < Element > elements = struct.readElements();
        List < AbstractSection > gensections = struct.readSections(LinearGeneralSection.class);
        List < AbstractSection > recsections = struct.readSections(LinearRectangleSection.class);
        List < AbstractSection > iprofsections = struct.readSections(LinearIProfileSection.class);
        List < AbstractSection > sandsections = struct.readSections(LinearSandwichSection.class);
        String structHTML = getHtmlHead("ruBeam structure");
        structHTML += "<body>" + "<h1>Structure</h1>";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<!-- Node --------------------------------------------------------------------->";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<h2><b><font color=blue><a name=&quotNode&quot>Node</a></font></b></h2>";

        if (nodes.isEmpty()) {
            structHTML += "<b>No nodes avaiable!</b>";
        }
        else{
            structHTML += "<table rules=all border=1 cellpadding=7 width=100%>";
            structHTML += "<tr><th>Name</th> <th align=right> x </th><th align=right> y </th><th align=right> dx </th><th  align=right> dy </th><th align=right> d&phi </th></tr>";
            for (Node n : nodes) {
                Point2D p = n.getPoint();
                structHTML += "<tr><th>";
                structHTML += n.getName();
                structHTML += "</th> <th align=right>";
                structHTML += String.format(Locale.ENGLISH, "%.5f", p.getX());
                structHTML += "</th> <th align=right>";
                structHTML +=String.format(Locale.ENGLISH, "%.5f", p.getY());
                structHTML += "</th> <th align=right>";
                structHTML += String.format(Locale.ENGLISH, "%5.4f", n.getContraints()[0]);
                structHTML += "</th> <th align=right>";
                structHTML += String.format(Locale.ENGLISH, "%5.4f", n.getContraints()[1]);
                structHTML += "</th> <th align=right>";
                structHTML += String.format(Locale.ENGLISH, "%5.4f", n.getContraints()[2]);
                structHTML += "</th></tr>";
            }
        }
        structHTML += " </table>";

        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<!-- Nodal Loads--------------------------------------------------------------->";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<h2><b><font color=blue><a name=&quotNodal Loads&quot>Nodal Loads</a></font></b></h2>";
        if (struct.readNodeLoads().isEmpty()) {structHTML += "<b>No load assinged to nodes!</b>";}
        else{ 
            structHTML += "<table rules=all border=1 cellpadding=7 width=100%>";
            for (Node n : nodes) {      
                if (struct.getNodeLoads(n) != null) {            
                    for (AbstractNodeLoad nl : struct.getNodeLoads(n)) {
                        if (nl instanceof NodalLoadAng) {
                            NodalLoadAng l = (NodalLoadAng) nl;
                            structHTML += "<tr><th colspan=4 align=left><b>Nodal Load Angular</b></th></tr>";
                            structHTML += "<tr> <th>Node</th><th>Name</th><th align=right>P</th><th align=right> &alpha </tr>";
                            structHTML += "<tr><th>";
                            structHTML += n.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getP();
                            structHTML += "</th> <th align=right>";
                            structHTML += Math.toDegrees(l.getAngle());
                            structHTML += "</th> </tr>";   
                        }
                        else if (nl instanceof NodalMoment) {
                            NodalMoment l = (NodalMoment) nl;
                            structHTML += "<tr><th colspan=4 align=left> Nodal Moment</th></tr>";
                            structHTML += "<tr> <th>Node</th><th>Name</th> <th align=right>M</th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += n.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getM();
                            structHTML += "</th><th></th> </tr>";   
                        }
                    }
                }
            }
        }
        structHTML += " </table>";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<!-- Sections ------------------------------------------------------------------>";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<h2><b><font color=blue><a name=Sections>Sections</a></font></b></h2>";
        structHTML += "<table rules=all border=1 cellpadding=7 width=100%>";

        if (!gensections.isEmpty()){
            structHTML +="<tr><th colspan=7 align=left>General Section</th></tr>";
            structHTML +="<tr><th> Name </th><th> Ai </th><th> Ak </th><th> Ii </th><th> E </th> <th></th><th></th> </tr>";
        }
        for (AbstractSection sec :  gensections) {

            structHTML +="<tr><th>" + sec.getName() + "</th><th>" + sec.getAi() + "</th><th>"+ sec.getAk() + "</th><th>"+ sec.getIi() + "</th><th>"+ sec.getE()+ "</th><th></th><th></th></tr>";  
        }
        if (!recsections.isEmpty()){
            structHTML +="<tr><th colspan=7 align=left>Rectangular Section</th></tr>";
            structHTML +="<tr><th> Name </th><th> hi </th><th> hk </th><th> b </th><th> E </th><th></th><th></th></tr>";
        }
        for (AbstractSection sec :  recsections) {

            structHTML +="<tr><th>" + sec.getName() + "</th><th>" + sec.getHi() + "</th><th>"+ sec.getHk() + "</th><th>"+ sec.getB() + "</th><th>"+ sec.getE()+ "</th><th></th><th></th></tr>";   
        }
        if (!iprofsections.isEmpty()){
            structHTML +="<tr><th colspan=7 align=left>I-Profil Section</th></tr>";
            structHTML +="<tr><th> Name </th><th> hi </th><th> hk </th><th> s </th><th> b </th><th> t </th><th> E </th></tr>";
        }
        for (AbstractSection sec :  iprofsections) {

            structHTML +="<tr><th>" + sec.getName() + " </th><th> " + sec.getHi() + "</th><th>"+ sec.getHk() + "</th><th>"+ sec.getS() + "</th><th>"+ sec.getB() + "</th><th>"+ sec.getT() +  "</th><th>"+ sec.getE()+ "</th></tr>";   
        }
        if (!sandsections.isEmpty()){
            structHTML +="<tr><th colspan=7 align=left>Sandwich Section</th></tr>";
            structHTML +="<tr><th> Name </th><th> hi </th><th> hk </th><th> b </th><th> t </th><th> E </th><th></th></tr>";
        }
        for (AbstractSection sec :  sandsections) {

            structHTML +="<tr><th>" + sec.getName() + "</th><th>" + sec.getHi() + "</th><th>"+ sec.getHk() + "</th><th>"+ sec.getB() + "</th><th>"+ sec.getT() +  "</th><th>"+ sec.getE()+ "</th><th></th></tr>";   
        }
        structHTML += " </table>";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<!-- Element ------------------------------------------------------------------>";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<h2><b><font color=blue><a name=&quotElement&quot>Element</a></font></b></h2>";


        if (elements.isEmpty()) {
            structHTML += "<b>No elements avaiable!</b>";
        }
        else{
            structHTML += "<table rules=all border=1 cellpadding=9 width=100%>";
            structHTML += "<tr><th>Name</th> <th>Node i</th> <th>Node j</th> <th>Section</th><th>Hinged I</th><th>Hinged K</th></tr>";
            for (Element e : elements) {
                structHTML += "<tr><th>";
                structHTML += e.getName();
                structHTML += "</th> <th>";
                structHTML += e.getNodeI().getName();
                structHTML += "</th> <th>";
                structHTML += e.getNodeK().getName();
                structHTML += "</th> <th>";
                structHTML += e.getSection().getName();
                structHTML += "</th> <th>";
                structHTML += e.getHingeI();
                structHTML += "</th> <th>";
                structHTML += e.getHingeK();
                structHTML += "</th> </tr>";
            }
        }
        structHTML += " </table>";        

        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<!-- Element Loads--------------------------------------------------------------->";
        structHTML += "<!----------------------------------------------------------------------------->";
        structHTML += "<h2><b><font color=blue><a name=&quotElement Loads&quot>Element Loads</a></font></b></h2>";

        if (struct.readElementLoads().isEmpty()) {structHTML += "<b>No load assinged to elements!</b>";}
        else{   
            structHTML += "<table rules=all border=1 cellpadding=&quot5&quot width=100%>";
            for (Element e : elements) {
                if (struct.getElementLoads(e) != null) {
                    for (AbstractElementLoad el : struct.getElementLoads(e)) {

                        if (el instanceof ElementBend) {
                            ElementBend l = (ElementBend) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Bend</b> </th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>&Phi E</th> <th align=right> li</th> </tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getPhiE();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th> </tr>";
                        }
                        else if (el instanceof ElementConLoadAng) {
                            ElementConLoadAng l = (ElementConLoadAng) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Concentrated Angular Load</b> </th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>F</th><th align=right>&alpha</th> <th align=right> li</th> </tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getF();
                            structHTML += "</th> <th align=right>";
                            structHTML += Math.toDegrees(l.getAngle());
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th> </tr>";
                        }
                        else if (el instanceof ElementLeapDir) {
                            ElementLeapDir l = (ElementLeapDir) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Leap Directional</b> </th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>UE</th> <th align=right> li </th><th></th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getUE();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th><th></th></tr>";
                        }
                        else if (el instanceof ElementLeapNor) {
                            ElementLeapNor l = (ElementLeapNor) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Leap Normal</b> </th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>WE</th> <th align=right> li </th><th></th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getWE();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th><th></th></tr>";
                        }
                        else if (el instanceof ElementMoment) {
                            ElementMoment l = (ElementMoment) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Piont Moment</b> </th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>M</th> <th align=right> li </th><th></th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getM();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th><th></th></tr>";
                        }
                        else if (el instanceof ElementUniDistribLoadNor) {
                            ElementUniDistribLoadNor l = (ElementUniDistribLoadNor) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Uniformal Distributed Load</b></th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>q</th><th></th><th></th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getQ0();
                            structHTML += "</th><th></th><th></th></tr>";
                        }
                        else if (el instanceof ElementUniDistribLoadSecNor) {
                            ElementUniDistribLoadSecNor l = (ElementUniDistribLoadSecNor) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Uniformly Distributed Sectioned Normal Load</b></th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>dq</th><th align=right>li</th> <th align=right> lj </th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getDq();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLj();
                            structHTML += "</th> </tr>";
                        }
                        else if (el instanceof ElementUniDistribLoadSecDir) {
                            ElementUniDistribLoadSecDir l = (ElementUniDistribLoadSecDir) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Uniformly Distributed Sectioned Directional Load</b></th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>dn</th><th align=right>li</th> <th align=right> lj </th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getDn();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLj();
                            structHTML += "</th> </tr>";
                        }
                        else if (el instanceof ElementUniDistribMomentSec) {
                            ElementUniDistribMomentSec l = (ElementUniDistribMomentSec) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Uniformly Distributed Sectioned Moment</b></th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>dm</th><th align=right>li</th> <th align=right> lj </th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getDm();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLi();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getLj();
                            structHTML += "</th> </tr>";
                        }
                        else if (el instanceof ElementUniPreStress) {
                            ElementUniPreStress l = (ElementUniPreStress) el;
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Uniformal Prestress</b></th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>eE</th><th></th><th></th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += l.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += l.getEE();
                            structHTML += "</th><th></th><th></th></tr>";
                        }
                        else if (el instanceof ElementSecOrder) {
                            structHTML += "<tr><th colspan=5 align=left> <b>Element Secound Order Influence Included</b></th></tr>";
                            structHTML += "<tr> <th>Element</th><th>Name</th> <th align=right>NII</th> <th align=right> w0 </tr><th align=right> psi0 </th></tr>";
                            structHTML += "<tr><th>";
                            structHTML += e.getName();
                            structHTML += "</th> <th>";
                            structHTML += el.getName();
                            structHTML += "</th> <th align=right>";
                            structHTML += el.getElement().getNII();
                            structHTML += "</th> <th align=right>";
                            structHTML += el.getElement().getW0();
                            structHTML += "</th> <th align=right>";
                            structHTML += el.getElement().getPsi0();
                            structHTML += "</th> </tr>";
                        }
                    }
                }
            }
        }
        structHTML += " </table>"; 
        structHTML += "</body> </html>";
        return structHTML;
    }

    public static String resultsToHTML(Structure structure) {
        Structure struct = Structure.getStructure();
        List < Node > nodes = struct.readNodes();
        List < Element > elements = struct.readElements();
        int steps = struct.getResultSteps();
        StringBuffer resultsHTML = new StringBuffer(getHtmlHead("ruBeam Results"));
        resultsHTML.append("<body> <h1>Results</h1>");
        if (struct.isValid() == false) {
            resultsHTML.append("<b>No results avaiable, system has to be analysed in order to obtain results!</b>");
        }
        else{
            resultsHTML.append("<!----------------------------------------------------------------------------->");
            resultsHTML.append("<!-- Nodal Results ------------------------------------------------------------>");
            resultsHTML.append("<!----------------------------------------------------------------------------->");
            resultsHTML.append("<h2><b><font color=blue><a name=&quotNodal Results &quot>Nodal Results</a></font></b></h2>");
            resultsHTML.append("<table rules=all border=1 cellpadding=7 width=100%>");
            resultsHTML.append("<tr><th>Name</th> <th align=right>&phi</th> <th align=right>w</th> <th align=right> u </th> <th align=right> M </th> <th align=right>V</th><th align=right> H </th></tr>");
            for (Node n : nodes) {
                resultsHTML.append("<tr><th>");
                resultsHTML.append(n.getName());
                resultsHTML.append("</th> <th align=right>");
                resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getNodeResult(n).get(0,0)));
                resultsHTML.append("</th> <th align=right>");
                resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getNodeResult(n).get(1,0)));
                resultsHTML.append("</th> <th align=right>");
                resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getNodeResult(n).get(2,0)));
                resultsHTML.append("</th> <th align=right>");
                resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getNodeResult(n).get(3,0)));
                resultsHTML.append("</th> <th align=right>");
                resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getNodeResult(n).get(4,0)));
                resultsHTML.append("</th> <th align=right>");
                resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getNodeResult(n).get(5,0)));
                resultsHTML.append("</th>  </tr>");
            }
            resultsHTML.append("</table>");
            resultsHTML.append("<!----------------------------------------------------------------------------->");
            resultsHTML.append("<!-- Element Results ------------------------------------------------------------>");
            resultsHTML.append("<!----------------------------------------------------------------------------->");
            resultsHTML.append("<h2><b><font color=blue><a name=&quotElement Results &quot>Element Results</a></font></b></h2>");
            resultsHTML.append("<table rules=all border=1 cellpadding=7 width=100%>");
            for (Element e : elements) {
                resultsHTML.append("<tr><th colspan=4 align=left><b><font color=green> Global Element Condition Matrix </font></b></th><th colspan=4 align=left><b>  Element Name: ");
                resultsHTML.append(e.getName());
                resultsHTML.append("</b></th></tr>");
                resultsHTML.append("<tr><th>Point</th><th>Position, xs</th> <th align=right>u</th> <th align=right>w</th> <th align=right> &phi </th> <th align=right> M </th> <th align=right>V</th><th align=right> H </th></tr>");
                for (int i = 0; i <= steps+1; i++){
                    resultsHTML.append("<tr><th>");
                    if (i == 0) { resultsHTML.append(e.getNodeI().getName());}
                    else if (i == steps+1) { resultsHTML.append(e.getNodeK().getName());}
                    else { resultsHTML.append(i);}
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",e.getLength()/steps*i));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[0].get(0,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[0].get(1,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[0].get(2,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[0].get(3,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[0].get(4,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[0].get(5,i)));
                    resultsHTML.append("</th>  </tr>");
                }
                resultsHTML.append("<tr><th colspan=4 align=left><b><font color=green> Local Element Condition Matrix </font></b></th><th colspan=4 align=left><b> Element Name: ");
                resultsHTML.append(e.getName());
                resultsHTML.append("</b></th></tr>");
                resultsHTML.append("<tr><th>Point</th><th>Position, xs</th><th align=right> u local</th> <th align=right>w local</th> <th align=right> &phi local</th> <th align=right> M local</th> <th align=right>Q local</th><th align=right> N local</th></tr>");
                for (int i = 0; i <= steps+1; i++){
                    resultsHTML.append("<tr><th>");
                    if (i == 0) { resultsHTML.append(e.getNodeI().getName());}
                    else if (i == steps+1) { resultsHTML.append(e.getNodeK().getName());}
                    else { resultsHTML.append(i);}
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",e.getLength()/steps*i));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[1].get(0,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[1].get(1,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[1].get(2,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[1].get(3,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[1].get(4,i)));
                    resultsHTML.append("</th> <th align=right>");
                    resultsHTML.append(String.format(Locale.ENGLISH, "%10.5f",struct.getElementResults(e, steps)[1].get(5,i)));
                    resultsHTML.append("</th>  </tr>");
                }
            }
        }
        resultsHTML.append("</body> </html>");
        return resultsHTML.toString();
    }

    public static String getHtmlHead(String title) {
        return "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.01 Transitional//EN\"\n"
        + "\"http://www.w3.org/TR/html4/loose.dtd\">\n"
        + "<html>\n"
        + "<head>\n"
        + "<title>" + title + "</title>\n"
        + HMTL_STYLE + "\n"
        + "</head>\n";
    }
    public static String format (double d) {
        return String.format(
                Locale.ENGLISH, Util.NUMBER_STRING_FORMAT, d);
    }
}
