/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.model;

import rubeam.Util;
import rubeam.model.sections.AbstractSection;
import Jama.Matrix;

public class Element extends ModelObject implements IChangeListener {

    private static final long serialVersionUID = 1L;
    
    private double NII = 0;
    private Node ni, nk;
    private boolean hingeI = false, hingeK = false; // hinged connection
    private double w0 = 0;
    private double psi0 = 0;
    
    public Element(Node ni, Node nk, String name) {
        super(name);
        this.ni = ni;
        this.nk = nk;
        Structure.getStructure().putSection(this, 
                AbstractSection.getDefaultSection());
    }

    public Object clone() {
        Element clone = (Element) super.clone(); // Makes a shallow copy
        return clone;
    }

    public Node getNodeI() {
        return this.ni;
    }

    public void setNodeI(Node ni) {
        this.ni = ni;
        _fireWasChanged();
    }

    public AbstractSection getSection() {
        return Structure.getStructure().getSection(this);
    }

    public double getNII() {
        return NII;
    }

    public void setNII(double nii) {
        NII = nii;
        _fireWasChanged();
    }

    public boolean getHingeI() {
        return hingeI;
    }

    public void setHingeI(boolean hingeI) {
        this.hingeI = hingeI;
        _fireWasChanged();
    }

    public boolean getHingeK() {
        return hingeK;
    }

    public void setHingeK(boolean hingeK) {
        this.hingeK = hingeK;
        _fireWasChanged();
    }

    public Node getNodeK() {
        return this.nk;
    }

    public void setNodeK(Node nk) {
        this.nk = nk;
        _fireWasChanged();
    }

    public double getLength() {
        return ni.getPoint().distance(nk.getPoint());
    }

    public double getW0() {
        return this.w0;
    }

    public void setW0(double w0) {
        this.w0 = w0;
        _fireWasChanged();
    }

    public double getPsi0() {
        return this.psi0;
    }

    public void setPsi0(double psi0) {
        this.psi0 = psi0;
        _fireWasChanged();
    }

    public Matrix getBj(double xs) {
        return Util.getBj(this, xs);
    }

    public Matrix[] getStiffMatrix() {

        Matrix[] stiff = new Matrix[7];
        // stiff[0] = Kik
        // stiff[1] = Kis
        // stiff[2] = Kks
        // stiff[3] = Kki
        // stiff[4] = E = E1^-1
        // stiff[5] = EE = E1^-1*E2
        // stiff[6] = EEE = E3 in case 4 = E3*E1^-1

        // Bj
        Matrix Bj = getBj(getLength());
        AbstractSection section = getSection();
        Matrix D = Util.getRotMatrix(ni, nk);
        double E = section.getE();
        double EI = section.getEI();

        double f0 = getLength() / (E * section.getAi());
        double f1 = Bj.get(0, 1);
        double f2 = -Bj.get(0, 2) / EI;
        double f3 = -Bj.get(0, 3) / EI;
        double f4 = Bj.get(0, 0);
        double f5 = -Bj.get(0, 1) / EI;
        double f6 = -NII * Bj.get(0, 1);

        double[][] _Fji = { { 1, 0, 0, 0, 0, f0 }, { 0, 1, f1, f2, f3, 0 },
                { 0, 0, f4, f5, f2, 0 }, { 0, 0, f6, f4, f1, 0 },
                { 0, 0, 0, 0, 1, 0 }, { 0, 0, 0, 0, 0, 1 } };
        Matrix Fji = new Matrix(_Fji);

        // Gji = D*Fji*D';
        Matrix Gji = D.times(Fji).times(D.transpose());

        // element cases
        // case on both parts fixed!
        if (!getHingeI() && !getHingeK()) {

            double[][] _E1 = { { Gji.get(2, 3), Gji.get(2, 4), Gji.get(2, 5) },
                    { -Gji.get(1, 3), -Gji.get(1, 4), -Gji.get(1, 5) },
                    { -Gji.get(0, 3), -Gji.get(0, 4), -Gji.get(0, 5) } };
            Matrix E1 = new Matrix(_E1);

            double[][] _E2 = {
                    { Gji.get(2, 2), -Gji.get(2, 1), -Gji.get(2, 0) },
                    { -Gji.get(1, 2), Gji.get(1, 1), Gji.get(1, 0) },
                    { -Gji.get(0, 2), Gji.get(0, 1), Gji.get(0, 0) } };
            Matrix E2 = new Matrix(_E2);

            double[][] _E3 = {
                    { -Gji.get(3, 3), -Gji.get(3, 4), -Gji.get(3, 5) },
                    { -Gji.get(4, 3), -Gji.get(4, 4), -Gji.get(4, 5) },
                    { -Gji.get(5, 3), -Gji.get(5, 4), -Gji.get(5, 5) } };
            Matrix E3 = new Matrix(_E3);

            Matrix Kik = E1.inverse();
            stiff[0] = Kik;

            Matrix Kis = Kik.uminus().times(E2);
            stiff[1] = Kis;

            Matrix Kks = E3.times(Kik);
            stiff[2] = Kks;

            Matrix Kki = Kik.transpose();
            stiff[3] = Kki;

            // E
            stiff[4] = E1.inverse();

            // EE
            stiff[5] = E1.inverse().times(E2);

            // EEE
            stiff[6] = E3;
        }
        // case 2 join on a point Ci
        else if (getHingeI() && !getHingeK()) {

            double[][] _E1 = { { Gji.get(2, 2), Gji.get(2, 4), Gji.get(2, 5) },
                    { -Gji.get(1, 2), -Gji.get(1, 4), -Gji.get(1, 5) },
                    { -Gji.get(0, 2), -Gji.get(0, 4), -Gji.get(0, 5) } };
            Matrix E1 = new Matrix(_E1);

            double[][] _E2 = { { 0, -Gji.get(2, 1), -Gji.get(2, 0) },
                    { 0, Gji.get(1, 1), Gji.get(1, 0) },
                    { 0, Gji.get(0, 1), Gji.get(0, 0) } };
            Matrix E2 = new Matrix(_E2);

            double[][] _E3 = {
                    { -Gji.get(3, 2), -Gji.get(3, 4), -Gji.get(3, 5) },
                    { -Gji.get(4, 2), -Gji.get(4, 4), -Gji.get(4, 5) },
                    { -Gji.get(5, 2), -Gji.get(5, 4), -Gji.get(5, 5) } };
            Matrix E3 = new Matrix(_E3);

            // chancel matrix
            double[][] _R = { { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
            Matrix R = new Matrix(_R);

            Matrix Kik = R.times(E1.inverse());
            stiff[0] = Kik;

            Matrix Kis = Kik.uminus().times(E2);
            stiff[1] = Kis;

            Matrix Kks = E3.times(E1.inverse());
            stiff[2] = Kks;

            Matrix Kki = Kik.transpose();
            stiff[3] = Kki;

            // E
            stiff[4] = E1.inverse();

            // EE
            stiff[5] = E1.inverse().times(E2);

            // EEE
            stiff[6] = E3;
        }
        // case 3 join on point Ck
        else if (!getHingeI() && getHingeK()) {

            double[][] _E1 = {
                    { -Gji.get(3, 3), -Gji.get(3, 4), -Gji.get(3, 5) },
                    { -Gji.get(1, 3), -Gji.get(1, 4), -Gji.get(1, 5) },
                    { -Gji.get(0, 3), -Gji.get(0, 4), -Gji.get(0, 5) } };
            Matrix E1 = new Matrix(_E1);

            double[][] _E2 = {
                    { -Gji.get(3, 2), Gji.get(3, 1), Gji.get(3, 0) },
                    { -Gji.get(1, 2), Gji.get(1, 1), Gji.get(1, 0) },
                    { -Gji.get(0, 2), Gji.get(0, 1), Gji.get(0, 0) } };
            Matrix E2 = new Matrix(_E2);

            double[][] _E3 = { { 0, 0, 0 },
                    { -Gji.get(4, 3), -Gji.get(4, 4), -Gji.get(4, 5) },
                    { -Gji.get(5, 3), -Gji.get(5, 4), -Gji.get(5, 5) } };
            Matrix E3 = new Matrix(_E3);

            // chancel matrix
            double[][] _R = { { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
            Matrix R = new Matrix(_R);

            Matrix Kik = E1.inverse().times(R);
            stiff[0] = Kik;

            Matrix Kis = E1.inverse().uminus().times(E2);
            stiff[1] = Kis;

            Matrix Kks = E3.times(Kik);
            stiff[2] = Kks;

            Matrix Kki = Kik.transpose();
            stiff[3] = Kki;

            // E
            stiff[4] = E1.inverse();

            // EE
            stiff[5] = E1.inverse().times(E2);

            // EEE
            stiff[6] = E3;
        }
        // case 4 join at point Ci and Ck
        else {

            double[][] _E1 = {
                    { -Gji.get(3, 2), -Gji.get(3, 4), -Gji.get(3, 5) },
                    { -Gji.get(1, 2), -Gji.get(1, 4), -Gji.get(1, 5) },
                    { -Gji.get(0, 2), -Gji.get(0, 4), -Gji.get(0, 5) } };
            Matrix E1 = new Matrix(_E1);

            double[][] _E2 = { { 0, Gji.get(3, 1), Gji.get(3, 0) },
                    { 0, Gji.get(1, 1), Gji.get(1, 0) },
                    { 0, Gji.get(0, 1), Gji.get(0, 0) } };
            Matrix E2 = new Matrix(_E2);

            double[][] _E3 = { { 0, 0, 0 },
                    { -Gji.get(4, 2), -Gji.get(4, 4), -Gji.get(4, 5) },
                    { -Gji.get(5, 2), -Gji.get(5, 4), -Gji.get(5, 5) } };
            Matrix E3 = new Matrix(_E3);

            // cancel matrix
            double[][] _R = { { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
            Matrix R = new Matrix(_R);

            Matrix Kik = R.times(E1.inverse()).times(R);
            stiff[0] = Kik;

            Matrix Kis = R.times(E1.inverse().uminus()).times(E2);
            stiff[1] = Kis;

            Matrix Kks = E3.times(Kik);
            stiff[2] = Kks;

            Matrix Kki = Kik.transpose();
            stiff[3] = Kki;

            // E
            stiff[4] = E1.inverse();

            // EE
            stiff[5] = E1.inverse().times(E2);

            // EEE
            stiff[6] = E3.times(E1.inverse());
        }
        return stiff;
    }

    public Matrix getTransferG(double xs) {

        // Bj
        Matrix Bj = getBj(xs);
        AbstractSection section = getSection();
        Matrix D = Util.getRotMatrix(ni, nk);
        double E = section.getE();
        double EI = section.getEI();

        double f0 = xs / (E * section.getAi());
        double f1 = Bj.get(0, 1);
        double f2 = -Bj.get(0, 2) / EI;
        double f3 = -Bj.get(0, 3) / EI;
        double f4 = Bj.get(0, 0);
        double f5 = -Bj.get(0, 1) / EI;
        double f6 = -NII * Bj.get(0, 1);

        double[][] _Fji = { { 1, 0, 0, 0, 0, f0 }, { 0, 1, f1, f2, f3, 0 },
                { 0, 0, f4, f5, f2, 0 }, { 0, 0, f6, f4, f1, 0 },
                { 0, 0, 0, 0, 1, 0 }, { 0, 0, 0, 0, 0, 1 } };
        Matrix Fji = new Matrix(_Fji);

        // Gji = D*Fji*D';
        Matrix Gji = D.times(Fji).times(D.transpose());

        // check if det is equal to -1
        double[][] _G = {
                { Gji.get(0, 0), Gji.get(0, 1), -Gji.get(0, 2), -Gji.get(0, 3),
                        -Gji.get(0, 4), -Gji.get(0, 5) },
                { Gji.get(1, 0), Gji.get(1, 1), -Gji.get(1, 2), -Gji.get(1, 3),
                        -Gji.get(1, 4), -Gji.get(1, 5) },
                { -Gji.get(2, 0), -Gji.get(2, 1), Gji.get(2, 2), Gji.get(2, 3),
                        Gji.get(2, 4), Gji.get(2, 5) },
                { Gji.get(3, 0), Gji.get(3, 1), -Gji.get(3, 2), -Gji.get(3, 3),
                        -Gji.get(3, 4), -Gji.get(3, 5) },
                { Gji.get(4, 0), Gji.get(4, 1), -Gji.get(4, 2), -Gji.get(4, 3),
                        -Gji.get(4, 4), -Gji.get(4, 5) },
                { Gji.get(5, 0), Gji.get(5, 1), -Gji.get(5, 2), -Gji.get(5, 3),
                        -Gji.get(5, 4), -Gji.get(5, 5) } };
        Matrix G = new Matrix(_G);

        return G;
    }

    // Listener
    public void wasChanged() {
        _fireWasChanged();
    }
}
