/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.model;

import java.awt.geom.Point2D;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rubeam.Util;
import rubeam.model.loads.AbstractElementLoad;
import rubeam.model.loads.AbstractNodeLoad;
import rubeam.model.sections.AbstractSection;
import Jama.Matrix;
import cib.util.geo.Vector2D;

/**
 * @author Emanuel Bombasaro, Christian Koch
 *
 */
public class Structure implements Serializable, IChangeListener {

    private static final long serialVersionUID = 1L;
    private static final int[] COL0 = new int[] { 0 };
    private static Structure structure = null;

    private List<Element> elements;
    private List<Node> nodes;

    // Load assignment
    private Map<Element, List<AbstractElementLoad>> elementLoadsMap; 
    private Map<Node, List<AbstractNodeLoad>> nodeLoadsMap;
    private Map<AbstractElementLoad, Element> loadElementMap;
    private Map<AbstractNodeLoad, Node> loadNodeMap;

    // Section assignment
    private List<AbstractSection> sections;
    private Map<Element, AbstractSection> elementSectionMap;

    private int visualizeSteps = 50;
    private int resultSteps = 10;

    private transient Matrix globK = null;
    private transient Matrix globS = null;
    private transient Matrix globV = null;
    private transient Matrix resultVg = null;
    private transient Map<Element, Matrix> elementResultsGlobal;
    private transient Map<Element, Matrix> elementResultsLocal;
    private transient Map<Element, Matrix> elementSik0s;
    private transient boolean valid = false;

    public Structure() {
        elements = new ArrayList<Element>();
        nodes = new ArrayList<Node>();
        sections = new ArrayList<AbstractSection>();
        sections.add(AbstractSection.getDefaultSection());
        elementLoadsMap = new HashMap<Element, List<AbstractElementLoad>>();
        nodeLoadsMap = new HashMap<Node, List<AbstractNodeLoad>>();
        loadElementMap = new LinkedHashMap<AbstractElementLoad, Element>();
        loadNodeMap = new LinkedHashMap<AbstractNodeLoad, Node>();
        elementSectionMap = new HashMap<Element, AbstractSection>();
        elementResultsGlobal = new HashMap<Element, Matrix>();
        elementResultsLocal = new HashMap<Element, Matrix>();
        elementSik0s = new HashMap<Element, Matrix>();
        structure = this;
    }

    public static Structure getStructure() {
        if (structure == null)
            structure = new Structure();
        return structure;
    }

    public void addElement(Element element) {
        if (!elements.contains(element)) {
            elements.add(element);
            Node ni = element.getNodeI();
            Node nk = element.getNodeK();
            if (!nodes.contains(ni))
                nodes.add(ni);
            if (!nodes.contains(nk))
                nodes.add(nk);
            // section
            if (!elementSectionMap.containsKey(element)) {
                putSection(element, AbstractSection.getDefaultSection());
            }
            element.setChangeListener(this);
            wasChanged();
        }
    }

    /**
     * Adds a node n.
     * @param n the node
     */
    public void addNode(Node n) {
        if (nodes.add(n)) {
            n.setChangeListener(this);
            wasChanged();
        }
    }

    public void addSection(AbstractSection s) {
        if (sections.add(s)) {
            s.setChangeListener(this);
            wasChanged();
        }
    }

    public void removeNode(Node node) {
        if (nodes.remove(node)) {
            // remove assigned elements?
            /*Set<Element> elmList = getElements(node);
            for (Element elm : elmList) {
                removeElement(elm);
            }*/
            // remove assigned node loads
            List<AbstractNodeLoad> loads = nodeLoadsMap.get(node);
            if (loads != null) {
                for (AbstractNodeLoad load : loads)
                    loadNodeMap.put(load, null);
            }
            nodeLoadsMap.remove(node);
            node.setChangeListener(null);
            wasChanged();
        }
    }

    public void removeElement(Element element) {
        if (elements.remove(element)) {
            List<AbstractElementLoad> loads = elementLoadsMap.get(element);
            if (loads != null) {
                for (AbstractElementLoad load : loads)
                    loadElementMap.put(load, null);
            }
            elementLoadsMap.remove(element);
            element.setChangeListener(null);
            wasChanged();
        }
    }

    public void removeSection(AbstractSection section) {
        if (section == AbstractSection.getDefaultSection())
            return;
        if (sections.remove(section)) {
            for (Element elm : elementSectionMap.keySet()) {
                if (elm.getSection() == section) {
                    putSection(elm, AbstractSection.getDefaultSection());
                }
            }
            section.setChangeListener(null);
            wasChanged();
        }
    }
    public void putSection(Element elm, AbstractSection section) {
        if (elements.contains(elm)
                && (sections.contains(section) || section == AbstractSection
                        .getDefaultSection())) {
            elementSectionMap.put(elm, section);
            wasChanged();
        }
    }
    public AbstractSection getSection(Element elm) {
        return elementSectionMap.get(elm);
    }

    public List<Element> readElements() {
        return Collections.unmodifiableList(elements);
    }
    public List<Node> readNodes() {
        return Collections.unmodifiableList(nodes);
    }
    public List<AbstractNodeLoad> readNodeLoads() {
        return readNodeLoads(AbstractNodeLoad.class);
    }
    public List<AbstractNodeLoad> readNodeLoads(Class<? extends AbstractNodeLoad> cl) {
        List<AbstractNodeLoad> nodeLoadList = new ArrayList<AbstractNodeLoad>();
        for (AbstractNodeLoad load : loadNodeMap.keySet()) {
            if (cl.isAssignableFrom(load.getClass()))
                nodeLoadList.add(load);
        }
        return nodeLoadList;
    }
    public List<AbstractElementLoad> readElementLoads() {
        return readElementLoads(AbstractElementLoad.class);
    }
    public List<AbstractElementLoad> readElementLoads(Class<? extends AbstractElementLoad> cl) {
        List<AbstractElementLoad> elmLoadList = new ArrayList<AbstractElementLoad>();
        for (AbstractElementLoad load : loadElementMap.keySet()) {
            if (cl.isAssignableFrom(load.getClass()))
                elmLoadList.add(load);
        }
        return elmLoadList;
    }
    public List<AbstractSection> readSections() {
        return readSections(AbstractSection.class);
    }
    public List<AbstractSection> readSections(Class<? extends AbstractSection> cl) {
        List<AbstractSection> sectionList = new ArrayList<AbstractSection>();
        for (AbstractSection sec : sections) {
            if (cl.isAssignableFrom(sec.getClass()))
                sectionList.add(sec);
        }
        return sectionList;
    }
    /**
     * @return the number of nodes in the structure
     */
    public int getNodeCount() {
        return nodes.size();
    }

    public Set<Element> getElements(Node node) {
        Set<Element> elementSet = new HashSet<Element>();
        for (Element element : elements) {
            if (element.getNodeI() == node || element.getNodeK() == node)
                elementSet.add(element);
        }
        return elementSet;
    }

    public boolean containsElement(Element element) {
        for (Element b : elements) {
            if ((b.getNodeI() == element.getNodeI() && b.getNodeK() == element
                    .getNodeK())
                    || (b.getNodeI() == element.getNodeK() && b.getNodeK() == element
                            .getNodeI()))
                return true;
        }
        return false;
    }

    public void addNodeLoad(AbstractNodeLoad load) {
        if (! loadNodeMap.containsKey(load)) {
            loadNodeMap.put(load, null);
            load.setChangeListener(this);
            wasChanged();
        }
    }
    public void removeNodeLoad(AbstractNodeLoad load) {
        if (loadNodeMap.containsKey(load)) {
            Node node = loadNodeMap.get(load);
            loadNodeMap.remove(load);
            List<AbstractNodeLoad> loads = nodeLoadsMap.get(node);
            if (loads != null)
                loads.remove(load);
            wasChanged();            
        }
    }
    public boolean containsNodeLoad(AbstractNodeLoad load) {
        return loadNodeMap.containsKey(load);
    }
    public Node getAssignedNode(AbstractNodeLoad load) {
        return loadNodeMap.get(load);
    }
    public void putNodeLoad(Node node, AbstractNodeLoad load) {
        if (nodes.contains(node) && loadNodeMap.containsKey(load)) {
            if (! nodeLoadsMap.containsKey(node)) { // not assigned load yet
                List<AbstractNodeLoad> loads = new ArrayList<AbstractNodeLoad>();
                nodeLoadsMap.put(node, loads);
                loads.add(load);
            }
            else { // already assigned load
                List<AbstractNodeLoad> loads = nodeLoadsMap.get(node);
                if (loads == null) {
                    loads = new ArrayList<AbstractNodeLoad>();
                }
                loads.add(load);
            }
            // remove old assignment
            Node oldNode = loadNodeMap.get(load);
            if (oldNode != null) {
                List<AbstractNodeLoad> oldLoads = nodeLoadsMap.get(oldNode);
                if (oldLoads != null) {
                    oldLoads.remove(load);
                }
            }
            loadNodeMap.put(load, node);
            wasChanged();
        }
    }
    public List<AbstractNodeLoad> getNodeLoads(Node node) {
        if (nodeLoadsMap.containsKey(node))
            return Collections.unmodifiableList(nodeLoadsMap.get(node)); // do not allow for change!
        else
            return null;
    }
    public void addElementLoad(AbstractElementLoad load) {
        if (! loadElementMap.containsKey(load)) {
            loadElementMap.put(load, null);
            load.setChangeListener(this);
            wasChanged();
        }
    }
    public void removeElementLoad(AbstractElementLoad load) {
        if (loadElementMap.containsKey(load)) {
            Element element = loadElementMap.get(load);
            loadElementMap.remove(load);
            List<AbstractElementLoad> loads = elementLoadsMap.get(element);
            if (loads != null)
                loads.remove(load);
            wasChanged();            
        }
    }
    public boolean containsElementLoad(AbstractElementLoad load) {
        return loadElementMap.containsKey(load);
    }
    public Element getAssignedElement(AbstractElementLoad load) {
        return loadElementMap.get(load);
    }
    public void putElementLoad(Element element, AbstractElementLoad load) {
        if (elements.contains(element) && loadElementMap.containsKey(load)) {

            if (! elementLoadsMap.containsKey(element)) { // not assigned load yet
                List<AbstractElementLoad> loads = new ArrayList<AbstractElementLoad>();
                elementLoadsMap.put(element, loads);
                loads.add(load);
            }
            else { // already assigned load 
                List<AbstractElementLoad> loads = elementLoadsMap.get(element);
                if (loads == null) {
                    loads = new ArrayList<AbstractElementLoad>();
                }
                loads.add(load);
            }
            // remove old assignment
            Element oldElement = loadElementMap.get(load);
            if (oldElement != null) {
                List<AbstractElementLoad> oldLoads = elementLoadsMap.get(oldElement);
                if (oldLoads != null) {
                    oldLoads.remove(load);
                }
            }
            loadElementMap.put(load, element);
            wasChanged();
        }
    }
    public List<AbstractElementLoad> getElementLoads(Element element) {
        if (elementLoadsMap.containsKey(element))
            return Collections.unmodifiableList(elementLoadsMap.get(element)); // do not allow for change!
        else
            return null;
    }

    public void setValid(boolean valid) {
        this.valid = valid;
    }

    public boolean isValid() {
        return this.valid;
    }

    public int getVisualizeSteps() {
        return visualizeSteps;
    }
    public void setVisualizeSteps(int visualizeSteps) {
        this.visualizeSteps = visualizeSteps;
        if (!(visualizeSteps >= 0)){
            System.err.println("visualizeSteps not valid, set to 0!");
            this.visualizeSteps = 0;
        }
        wasChanged();
    }
    public int getResultSteps() {
        return resultSteps;
    }
    public void setResultSteps(int resultSteps) {
        this.resultSteps = resultSteps;
        if (!(resultSteps >= 0)){
            System.err.println("resultSteps not valid, set to 0!");
            this.resultSteps = 0;
        }
        else if (resultSteps > 100) {
            System.err.println("resultSteps not valid, set to 100!");
            this.resultSteps = 100;
        }
    }

    // globK, globS, globV
    private Matrix[] _buildGlobals() {

        if (elements.isEmpty())
            throw new IllegalStateException("Empty structure!");

        int dim = 3 * nodes.size();
        globK = new Matrix(dim, dim, 0);
        globS = new Matrix(dim, 1, 0);
        globV = new Matrix(dim, 1, Double.NaN);
        Set<Node> markedNodes = new HashSet<Node>();

        int[] col0 = new int[] { 0 };

        for (Element element : elements) {
            Matrix[] mat = element.getStiffMatrix();
            Matrix Kik = mat[0];
            Matrix Kis = mat[1];
            Matrix Kks = mat[2];
            Matrix Kki = mat[3];
            Matrix E = mat[4];
            // Matrix EE = mat[5];
            Matrix EEE = mat[6];

            
            Node ni = element.getNodeI();
            int idx = nodes.indexOf(ni);
            Node nk = element.getNodeK();
            int kdx = nodes.indexOf(nk);
            int[] rows = new int[] { 3 * idx + 0, 3 * idx + 1, 3 * idx + 2 };
            int[] cols = new int[] { 3 * kdx + 0, 3 * kdx + 1, 3 * kdx + 2 };

            // globK
            // Kik
            Util.addSubMatrix(globK, rows, cols, Kik);
            // Kis
            Util.addSubMatrix(globK, rows, rows, Kis);
            // Kks
            Util.addSubMatrix(globK, cols, cols, Kks);
            // Kki
            Util.addSubMatrix(globK, cols, rows, Kki);

            // globV and node forces
            if (!markedNodes.contains(ni)) {
                Matrix sub = new Matrix(3, 1, Double.NaN);
                double[] constraints = ni.getContraints();
                if (! Double.isNaN(constraints[0]))
                    sub.set(0, 0, constraints[0]);
                if (! Double.isNaN(constraints[1]))
                    sub.set(1, 0, constraints[1]);
                if (! Double.isNaN(constraints[2]))
                    sub.set(2, 0, constraints[2]);
                globV.setMatrix(rows, col0, sub);

                // Node forces (M, V, H)
                List<AbstractNodeLoad> loadSet = getNodeLoads(ni); //ni.getLoadSet();
                if (loadSet != null) {
                    for (AbstractNodeLoad load : loadSet) {
                        Util.addSubMatrix(globS, rows, col0, load.getLoadMatrix());
                    }
                }

                markedNodes.add(ni);
            }
            if (!markedNodes.contains(nk)) {
                Matrix sub = new Matrix(3, 1, Double.NaN);
                double[] constraints = nk.getContraints();
                if (! Double.isNaN(constraints[0]))
                    sub.set(0, 0, constraints[0]);
                if (! Double.isNaN(constraints[1]))
                    sub.set(1, 0, constraints[1]);
                if (! Double.isNaN(constraints[2]))
                    sub.set(2, 0, constraints[2]);
                globV.setMatrix(cols, col0, sub);

                // Node forces (M, V, H)
                List<AbstractNodeLoad> loadSet = getNodeLoads(nk); //nk.getLoadSet();
                if (loadSet != null) {
                    for (AbstractNodeLoad load : loadSet) {
                        Util.addSubMatrix(globS, cols, col0, load.getLoadMatrix());
                    }
                }

                markedNodes.add(nk);
            }

            // Load elements per element
            Matrix L1 = new Matrix(3, 1, 0.0);
            Matrix L2 = new Matrix(3, 1, 0.0);
            Matrix Sik0 = new Matrix(3, 1, 0.0);
            Matrix Ski0 = new Matrix(3, 1, 0.0);

            List<AbstractElementLoad> loads = getElementLoads(element); //element.getLoadSet();
            if (loads != null) {
                for (AbstractElementLoad load : loads) {
                    Matrix[] L = load.getLoadVectors(element.getLength());

                    L1 = L1.plus(L[0]);
                    L2 = L2.plus(L[1]);
                }
                // case 1 & 2
                if (!element.getHingeI() && !element.getHingeK() || element.getHingeI()
                        && !element.getHingeK()) {
                    Sik0 = Kik.uminus().times(L1);
                    elementSik0s.put(element, Sik0);

                    Ski0 = L2.minus(Kks.times(L1));
                    // elementSki0s.put(element, Ski0);

                }
                // case 3
                else if (!element.getHingeI() && element.getHingeK()) {
                    Sik0 = E.uminus().times(L1);
                    elementSik0s.put(element, Sik0);

                    Ski0 = L2.plus(EEE.times(Sik0));
                    // elementSki0s.put(element, Ski0);

                }
                // case 4
                else {
                    // cancel matrix
                    double[][] _R = { { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
                    Matrix R = new Matrix(_R);

                    Sik0 = (R.times(E)).uminus().times(L1);
                    elementSik0s.put(element, Sik0);

                    Ski0 = L2.minus(EEE.times(L1));
                    // elementSki0s.put(element, Ski0);

                }
            }
            // Sik0
            Util.addSubMatrix(globS, rows, col0, Sik0);
            // Ski0
            Util.addSubMatrix(globS, cols, col0, Ski0);
        }

        return new Matrix[] { globK, globS, globV};
    }

    // resultVg
    public Matrix getResultVg() {
        if (! valid)
            solve();
        return resultVg;
    }

    public Matrix getNodeResult(Node node) {
        Matrix Nr = new Matrix(6, 1, 0.0);

        Matrix Vg = getResultVg();
        int nodeIdx = nodes.indexOf(node);
        if (nodeIdx != -1) {
            int[] rows = new int[] { 3 * nodeIdx, 3 * nodeIdx + 1,
                    3 * nodeIdx + 2 };
            Nr.setMatrix(new int[] { 0, 1, 2 }, new int[] { 0 }, Vg.getMatrix(rows, new int[] { 0 }));
        }

        Iterator<Element> ele = getElements(node).iterator();
        Element l_element;
        while (ele.hasNext()){
            l_element = (Element) ele.next();
            if ( l_element.getNodeI() == node) {
                Matrix add = new Matrix(6, 1, 0.0);
                add.setMatrix(new int[] { 3, 4, 5 }, new int[] { 0 }, getElementResultNi(l_element).getMatrix(new int[] { 3, 4, 5 }, new int[] { 0 }));
                Nr.plusEquals(add);
            }
            else if ( l_element.getNodeK() == node) {
                Matrix add = new Matrix(6, 1, 0.0);
                add.setMatrix(new int[] { 3, 4, 5 }, new int[] { 0 }, getElementResultNk(l_element).getMatrix(new int[] { 3, 4, 5 }, new int[] { 0 }));
                Nr.plusEquals(add);
            }
            else {
                throw new IllegalStateException();
            }
        }; 

        Matrix add = new Matrix(6, 1, 0.0);
        List<AbstractNodeLoad> loadSet = getNodeLoads(node);
        if (loadSet != null) {
            for (AbstractNodeLoad load : loadSet) {
                add.setMatrix(new int[] { 3, 4, 5 }, new int[] { 0 }, load.getLoadMatrix());
                Nr.plusEquals(add);
            }
        }
        Nr.set(3, 0, - Nr.get(3,0));

        return Nr;

    }

    public Matrix getNodeResultV(Node node) {
        Matrix Vni = null;
        Matrix Vg = getResultVg();
        int nodeIdx = nodes.indexOf(node);
        if (nodeIdx != -1) {
            int[] rows = new int[] { 3 * nodeIdx, 3 * nodeIdx + 1,
                    3 * nodeIdx + 2 };
            Vni = Vg.getMatrix(rows, new int[] { 0 });
        }

        return Vni;
    }
    // solve
    public void solve() {
        clean();
        try {
            Matrix[] globals = _buildGlobals();
            resultVg = EquationSystem.solve(globals);
            valid = true;
        }
        catch (Exception e) {
            valid = false;
            throw new IllegalStateException(e.getMessage());
        }

    }

    // max values
    public double getMaxResultVg() {
        Matrix resultVg = getResultVg();
        double max = 0.0;
        for (int i=0; i < resultVg.getRowDimension(); i++) {
            double val = resultVg.get(i, 0);
            if (Math.abs(val) > max)
                max = Math.abs(val);
        }
        return max;
    }

    public double getDeltaPhi(Node node) {
        if (valid) {
            int nodeIdx = nodes.indexOf(node);
            if (nodeIdx != -1)
                return getResultVg().get(3 * nodeIdx, 0);
            else
                return 0.0;
        }
        else
            return Double.NaN;
    }

    public double getDeltaY(Node node) {
        if (valid) {
            int nodeIdx = nodes.indexOf(node);
            if (nodeIdx != -1)
                return getResultVg().get(3 * nodeIdx + 1, 0);
            else
                return 0.0;
        }
        else
            return Double.NaN;
    }

    public double getDeltaX(Node node) {
        if (valid) {
            int nodeIdx = nodes.indexOf(node);
            if (nodeIdx != -1)
                return getResultVg().get(3 * nodeIdx + 2, 0);
            else
                return 0.0;
        }
        else
            return Double.NaN;
    }

    public Matrix getElementResultNi(Element element) {
        if (!valid)
            return null;

        // [Kik, Kis, Kks, Kki, E, EE, EEE]
        Matrix[] stiffs = element.getStiffMatrix();
        Matrix Kik = stiffs[0];
        Matrix Kis = stiffs[1];
        // Matrix Kks = stiffs[2];
        // Matrix Kki = stiffs[3];
        Matrix E = stiffs[4];
        Matrix EE = stiffs[5];
        // Matrix EEE = stiffs[6];

        Matrix L1 = new Matrix(3, 1, 0.0);
        Matrix L2 = new Matrix(3, 1, 0.0);
        List<AbstractElementLoad> loads = getElementLoads(element); //element.getLoadSet();
        if (loads != null) {
            for (AbstractElementLoad load : loads) {
                Matrix[] L = load.getLoadVectors(element.getLength());

                L1 = L1.plus(L[0]);
                L2 = L2.plus(L[1]);
            }
        }

        Matrix Vik;
        Matrix Vki;
        Matrix VSik;
        Matrix Sik;
        Matrix Sik0;
        Matrix Yik = new Matrix(6, 1, 0.0);

        Vik = getNodeResultV(element.getNodeI());
        Vki = getNodeResultV(element.getNodeK());

        if (elementSik0s.containsKey(element))
            Sik0 = elementSik0s.get(element);
        else
            Sik0 = new Matrix(3, 1);

        // case 1 and 3
        if (!element.getHingeI()) {
            Sik = Kis.times(Vik).plus(Kik.times(Vki)).plus(Sik0);

            Yik.setMatrix(new int[] { 2, 1, 0 }, COL0, Vik);
            Yik.setMatrix(new int[] { 3, 4, 5 }, COL0, Sik);
        }
        // 2
        else if (element.getHingeI() && !element.getHingeK()) {
            VSik = EE.uminus().times(Vik).plus(E.times(Vki)).minus(E.times(L1));
            Yik.set(0, 0, Vik.get(2, 0));
            Yik.set(1, 0, Vik.get(1, 0));
            Yik.set(2, 0, VSik.get(0, 0));
            Yik.set(3, 0, 0.0);
            Yik.set(4, 0, VSik.get(1, 0));
            Yik.set(5, 0, VSik.get(2, 0));

        }
        // 4
        else if (element.getHingeI() && element.getHingeK()) {
            double[][] _R = { { 0, 0, 0 }, { 0, 1, 0 }, { 0, 0, 1 } };
            Matrix R = new Matrix(_R);

            VSik = EE.uminus().times(Vik).plus(E.times(R).times(Vki)).minus(
                    E.times(L1));
            Yik.set(0, 0, Vik.get(2, 0));
            Yik.set(1, 0, Vik.get(1, 0));
            Yik.set(2, 0, VSik.get(0, 0));
            Yik.set(3, 0, 0.0);
            Yik.set(4, 0, VSik.get(1, 0));
            Yik.set(5, 0, VSik.get(2, 0));

        }
        else {
            throw new IllegalStateException();
        }

        return Yik;
    }

    public Matrix getElementResultNk(Element element) {

        Matrix Yik = getElementResultNi(element);
        Matrix G = element.getTransferG(element.getLength());
        Matrix L = Util.getTransferL(getElementLoads(element), element.getLength());

        Matrix Yki = G.times(Yik).plus(L);

        return Yki;
    }

    public Matrix[] getElementResults(Element element, int interSteps) {

        if (!valid)
            return null;

        Matrix Yik = getElementResultNi(element);
        Matrix Yki = getElementResultNk(element);

        Matrix[] results = new Matrix[2];
        Matrix resultsGlobal = new Matrix(6, interSteps+2);
        Matrix resultsLocal = new Matrix(6, interSteps+2);

        Point2D Pi = element.getNodeI().getPoint();
        Point2D Pk = element.getNodeK().getPoint();
        Matrix D = Util.getRotMatrix(element.getNodeI(), element.getNodeK());
        Vector2D vec = new Vector2D(Pi, Pk);
		// u, w, phi, M, R , N
        int[] rows = new int[] { 0, 1, 2, 3, 4, 5 };
        if (interSteps > 0){
            for (int i = 1; i <= interSteps; i++) {
                Vector2D step = new Vector2D(vec);
                double scale = ((double) (i)) / ((double) interSteps+1);
                step.scaleBy(scale);
                double xs = step.length();

                Matrix G = element.getTransferG(xs);
                Matrix L = Util.getTransferL(getElementLoads(element), xs);

                Matrix Yhc = G.times(Yik).plus(L);
                Matrix Zhc = D.transpose().times(Yhc).uminus();
                Zhc.set(1, 0, - Zhc.get(1, 0));
                resultsGlobal.setMatrix(rows, new int[] { i }, Yhc);
                resultsLocal.setMatrix(rows, new int[] { i }, Zhc);

            }
        }
        Yik.set(3, 0, -Yik.get(3, 0));
        Yik.set(4, 0, -Yik.get(4, 0));
        Yik.set(5, 0, -Yik.get(5, 0));
        resultsGlobal.setMatrix(rows, new int[] { 0 }, Yik);
        resultsGlobal.setMatrix(rows, new int[] { interSteps+1 }, Yki);
        resultsLocal.setMatrix(rows, new int[] { 0 }, D.transpose().times(Yik).uminus());
        Matrix Zki=D.transpose().times(Yki).uminus();
        Zki.set(1, 0, -Zki.get(1, 0));
        resultsLocal.setMatrix(rows, new int[] { interSteps+1 }, Zki);

        results[0] = resultsGlobal;
        results[1] = resultsLocal;

        elementResultsGlobal.put(element, resultsGlobal);
        elementResultsLocal.put(element, resultsLocal);
        
        return results;

    }

    // Listener
    public void wasChanged() {
        valid = false;
    }

    // Serialization
    private void readObject(ObjectInputStream in) throws IOException,
    ClassNotFoundException {

        // Deserialize non-transient attributes
        in.defaultReadObject();

        elementResultsGlobal = new HashMap<Element, Matrix>();
        elementResultsLocal = new HashMap<Element, Matrix>();
        elementSik0s = new HashMap<Element, Matrix>();
        structure = this;
        valid = false;
    }

    public void clean() {

        globK = null;
        globS = null;
        globV = null;
        resultVg = null;
        elementSik0s.clear();
        elementResultsGlobal.clear();
        elementResultsLocal.clear();
        valid = false;

    }

    public void clear() {
        nodes.clear();
        elements.clear();
        sections.clear();
        sections.add(AbstractSection.getDefaultSection());
        nodeLoadsMap.clear();
        loadNodeMap.clear();
        elementLoadsMap.clear();
        loadElementMap.clear();
        elementSectionMap.clear();

        globK = null;
        globS = null;
        globV = null;
        resultVg = null;
        elementSik0s.clear();
        elementResultsGlobal.clear();
        elementResultsLocal.clear();
        valid = false;
    }

    public String toString() {
        return Util.structureToString(this);
    }
    public String resultsToString() {
        return Util.resultsToString(this);
    }
    public String exportToString() {
        return Util.exportToString(this);
    }
    public String toHTML() {
        return Util.structureToHTML(this);
    }
    public String resultsToHTML() {
        return Util.resultsToHTML(this);
    }
}
