/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmps;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.font.TextAttribute;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.Iterator;
import java.util.NoSuchElementException;

import rubeam.RuBeam;
import rubeam.model.Element;
import rubeam.model.Node;
import rubeam.model.Structure;
import Jama.Matrix;
import cib.cad.db.comp.Component;
import cib.cad.db.comp.ComponentAdapter;
import cib.cad.db.feature.BooleanFeature;
import cib.cad.db.feature.DoubleFeature;
import cib.cad.db.feature.Feature;
import cib.cad.db.feature.StringFeature;
import cib.util.AttributedShape;
import cib.util.AttributedText;
import cib.util.CoordSpace;
import cib.util.coll.NamedListIterator;
import cib.util.coll.NamedListIteratorAdapter;
import cib.util.geo.Geo2D;
import cib.util.geo.NoIntersectionException;
import cib.util.geo.Vector2D;

/**
 * @author Christian Koch
 */
public class ComponentElement extends ComponentAdapter {

    private static final long serialVersionUID = 0L;
    protected static final double TEXT_DIST = 1.5; // [mm]
    public static final Color RESULT_COLOR = new Color(50, 180, 0);
    public static final double RESULT_SHAPE_LINE_WIDTH = .1; // [mm]
    private Element m_element;

    public ComponentElement(Element beam) {
        m_element = beam;
    }

    public Object clone() {
        ComponentElement clone = (ComponentElement) super.clone();
        clone.m_element = (Element) this.m_element.clone();
        return clone;
    }

    public void assign(Component rhs) {
        ComponentElement rhsComp = (ComponentElement) rhs;
        Element rhsElement = rhsComp.m_element;
        m_element.setNodeI(rhsElement.getNodeI());
        m_element.setNodeK(rhsElement.getNodeK());
        m_element.setName(rhsElement.getName());
        m_element.setHingeI(rhsElement.getHingeI());
        m_element.setHingeK(rhsElement.getHingeK());
        m_element.setW0(rhsElement.getW0());
        m_element.setPsi0(rhsElement.getPsi0());
        RuBeam.getStructure().putSection(m_element, rhsElement.getSection());
        super.assign(rhs);
    }

    public Element getElement() {
        return m_element;
    }

    public Point2D getStart() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = m_element.getNodeI().getPoint();
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public double getStartX() {
        return getStart().getX();
    }

    public double getStartY() {
        return getStart().getY();
    }

    public Point2D getEnd() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = m_element.getNodeK().getPoint();
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    private Point2D _getInterPoint(double scale) {
        Point2D start = m_element.getNodeI().getPoint();
        Point2D end = m_element.getNodeK().getPoint();
        Vector2D vec = new Vector2D(start, end);
        vec.scaleBy(scale);
        Vector2D inter = new Vector2D(start);
        inter.add(vec);
        return inter;
    }

    public double getEndX() {
        return getEnd().getX();
    }

    public double getEndY() {
        return getEnd().getY();
    }

    private AttributedShape getBeamShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();
        double delta;
        Point2D sp = getStart();
        Point2D ep = getEnd();
        Line2D line = new Line2D.Double(getStart(), getEnd());
        // start
        if (m_element.getHingeI())
            delta = ComponentNode.NODE_RADIUS_INNER * scale;
        else
            delta = ComponentNode.NODE_RADIUS * scale;
        Ellipse2D circle = new Ellipse2D.Double(sp.getX() - delta, sp.getY()
                - delta, delta + delta, delta + delta);
        try {
            Iterator<Point2D> it = Geo2D.intersection(line, circle);
            sp = it.next();
            // end
            if (m_element.getHingeK())
                delta = ComponentNode.NODE_RADIUS_INNER * scale;
            else
                delta = ComponentNode.NODE_RADIUS * scale;
            circle = new Ellipse2D.Double(ep.getX() - delta, ep.getY() - delta,
                    delta + delta, delta + delta);
            it = Geo2D.intersection(line, circle);
            ep = it.next();
            line = new Line2D.Double(sp, ep);
        }
        catch (NoIntersectionException e) {
            // do nothing
        }
        AttributedShape as = _attributeShape(new AttributedShape(line));
        BasicStroke stroke = (BasicStroke) as.getStroke();
        as.setStroke(new BasicStroke(stroke.getLineWidth(),
                BasicStroke.CAP_BUTT, stroke.getLineJoin(), stroke
                        .getMiterLimit(), stroke.getDashArray(), stroke
                        .getDashPhase()));
        return as;
    }

    private AttributedShape getResultShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

        Structure struct = RuBeam.getStructure();

        if (struct.isValid()) {
            Path2D path = new Path2D.Double();
            
            // [u, w, phi, M, V, H]
            Matrix[] results = struct.getElementResults(
                    m_element, struct.getVisualizeSteps());
            double max = struct.getMaxResultVg();
            if (max == 0.0)
                max = 1.0;
            
            // stepcount and scale
            int stepcount = results[0].getColumnDimension();
            double sf = 1.0 / (stepcount-1);
  
            Point2D pj = new Point2D.Double();
            for (int i = 0; i < stepcount; i++) {

                double dxj = results[0].get(0, i);// * 1e3;
                double dyj = results[0].get(1, i);// * 1e3;
                Point2D iP = _getInterPoint(sf * (i));
                double xj = iP.getX()
                        - (dxj * scale / max * ComponentNode.NODE_MAX_DISP);
                double yj = iP.getY()
                        + (dyj * scale / max * ComponentNode.NODE_MAX_DISP);
                pj.setLocation(xj, yj);
                try {
                    pj = ucrd.inverseTransform(pj, null);
                } catch (NoninvertibleTransformException e) {
                    // Should never happen
                }
                if (i == 0)
                    path.moveTo(pj.getX(), pj.getY());
                else
                    path.lineTo(pj.getX(), pj.getY());
            }
            AttributedShape as = _attributeShape(new AttributedShape(path));
            as.setDrawPaint(RESULT_COLOR);
            BasicStroke stroke = (BasicStroke) as.getStroke();
            as.setStroke(new BasicStroke(
                    (float) (RESULT_SHAPE_LINE_WIDTH * scale), stroke
                            .getEndCap(), stroke.getLineJoin(), stroke
                            .getMiterLimit(), stroke.getDashArray(), stroke
                            .getDashPhase()));
            return as;
        }
        else {
            Point2D p1 = getStart();
            Point2D p2 = getEnd();
            Line2D line = new Line2D.Double(p1, p2);
            AttributedShape as = _attributeShape(new AttributedShape(line)); 
            as.setDrawPaint(RESULT_COLOR);
            return as;
        }
    }

    public NamedListIterator<AttributedShape> shapeIterator() {
        return new NamedListIteratorAdapter<AttributedShape>() {
            protected int _size() {
                if (RuBeam.getStructure().isValid())
                    return 2;
                else 
                    return 1;
            }

            protected AttributedShape _get(int index) {
                switch (index) {
                case 0:
                    return getBeamShape();
                case 1:
                    return getResultShape();
                default:
                    throw new IllegalArgumentException();
                }
            }
        };
    }

    private AttributedText getText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        Vector2D vec = new Vector2D(getStart(), getEnd());
        Vector2D ext = new Vector2D(vec);
        ext.normalize();
        ext.left();
        ext.scaleBy(TEXT_DIST * scale);
        vec.scaleBy(0.5);
        Vector2D vCtr = new Vector2D(getStart());
        vCtr.add(vec);
        vCtr.add(ext);
        Point2D ptW = u2w.transform(vCtr, null);

        AffineTransform trf = AffineTransform.getTranslateInstance(ptW.getX(),
                ptW.getY());

        AttributedString string = new AttributedString(m_element.getName());
        AttributedText text = new AttributedText(string, trf);
        text.setJustification(AttributedText.JUST_CENTER_CENTER);
        text.transformBy(w2u);
        text = _attributeText(text);
        AttributedString as = text.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(ComponentNode.NODE_TEXT_SIZE 
                            *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        return text;
    }

    public NamedListIterator<AttributedText> textIterator() {
        return NamedListIteratorAdapter.singletonNamedListIterator(getText());
    }

    public void transformBy(AffineTransform mat) {
        // do nothing
    }

    public Node getNodeI() {
        return m_element.getNodeI();
    }

    public Node getNodeK() {
        return m_element.getNodeK();
    }

    // Features
    private static final String[] FEATURE_NAMES = { 
            "NODE_I", // 0
            "NODE_K", // 1
            "LENGTH", // 2
            "NAME", // 3
            "HINGE_I", // 4
            "HINGE_K", // 5
            "SECTION"}; // 6

    public boolean hasFeature(String name) {
        for (int i = 0; i < FEATURE_NAMES.length; i++) {
            if (name.equals(FEATURE_NAMES[i]))
                return true;
        }
        return super.hasFeature(name);
    }

    public Feature getFeature(String name) {

        if (name.equals(FEATURE_NAMES[0])) {
            StringFeature sf = new StringFeature(FEATURE_NAMES[0], m_element
                    .getNodeI().getName());
            sf.setChangeable(false);
            return sf;
        }
        else if (name.equals(FEATURE_NAMES[1])) {
            StringFeature sf = new StringFeature(FEATURE_NAMES[1], m_element
                    .getNodeK().getName());
            sf.setChangeable(false);
            return sf;
        }
        else if (name.equals(FEATURE_NAMES[2])) {
            DoubleFeature df = new DoubleFeature(FEATURE_NAMES[2], m_element
                    .getLength());
            df.setChangeable(false);
            return df;
        }
        else if (name.equals(FEATURE_NAMES[3])) {
            StringFeature sf = new StringFeature(FEATURE_NAMES[3],
                    (String) getElement().getName());
            return sf;
        }
        else if (name.equals(FEATURE_NAMES[4])) {
            return new BooleanFeature(FEATURE_NAMES[4], m_element.getHingeI());
        }
        else if (name.equals(FEATURE_NAMES[5])) {
            return new BooleanFeature(FEATURE_NAMES[5], m_element.getHingeK());
        }
        else if (name.equals(FEATURE_NAMES[6])) {
            StringFeature sf = new StringFeature(FEATURE_NAMES[6], m_element.getSection().getName());
            sf.setChangeable(false);
            return sf;
        }
        else
            return super.getFeature(name);
    }

    public void setFeature(Feature feature) {
        if (feature.getName().equals(FEATURE_NAMES[3])) {
            m_element.setName(feature.valueToString());
        }
        else if (feature.getName().equals(FEATURE_NAMES[4])) {
            m_element.setHingeI(Boolean.parseBoolean(feature.valueToString()));
        }
        else if (feature.getName().equals(FEATURE_NAMES[5])) {
            m_element.setHingeK(Boolean.parseBoolean(feature.valueToString()));
        }
        else {
            super.setFeature(feature);
            return;
        }
        _notifyWasChanged();
    }

    public void addFeature(Feature feature) {
        throw new UnsupportedOperationException();
    }

    public Iterator<Feature> featureIterator() {

        final Iterator<Feature> itThis = new Iterator<Feature>() {
            private int m_index = 0;

            public boolean hasNext() {
                return m_index < FEATURE_NAMES.length;
            }

            public Feature next() {
                if (hasNext())
                    return getFeature(FEATURE_NAMES[m_index++]);
                throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };

        final Iterator<Feature> itBase = super.featureIterator();

        return new Iterator<Feature>() {
            public boolean hasNext() {
                return itBase.hasNext() | itThis.hasNext();
            }

            public Feature next() {
                if (itBase.hasNext())
                    return itBase.next();
                else if (itThis.hasNext())
                    return itThis.next();
                else
                    throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

}
