/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmps;

import java.awt.Color;
import java.awt.geom.AffineTransform;
import java.awt.geom.Path2D;
import java.awt.geom.Line2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.NoSuchElementException;

import rubeam.model.Node;
import rubeam.model.loads.AbstractNodeLoad;
import rubeam.model.loads.NodalLoadAng;
import cib.cad.db.comp.Component;
import cib.cad.db.comp.ComponentAdapter;
import cib.cad.db.feature.AngleFeature;
import cib.cad.db.feature.DoubleFeature;
import cib.cad.db.feature.Feature;
import cib.util.AttributedShape;
import cib.util.CoordSpace;
import cib.util.coll.NamedListIterator;
import cib.util.coll.NamedListIteratorAdapter;
import cib.util.geo.Vector2D;

/**
 * @author Christian Koch
 */
public class ComponentNodalLoadAng extends ComponentAdapter implements ComponentNodalLoad {

    private static final long serialVersionUID = 0L;
    private NodalLoadAng m_load;
    public static final double NODE_RADIUS = .4; // [mm]
    public static final double NODE_RADIUS_INNER = 0.8; // [mm]
    public static final double NODE_SHAPE_LINE_WIDTH = .1; // [mm]
    public static final double NODE_SUPPORT_LINE_WIDTH = .4; // [mm]
    public static final double NODE_CONN_LINE_WIDTH = .15; // [mm]
    public double m_x, m_y;

    // Constructors
    public ComponentNodalLoadAng(NodalLoadAng load, Point2D pt) {
        this.m_load = load;
        Node node = load.getNode();
        if (node != null) {
            setNode(node.getPoint());
        }
        else {
            setNode(pt);
        }
    }

    public Object clone() {
        ComponentNodalLoadAng clone = (ComponentNodalLoadAng) super.clone();
        clone.m_load = (NodalLoadAng) this.m_load.clone();
        return clone;
    }

    public void assign(Component rhs) {
        ComponentNodalLoadAng rhsComp = (ComponentNodalLoadAng) rhs;
        NodalLoadAng rhsLoad = rhsComp.m_load;
        m_load.setP(rhsLoad.getP());
        m_load.setAngle(rhsLoad.getAngle());
        super.assign(rhs);
    }

    public Point2D getPoint() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = null;
        if (m_load.getNode() == null)
            ptU = new Point2D.Double(m_x, m_y);
        else {
            ptU = m_load.getNode().getPoint();
            m_x = ptU.getX();
            m_y = ptU.getY();
        }
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public void setNode(double x, double y) {
        setNode(new Point2D.Double(x, y));
    }
    public void setNode(Point2D pt) {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D _pt = ucrd.transform(pt, null);
        m_x = _pt.getX();
        m_y = _pt.getY();

        // Notification
        _notifyWasChanged();
    }

    public AbstractNodeLoad getLoad() {
        return m_load;
    }

    private AttributedShape getLoadShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

        Point2D ctr = getPoint();
        double _delta = NODE_RADIUS * 2 * scale;

        Path2D path = new Path2D.Double();

        AttributedShape as = _attributeShape(new AttributedShape(path));
        if (m_load.getNode() != null)
            as.setDrawPaint(ComponentNodalLoad.NODALLOAD_COLOR);
        else
            as.setDrawPaint(Color.LIGHT_GRAY);

                double v = m_load.getP();
                double ang = m_load.getAngle();
                try {
                    Point2D _X = ucrd.inverseTransform(Vector2D.X_UNIT, null);
                    ang = Vector2D.X_UNIT.getAngleCCW(new Vector2D(_X)) + ang;
                } catch (NoninvertibleTransformException e) {
                    // Should never happen
                }
                Vector2D vec = new Vector2D(ang);
                vec.scaleBy(_delta + _delta);
                Vector2D cv = new Vector2D(ctr);
                cv.add(vec);
                Vector2D cv2 = new Vector2D(cv);
                cv2.add(vec).add(vec);
                Line2D line = new Line2D.Double(cv, cv2);
                path.append(line, false);
                if (v > 0) {
                    Vector2D ar1 = new Vector2D(ang + Math.PI
                            - Math.toRadians(30));
                    ar1.scaleBy(_delta);
                    Vector2D p1 = new Vector2D(cv2);
                    p1.add(ar1);
                    Line2D l1 = new Line2D.Double(cv2, p1);
                    path.append(l1, false);
                    Vector2D ar2 = new Vector2D(ang + Math.PI
                            + Math.toRadians(30));
                    ar2.scaleBy(_delta);
                    Vector2D p2 = new Vector2D(cv2);
                    p2.add(ar2);
                    Line2D l2 = new Line2D.Double(cv2, p2);
                    path.append(l2, false);
                }
                else {
                    Vector2D ar1 = new Vector2D(ang + Math.toRadians(30));
                    ar1.scaleBy(_delta);
                    Vector2D p1 = new Vector2D(cv);
                    p1.add(ar1);
                    Line2D l1 = new Line2D.Double(cv, p1);
                    path.append(l1, false);
                    Vector2D ar2 = new Vector2D(ang - Math.toRadians(30));
                    ar2.scaleBy(_delta);
                    Vector2D p2 = new Vector2D(cv);
                    p2.add(ar2);
                    Line2D l2 = new Line2D.Double(cv, p2);
                    path.append(l2, false);
                }
            //}
//        }
        return as;
    }

    public NamedListIterator<AttributedShape> shapeIterator() {
        return NamedListIteratorAdapter
                .singletonNamedListIterator(getLoadShape());
    }
   
    public Point2D getControlPoint(int name) throws IllegalArgumentException {
        switch (name) {
        case 0:
            return getPoint();
        default:
            throw new IllegalArgumentException();
        }
    }

    public void setControlPoint(Point2D pnt, int name)
            throws UnsupportedOperationException {
        Point2D p = pnt;
        switch (name) {
        case 0:
            if (m_load.getNode() == null)
                setNode(p);
            break;
        default:
            throw new IllegalArgumentException();
        }
    }

    public NamedListIterator<Point2D> controlPointIterator() {
        return new NamedListIteratorAdapter<Point2D>() {
            protected int _size() {
                return 1;
            }

            protected Point2D _get(int index) {
                switch (index) {
                case 0:
                    return getPoint();
                default:
                    throw new IllegalArgumentException();
                }
            }

            protected void _set(int index, Point2D value) {
                Point2D p = (Point2D) value;
                switch (index) {
                case 0:
                    if (m_load.getNode() == null)
                        setNode(p.getX(), p.getY());
                    break;
                default:
                    throw new IllegalArgumentException();
                }
            }
        };
    }

    public void transformBy(AffineTransform mat) {
        if (m_load.getNode() == null) {
            // Transform point
            Point2D pt = mat.transform(getPoint(), null);
            setNode(pt);
        }
    }

    // Features
    private static final String[] FEATURE_NAMES = { 
            "VALUE", // 0
            "ANGLE", // 1
   };

    public boolean hasFeature(String name) {
        for (int i = 0; i < FEATURE_NAMES.length; i++) {
            if (name.equals(FEATURE_NAMES[i]))
                return true;
        }
        return super.hasFeature(name);
    }

    public Feature getFeature(String name) {
        if (name.equals(FEATURE_NAMES[0]))
            return new DoubleFeature(FEATURE_NAMES[0], m_load.getP());
        else if (name.equals(FEATURE_NAMES[1])) {
            return new AngleFeature(FEATURE_NAMES[1], m_load.getAngle());
        }
        else
            return super.getFeature(name);
    }

    public void setFeature(Feature feature) {
        // Get parameter from feature
        if (feature.getName().equals(FEATURE_NAMES[0])) {
            m_load.setP(((DoubleFeature)feature).getDoubleValue());
        }
        else if (feature.getName().equals(FEATURE_NAMES[1])) {
            m_load.setAngle(((AngleFeature)feature).getAngle());
        }
        else {
            super.setFeature(feature);
            return;
        }
        _notifyWasChanged();
    }

    public Iterator<Feature> featureIterator() {

        final Iterator<Feature> itThis = new Iterator<Feature>() {
            private int m_index = 0;

            public boolean hasNext() {
                return m_index < FEATURE_NAMES.length;
            }

            public Feature next() {
                if (hasNext())
                    return getFeature(FEATURE_NAMES[m_index++]);
                throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };

        final Iterator<Feature> itBase = super.featureIterator();

        return new Iterator<Feature>() {
            public boolean hasNext() {
                return itBase.hasNext() | itThis.hasNext();
            }

            public Feature next() {
                if (itBase.hasNext())
                    return itBase.next();
                else if (itThis.hasNext())
                    return itThis.next();
                else
                    throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }
    
}
