/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmps;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Shape;
import java.awt.font.TextAttribute;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Path2D;
import java.awt.geom.Line2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.Set;

import rubeam.RuBeam;
import rubeam.model.Element;
import rubeam.model.Node;
import rubeam.model.Structure;
import cib.cad.db.comp.Component;
import cib.cad.db.comp.ComponentAdapter;
import cib.cad.db.feature.DoubleFeature;
import cib.cad.db.feature.Feature;
import cib.cad.db.feature.Point2DFeature;
import cib.cad.db.feature.StringFeature;
import cib.util.AttributedShape;
import cib.util.AttributedText;
import cib.util.CoordSpace;
import cib.util.coll.NamedListIterator;
import cib.util.coll.NamedListIteratorAdapter;

/**
 * @author Christian Koch
 */
public class ComponentNode extends ComponentAdapter {

    private static final long serialVersionUID = 0L;
    private Node m_node;
    public static final double NODE_RADIUS = .4; // [mm]
    public static final double NODE_RADIUS_INNER = 0.8; // [mm]
    public static final double NODE_TEXT_OFFSET = 0.8; // [mm]
    public static final double NODE_SHAPE_LINE_WIDTH = .1; // [mm]
    public static final double NODE_SUPPORT_LINE_WIDTH = .4; // [mm]
    public static final double NODE_CONN_LINE_WIDTH = .15; // [mm]
    public static final double NODE_RESULT_OFFSET = .5; // [mm]
    public static final double NODE_TEXT_SIZE = 2;
    public static final double NODE_MAX_DISP = 10; // [mm] in paper units
    public static final Color SUPPORT_COLOR = Color.RED;

    // Constructors
    public ComponentNode(Node node) {
        this.m_node = node;
    }

    public Object clone() {
        ComponentNode clone = (ComponentNode) super.clone();
        clone.m_node = (Node) this.m_node.clone();
        /*List<NodeLoad> loadSet = clone.m_node.getLoadSet();
        if (loadSet != null) {
            for (NodeLoad nl : loadSet) {
                nl.setNode(clone.m_node);
            }
        }*/
        return clone;
    }

    public void assign(Component rhs) {
        ComponentNode rhsComp = (ComponentNode) rhs;
        Node rhsNode = rhsComp.m_node;
        double[] crds = rhsNode.getCoordinates();
        m_node.setCoordinates(crds[0], crds[1]);
        m_node.setName(rhsNode.getName());
        double[] conts = rhsNode.getContraints();
        m_node.setContraints(conts[0], conts[1], conts[2]);
        super.assign(rhs);
    }

    public Point2D getCenter() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = m_node.getPoint();
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public double getCenterX() {
        return getCenter().getX();
    }

    public double getCenterY() {
        return getCenter().getY();
    }

    public void setNode(double cntrX, double cntrY) {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D pt = ucrd.transform(new Point2D.Double(cntrX, cntrY), null);
        m_node.setCoordinates(pt.getX(), pt.getY());

        // Notification
        _notifyWasChanged();
    }

    public Node getNode() {
        return m_node;
    }

    private AttributedShape getNodeShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

        // Circle shape
        Point2D ctr = getCenter();
        double delta = NODE_RADIUS * scale;
        double delta2 = delta + delta;
        Ellipse2D circle = new Ellipse2D.Double(ctr.getX() - delta, ctr.getY()
                - delta, delta2, delta2);
        AttributedShape as = _attributeShape(new AttributedShape(circle));
        BasicStroke stroke = (BasicStroke) as.getStroke();
        as.setStroke(new BasicStroke(
        /* stroke.getLineWidth() */(float) (NODE_SHAPE_LINE_WIDTH * scale),
                stroke.getEndCap(), stroke.getLineJoin(), stroke
                        .getMiterLimit(), stroke.getDashArray(), stroke
                        .getDashPhase()));
        as.setFillPaint(Color.BLACK);
        return as;
    }

    private AttributedShape getNodeSupportShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        
        Path2D path = new Path2D.Double();
        // Center point
        Point2D ctr = m_node.getPoint();//getCenter();

        // Supported node
        double[] cons = m_node.getContraints();
        if (! Double.isNaN(cons[0])
                || ! Double.isNaN(cons[1]) 
                || ! Double.isNaN(cons[2])) {
            
            double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

            double delta = NODE_RADIUS * scale;
            double delta2 = delta + delta;
            double deltaInner = NODE_RADIUS_INNER * scale;
            double deltaInner2 = deltaInner + deltaInner;

            if (! Double.isNaN(cons[0])) {
                if (cons[0] == 0.0) {
                    Ellipse2D circle = new Ellipse2D.Double(
                            ctr.getX() - delta, 
                            ctr.getY() - delta, 
                            delta2, 
                            delta2);
                    path.append(circle, false);
                }
                else {
                    Arc2D arc = new Arc2D.Double(
                            ctr.getX() - deltaInner,
                            ctr.getY() - deltaInner, 
                            deltaInner2, 
                            deltaInner2, 
                            45.0, 
                            -180.0, 
                            Arc2D.OPEN);
                    path.append(arc, false);
                }
                
            }
            else {
                Ellipse2D circle = new Ellipse2D.Double(
                        ctr.getX() - deltaInner, 
                        ctr.getY() - deltaInner,
                        deltaInner2, 
                        deltaInner2);
                path.append(circle, false);
            }
            if (! Double.isNaN(cons[1])) { // fixed y |
                if (cons[1] != 0) { // disp y
                        path.append(new Line2D.Double(
                            ctr.getX(), 
                            ctr.getY() - deltaInner2, 
                            ctr.getX(), 
                            ctr.getY() - deltaInner2 - deltaInner2), 
                            false);
                }
                else {
                    if (Double.isNaN(cons[0]) || (cons[0] != 0)) { // phi free o
                        path.append(new Line2D.Double(
                            ctr.getX(), 
                            ctr.getY() - deltaInner, 
                            ctr.getX(), 
                            ctr.getY() - deltaInner2 - deltaInner2), 
                            false);
                    }
                    else {
                        path.append(new Line2D.Double( // phi fixed
                                ctr.getX(), 
                                ctr.getY() - delta, 
                                ctr.getX(), 
                                ctr.getY() - deltaInner2 - deltaInner2), 
                                false);
                    }
                }
                
            }
            
            if (! Double.isNaN(cons[2])) { // fixed x -
                if (cons[2] != 0) { // disp x
                        path.append(new Line2D.Double(
                            ctr.getX() - deltaInner2, 
                            ctr.getY(), 
                            ctr.getX() - deltaInner2 - deltaInner2, 
                            ctr.getY()), 
                            false);
                }
                else {
                    if (Double.isNaN(cons[0]) || (cons[0] != 0)) { // phi free o
                        path.append(new Line2D.Double(
                            ctr.getX() - deltaInner, 
                            ctr.getY(), 
                            ctr.getX() - deltaInner2 - deltaInner2, 
                            ctr.getY()), 
                            false);
                    }
                    else { // fixed phi
                        path.append(new Line2D.Double(
                                ctr.getX() - delta, 
                                ctr.getY(), 
                                ctr.getX() - deltaInner2 - deltaInner2, 
                                ctr.getY()), 
                                false);
                    }
                }
            }
            
            
            Shape shape = path;
            try {
                shape = ucrd.createInverse().createTransformedShape(path);
            } catch (NoninvertibleTransformException e) {
                // Should never happen
            }
            
            AttributedShape as = _attributeShape(new AttributedShape(shape));
            BasicStroke stroke = (BasicStroke) as.getStroke();
            as.setStroke(new BasicStroke(
                    /* stroke.getLineWidth() */(float) (NODE_SUPPORT_LINE_WIDTH * scale),
                    stroke.getEndCap(), 
                    stroke.getLineJoin(), 
                    stroke.getMiterLimit(), 
                    stroke.getDashArray(),
                    stroke.getDashPhase()));
            as.setDrawPaint(SUPPORT_COLOR);
            if (cons[0] == 0.0)
                as.setFillPaint(SUPPORT_COLOR);
            
            return as;
        }
        else {
            path.append(new Line2D.Double(ctr, ctr), false);
            Shape shape = path;
            try {
                shape = ucrd.createInverse().createTransformedShape(path);
            } catch (NoninvertibleTransformException e) {
                // Should never happen
            }
            AttributedShape as = _attributeShape(new AttributedShape(shape));
            BasicStroke stroke = (BasicStroke) as.getStroke();
            as.setStroke(new BasicStroke(0, stroke.getEndCap(), stroke
                    .getLineJoin(), stroke.getMiterLimit(), stroke
                    .getDashArray(), stroke.getDashPhase()));
            return as;
        }
    }

    private AttributedShape getNodeConnectionShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();
        Path2D path = new Path2D.Double();
        Point2D ctr = getCenter();

        Structure struct = RuBeam.getStructure();
        Set<Element> beams = struct.getElements(m_node);
        boolean hinged = false;
        for (Element b : beams) {
            if (b.getNodeI() == m_node) {
                if (b.getHingeI()) {
                    hinged = true;
                    break;
                }
            }
            else {
                if (b.getHingeK()) {
                    hinged = true;
                    break;
                }
            }
        }
        if (!hinged) {
            path.append(new Line2D.Double(ctr, ctr), false);
            Shape shape = path;
            AttributedShape as = _attributeShape(new AttributedShape(shape));
            BasicStroke stroke = (BasicStroke) as.getStroke();
            as.setStroke(new BasicStroke(0, stroke.getEndCap(), stroke
                    .getLineJoin(), stroke.getMiterLimit(), stroke
                    .getDashArray(), stroke.getDashPhase()));
            return as;
        }

        double deltaInner = NODE_RADIUS_INNER * scale;
        double deltaInner2 = deltaInner + deltaInner;

        // Circles
        Ellipse2D circleInner = new Ellipse2D.Double(ctr.getX() - deltaInner,
                ctr.getY() - deltaInner, deltaInner2, deltaInner2);
        path.append(circleInner, false);

        AttributedShape as = _attributeShape(new AttributedShape(path));
        BasicStroke stroke = (BasicStroke) as.getStroke();
        as.setStroke(new BasicStroke(
        /* stroke.getLineWidth() */(float) (NODE_CONN_LINE_WIDTH * scale),
                stroke.getEndCap(), stroke.getLineJoin(), stroke
                        .getMiterLimit(), stroke.getDashArray(), stroke
                        .getDashPhase()));
        // as.setFillPaint(Color.BLACK);
        return as;
    }

    public NamedListIterator<AttributedShape> shapeIterator() {
        return new NamedListIteratorAdapter<AttributedShape>() {
            protected int _size() {
                return 3;
            }

            protected AttributedShape _get(int index) {
                switch (index) {
                case 0:
                    return getNodeShape();
                case 1:
                    return getNodeConnectionShape();
                case 2:
                    return getNodeSupportShape();
                default:
                    throw new IllegalArgumentException();
                }
            }
        };
    }

    private AttributedText getNameText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale()
                / (cs.getMillimetersPerNaturalUnit() * cs
                        .getNaturalUnitsPerWorldUnit());
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        Point2D ptW = u2w.transform(getCenter(), null);

        AffineTransform trf = AffineTransform.getTranslateInstance(
                ptW.getX() + NODE_TEXT_OFFSET * scale, 
                ptW.getY() + NODE_TEXT_OFFSET * scale);

        AttributedString string = new AttributedString(m_node.getName());
        AttributedText text = new AttributedText(string, trf);
        text.setJustification(AttributedText.JUST_LEFT_BOTTOM);
        text.transformBy(w2u);
        
        text = _attributeText(text);
        AttributedString as = text.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(NODE_TEXT_SIZE *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        
        return text;
    }

    // [u, w, phi, M, V, H]
    private AttributedText getDxText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale()
        / (cs.getMillimetersPerNaturalUnit() * cs
                .getNaturalUnitsPerWorldUnit());
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        
        Structure struct = RuBeam.getStructure();
        
        Point2D ctrU = getCenter();
        Point2D ctrW = u2w.transform(ctrU, null);
        Point2D pW = ctrW;
        double dx = 0.0;
        double dy = 0.0;
        if (struct.isValid()) {
            double max = struct.getMaxResultVg() * 1000;
            if (max == 0.0)
                max = 1;
            // [mm]
            dx = - struct.getDeltaX(getNode()) * 1e3;
            dy = struct.getDeltaY(getNode()) * 1e3;
            if (! Double.isNaN(dx) && ! Double.isNaN(dy)) {
                double x = ctrW.getX()
                        + (dx * scale / max * NODE_MAX_DISP);
                double y = ctrW.getY()
                        + (dy * scale / max * NODE_MAX_DISP);
                pW.setLocation(x, y);
            }
        }
        
        String textX = String.format(Locale.ENGLISH, "%.2f", dx)+ "e-3";
        AffineTransform trfX = AffineTransform.getTranslateInstance(
                pW.getX() + (NODE_RESULT_OFFSET * scale), pW.getY());
        
        AttributedText aTextX = new AttributedText(
            new AttributedString(textX), trfX);
        aTextX.setJustification(AttributedText.JUST_LEFT_CENTER);
        aTextX.transformBy(w2u);
        
        aTextX = _attributeText(aTextX);
        aTextX.setDrawPaint(ComponentElement.RESULT_COLOR);
        
        AttributedString as = aTextX.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(.5 * NODE_TEXT_SIZE *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        return aTextX;
    }
    private AttributedText getDyText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale()
        / (cs.getMillimetersPerNaturalUnit() * cs
                .getNaturalUnitsPerWorldUnit());
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        
        Structure struct = RuBeam.getStructure();
        
        Point2D ctrU = getCenter();
        Point2D ctrW = u2w.transform(ctrU, null);
        Point2D pW = ctrW;
        double dx = 0.0;
        double dy = 0.0;
        if (struct.isValid()) {
            // [mm]
            double max = struct.getMaxResultVg() * 1000;
            if (max == 0.0)
                max = 1.0;
            dx = - struct.getDeltaX(getNode()) * 1e3;
            dy = struct.getDeltaY(getNode()) * 1e3;
            if (! Double.isNaN(dx) && ! Double.isNaN(dy)) {
                double x = ctrW.getX()
                        + (dx * scale / max * NODE_MAX_DISP);
                double y = ctrW.getY()
                        + (dy * scale / max * NODE_MAX_DISP);
                pW.setLocation(x, y);
            }
        }
        
        String textY = String.format(Locale.ENGLISH, "%.2f", dy)+ "e-3";
        
        AffineTransform trfY = AffineTransform.getTranslateInstance(
                pW.getX(), pW.getY());
        trfY.rotate(.5 * Math.PI);
        trfY.translate(NODE_RESULT_OFFSET * scale, 0);
        
        AttributedText aTextY = new AttributedText(
            new AttributedString(textY), trfY);
        aTextY.setJustification(AttributedText.JUST_LEFT_CENTER);
        aTextY.transformBy(w2u);
        
        aTextY = _attributeText(aTextY);
        aTextY.setDrawPaint(ComponentElement.RESULT_COLOR);
        
        AttributedString as = aTextY.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(.5 * NODE_TEXT_SIZE *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        return aTextY;
    }
    private AttributedText getHText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale()
        / (cs.getMillimetersPerNaturalUnit() * cs
                .getNaturalUnitsPerWorldUnit());
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        
        Structure struct = RuBeam.getStructure();
        
        Point2D ctrU = getCenter();
        Point2D ctrW = u2w.transform(ctrU, null);
        Point2D pW = ctrW;
        
        double H = struct.getNodeResult(m_node).get(5, 0);
        
        String textH = String.format(Locale.ENGLISH, "%.2f", H);//+ "e3";
        AffineTransform trfX = AffineTransform.getTranslateInstance(
                pW.getX() - (2 * NODE_RESULT_OFFSET * scale), pW.getY());
        
        AttributedText aTextH = new AttributedText(
            new AttributedString(textH), trfX);
        aTextH.setJustification(AttributedText.JUST_RIGHT_TOP);
        aTextH.transformBy(w2u);
        
        aTextH = _attributeText(aTextH);
        aTextH.setDrawPaint(SUPPORT_COLOR);
        
        AttributedString as = aTextH.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(.5 * NODE_TEXT_SIZE *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        return aTextH;
    }
    private AttributedText getVText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale()
        / (cs.getMillimetersPerNaturalUnit() * cs
                .getNaturalUnitsPerWorldUnit());
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        
        Structure struct = RuBeam.getStructure();
        
        Point2D ctrU = getCenter();
        Point2D ctrW = u2w.transform(ctrU, null);
        Point2D pW = ctrW;
        
        double V = struct.getNodeResult(m_node).get(4, 0);
        
        String textV = String.format(Locale.ENGLISH, "%.2f", V);//+ "e3";
        
        AffineTransform trfY = AffineTransform.getTranslateInstance(
                pW.getX(), pW.getY());
        trfY.rotate(.5 * Math.PI);
        trfY.translate(- (2 * NODE_RESULT_OFFSET * scale), 0);
        
        AttributedText aTextV = new AttributedText(
            new AttributedString(textV), trfY);
        aTextV.setJustification(AttributedText.JUST_RIGHT_BOTTOM);
        aTextV.transformBy(w2u);
        
        aTextV = _attributeText(aTextV);
        aTextV.setDrawPaint(SUPPORT_COLOR);
        
        AttributedString as = aTextV.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(.5 * NODE_TEXT_SIZE *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        return aTextV;
    }
    private AttributedText getMText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale()
        / (cs.getMillimetersPerNaturalUnit() * cs
                .getNaturalUnitsPerWorldUnit());
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        
        Structure struct = RuBeam.getStructure();
        
        Point2D ctrU = getCenter();
        Point2D ctrW = u2w.transform(ctrU, null);
        Point2D pW = ctrW;
        
        double M = struct.getNodeResult(m_node).get(3, 0);
        
        String textM = String.format(Locale.ENGLISH, "%.2f", M);//+ "e3";
        
        AffineTransform trfY = AffineTransform.getTranslateInstance(
                pW.getX() - (2 * NODE_RESULT_OFFSET * scale), 
                pW.getY() - (2 * NODE_RESULT_OFFSET * scale));
        trfY.rotate(.25 * Math.PI);
        trfY.translate(- NODE_RESULT_OFFSET * scale, 0);
        
        AttributedText aTextM = new AttributedText(
            new AttributedString(textM), trfY);
        aTextM.setJustification(AttributedText.JUST_RIGHT_CENTER);
        aTextM.transformBy(w2u);
        
        aTextM = _attributeText(aTextM);
        aTextM.setDrawPaint(SUPPORT_COLOR);
        
        AttributedString as = aTextM.getAttributedString();
        AttributedCharacterIterator aci = as.getIterator();
        if (aci.getRunLimit() != 0) {
            final double PT_PER_IN = 72.;
            final double IN_PER_M = 39.37;
            final double M_TO_PT = PT_PER_IN * IN_PER_M;
            as.addAttribute(TextAttribute.SIZE, 
                    new Float(.5 * NODE_TEXT_SIZE *  M_TO_PT * cs.getScale() /
                            cs.getMillimetersPerNaturalUnit()));
        }
        return aTextM;
    }
    
    public NamedListIterator<AttributedText> textIterator() {
        /*
         * Names
         * 0 : getNameText()
         * 1 : getDxText()
         * 2 : getDyText()
         * 3 : getHText()
         * 4 : getVText()
         * 5 : getMText()
         * */
        return new NamedListIteratorAdapter<AttributedText>() {
            private List<Integer> names = new ArrayList<Integer>();
            private void _updateNames() {
                names.clear();
                names.add(0); // name
                double[] cons = getNode().getContraints();
                
                if (RuBeam.getStructure().isValid()) {
                    if (Double.isNaN(cons[2]) || cons[2] != 0.0) {
                        names.add(1); // Dx
                    }
                    if (Double.isNaN(cons[1]) || cons[1] != 0.0) {
                        names.add(2); // Dy
                    }
                    if (! Double.isNaN(cons[2])) {
                        names.add(3); // H
                    }
                    if (! Double.isNaN(cons[1])) {
                        names.add(4); // V
                    }
                    if (! Double.isNaN(cons[0])) {
                        names.add(5); // M
                    }
                }
            }
            private AttributedText _getText(int name) {
                switch (name) {
                case 0:
                    return getNameText();
                case 1:
                    return getDxText();
                case 2:
                    return getDyText();
                case 3:
                    return getHText();
                case 4:
                    return getVText();
                case 5:
                    return getMText();
                default :
                    throw new IllegalArgumentException();
                }
            }
            protected int _size() {
                _updateNames();
                return names.size();
            }
            protected AttributedText _get(int index) {
                _updateNames();
                switch (index) {
                case 0:
                    return getNameText();
                case 1:
                    return _getText(names.get(1));
                case 2:
                    return _getText(names.get(2));
                case 3:
                    return _getText(names.get(3));
                case 4:
                    return _getText(names.get(4));
                case 5:
                    return _getText(names.get(5));
                default:
                    throw new IllegalArgumentException();
                }
            }
            protected int _name(int index) {
                _updateNames();
                return names.get(index);
            }
        };
    }

    public Point2D getControlPoint(int name) throws IllegalArgumentException {
        switch (name) {
        case 0:
            return getCenter();
        case 1:
            final CoordSpace cs = CoordSpace.getCoordSpace();
            AffineTransform ucrd = cs.getUserCrdTransform();
            double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

            Structure struct = RuBeam.getStructure();
            Point2D p = getCenter();

            if (struct.isValid()) {
                double max = struct.getMaxResultVg();
                if (max == 0.0)
                    max = 1.0;
                //System.out.println("max=" + max + ", scale=" + scale);
                double dx = - struct.getDeltaX(getNode());// * 1e3;
                double dy = struct.getDeltaY(getNode());// * 1e3;
                if (! Double.isNaN(dx) && ! Double.isNaN(dy)) {
                    double x = getNode().getPoint().getX() 
                        + (dx * scale / max * NODE_MAX_DISP);  
                    double y = getNode().getPoint().getY() 
                        + (dy * scale  / max * NODE_MAX_DISP);
                    p.setLocation(x, y);
                    try {
                        p = ucrd.inverseTransform(p, null);
                    } catch (NoninvertibleTransformException e) {
                        // Should never happen
                    }
                }
            }
            return p;
        default:
            throw new IllegalArgumentException();
        }
    }

    public void setControlPoint(Point2D pnt, int name)
            throws UnsupportedOperationException {
        Point2D p = pnt;
        switch (name) {
        case 0:
            setNode(p.getX(), p.getY());
            break;
        case 1:
            break;
        default:
            throw new IllegalArgumentException();
        }
    }
    
    public NamedListIterator<Point2D> controlPointIterator() {
        return new NamedListIteratorAdapter<Point2D>() {
            protected int _size() {
                if (RuBeam.getStructure().isValid())
                    return 2;
                else
                    return 1;
            }

            protected Point2D _get(int index) {
                switch (index) {
                case 0:
                    return getCenter();
                case 1:
                    return getControlPoint(1);
                default:
                    throw new IllegalArgumentException();
                }
            }

            protected void _set(int index, Point2D value) {
                Point2D p = (Point2D) value;
                switch (index) {
                case 0:
                    setNode(p.getX(), p.getY());
                    break;
                case 1:
                    break;
                default:
                    throw new IllegalArgumentException();
                }
            }
        };
    }

    public void transformBy(AffineTransform mat) {
        // Transform point
        Point2D c = getCenter();
        Point2D newC = new Point2D.Double();
        mat.transform(c, newC);

        setNode(newC.getX(), newC.getY());
        _notifyWasChanged();
    }

    // Node interface
    public void setConstraints(double fphi, double fy, double fx) {
        m_node.setContraints(fphi, fy, fx);
        _notifyWasChanged();
    }
    
    // Features
    private static final String[] FEATURE_NAMES = { 
            "POINT", // 0
            "NAME", // 1
            "CONST_X", // 2
            "CONST_Y", // 3
            "CONST_PHI", // 4
            "DX [e-3]", // 5
            "DY [e-3]", // 6
            "DPHI [rad]" // 7
    };

    public boolean hasFeature(String name) {
        for (int i = 0; i < FEATURE_NAMES.length; i++) {
            if (name.equals(FEATURE_NAMES[i]))
                return true;
        }
        return super.hasFeature(name);
    }

    public Feature getFeature(String name) {
        Structure struct = RuBeam.getStructure();
        if (name.equals(FEATURE_NAMES[0]))
            return new Point2DFeature(FEATURE_NAMES[0], getCenter());
        else if (name.equals(FEATURE_NAMES[1])) {
            StringFeature sf = new StringFeature(FEATURE_NAMES[1],
                    (String) getNode().getName());
            return sf;
        }
        else if (name.equals(FEATURE_NAMES[2])) {
            String dx = Double.toString(m_node.getContraints()[2]);
            if (dx.startsWith("-"))
                dx = dx.substring(1);
            return new StringFeature(FEATURE_NAMES[2], dx);
        }
        else if (name.equals(FEATURE_NAMES[3])) {
            return new StringFeature(FEATURE_NAMES[3],
                    Double.toString(m_node.getContraints()[1]));
        }
        else if (name.equals(FEATURE_NAMES[4])) {
            return new StringFeature(FEATURE_NAMES[4],
                    Double.toString(m_node.getContraints()[0]));
        }
        else if (name.equals(FEATURE_NAMES[5])) {
            double d;
            if (struct.isValid()) {
                d = struct.getDeltaX(m_node) * 1e3;
            }
            else
                d = Double.NaN;
            DoubleFeature df = new DoubleFeature(FEATURE_NAMES[5], d);
            df.setChangeable(false);
            return df;
        }
        else if (name.equals(FEATURE_NAMES[6])) {
            double d;
            if (struct.isValid()) {
                d = struct.getDeltaY(m_node) * 1e3;
            }
            else
                d = Double.NaN;
            DoubleFeature df = new DoubleFeature(FEATURE_NAMES[6], d);
            df.setChangeable(false);
            return df;
        }
        else if (name.equals(FEATURE_NAMES[7])) {
            double d;
            if (struct.isValid()) {
                d = struct.getDeltaPhi(m_node);
            }
            else
                d = Double.NaN;
            DoubleFeature df = new DoubleFeature(FEATURE_NAMES[7], d);
            df.setChangeable(false);
            return df;
        }
        else
            return super.getFeature(name);
    }

    public void setFeature(Feature feature) {
        // Preset arc parameters
        Point2D center = getCenter();

        // Get parameter from feature
        if (feature.getName().equals(FEATURE_NAMES[0])) {
            center = (Point2D) feature.getValue();
            setNode(center.getX(), center.getY());
            return;
        }
        else if (feature.getName().equals(FEATURE_NAMES[1])) {
            m_node.setName(feature.valueToString());
        }
        else if (feature.getName().equals(FEATURE_NAMES[2])) {
            String value = feature.valueToString();
            if (value.equals(""))
                m_node.setContraintX(Double.NaN);
            else {
                double dispX = Double.NaN;
                try {
                    dispX =  - Double.parseDouble(value);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                m_node.setContraintX(dispX);
            }
        }
        else if (feature.getName().equals(FEATURE_NAMES[3])) {
            String value = feature.valueToString();
            if (value.equals(""))
                m_node.setContraintY(Double.NaN);
            else {
                double dispY = Double.NaN;
                try {
                    dispY = Double.parseDouble(value);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                m_node.setContraintY(dispY);
            }
        }
        else if (feature.getName().equals(FEATURE_NAMES[4])) {
            String value = feature.valueToString();
            if (value.equals(""))
                m_node.setContraintPhi(Double.NaN);
            else {
                double dispP = Double.NaN;
                try {
                    dispP = Double.parseDouble(value);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                m_node.setContraintPhi(dispP);
            }
        }
        else {
            super.setFeature(feature);
            return;
        }
        _notifyWasChanged();
    }

    public Iterator<Feature> featureIterator() {

        final Iterator<Feature> itThis = new Iterator<Feature>() {
            private int m_index = 0;

            public boolean hasNext() {
                return m_index < FEATURE_NAMES.length;
            }

            public Feature next() {
                if (hasNext())
                    return getFeature(FEATURE_NAMES[m_index++]);
                throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };

        final Iterator<Feature> itBase = super.featureIterator();

        return new Iterator<Feature>() {
            public boolean hasNext() {
                return itBase.hasNext() | itThis.hasNext();
            }

            public Feature next() {
                if (itBase.hasNext())
                    return itBase.next();
                else if (itThis.hasNext())
                    return itThis.next();
                else
                    throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

}
