/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.gui;

import java.awt.Color;
import java.util.List;

import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

import rubeam.RuBeam;
import rubeam.model.Element;
import cib.cad.kernel.InputDeviceAdapter;
import cib.cad.kernel.TextToken;

public class ElementTablePanel extends JScrollPane {

    private static final long serialVersionUID = 1L;
    private static final String[] colNames = new String[] {
        "Name", "Node i", "Node k", "Section", "Hinge i", "Hinge k"
    };
    
    // GUI
    private JTable table = new JTable();
    private TableModel model = new TableModel();
    
    // Input device
    private InputDeviceAdapter inputDevice = new InputDeviceAdapter();
    
    public ElementTablePanel(InputDeviceAdapter inputDevice) {
        super();
        this.inputDevice = inputDevice;
        initialize();
    }
    
    private void initialize() {
        this.setViewportView(table);
        this.setVerticalScrollBarPolicy(VERTICAL_SCROLLBAR_AS_NEEDED);
        this.setHorizontalScrollBarPolicy(HORIZONTAL_SCROLLBAR_AS_NEEDED);
        table.setModel(model);
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        table.setGridColor(Color.LIGHT_GRAY);
        setCellEditors();
    }
    
    public void updatePanel() {
        model.fireTableDataChanged();
    }
    
    public Element getSelectedElement() {
        int row = table.getSelectedRow();
        if (row != -1)
            return RuBeam.getStructure().readElements().get(row);
        else 
            return null;
    }
    
    private void setCellEditors() {
        TableColumn tc = table.getColumnModel().getColumn(1);
        tc.setCellEditor(new NodeListCellEditor());
        tc = table.getColumnModel().getColumn(2);
        tc.setCellEditor(new NodeListCellEditor());
        tc = table.getColumnModel().getColumn(3);
        tc.setCellEditor(new SectionListCellEditor());
    }
    
    class TableModel extends DefaultTableModel {

        private static final long serialVersionUID = 1L;
        public int getColumnCount() {
            return colNames.length;
        }

        public String getColumnName(int column) {
            return colNames[column];
        }
        
        public int getRowCount() {
            return RuBeam.getStructure().readElements().size();
        }
        
        public Class<?> getColumnClass(int columnIndex) {
            switch (columnIndex) {
            case 0:
                return String.class;
            case 1:
                return String.class;
            case 2:
                return String.class;
            case 3:
                return String.class;
            case 4:
                return Boolean.class;
            case 5: 
                return Boolean.class;
            default:
                throw new IllegalStateException();
            }
        }
        
        public boolean isCellEditable(int rowIndex, int columnIndex) {
            return true;
        }
        
        public Object getValueAt(int rowIndex, int columnIndex) {
            List<Element> elementList = RuBeam.getStructure().readElements();
            if (rowIndex >= elementList.size())
                return null;
            Element element = elementList.get(rowIndex);
            switch (columnIndex) {
            case 0:
                return element.getName();
            case 1:
                return element.getNodeI().getName();
            case 2:
                return element.getNodeK().getName();
            case 3:
                return element.getSection().getName();
            case 4:
                return new Boolean(element.getHingeI());
            case 5:
                return new Boolean(element.getHingeK());
            default:
                throw new IllegalStateException();
            }
        }
        
        public void setValueAt(Object value, int rowIndex, int columnIndex) {
            if (columnIndex == 0) { // set name
                String newName = value.toString();
                List<Element> elmList = RuBeam.getStructure().readElements();
                String oldName = elmList.get(rowIndex).getName();
                String fire = "setelementname " 
                    + TextToken.stringToLiteral(oldName) + " " 
                    + TextToken.stringToLiteral(newName);
                fireInput(fire);
                model.fireTableCellUpdated(rowIndex, columnIndex);
            }
            else if (columnIndex == 1) { // set node i
                try {
                    String newNodeI = value.toString();
                    List<Element> elmList = RuBeam.getStructure().readElements();
                    String elmName = elmList.get(rowIndex).getName();
                    String fire = "setelementnodei " 
                        + TextToken.stringToLiteral(elmName) + " " 
                        + TextToken.stringToLiteral(newNodeI);
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
            }
            else if (columnIndex == 2) { // set node k
                try {
                    String newNodeK = value.toString();
                    List<Element> elmList = RuBeam.getStructure().readElements();
                    String elmName = elmList.get(rowIndex).getName();
                    String fire = "setelementnodek " 
                        + TextToken.stringToLiteral(elmName) + " " 
                        + TextToken.stringToLiteral(newNodeK);
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
            }
            else if (columnIndex == 3) { // set section
                try {
                    String newSection = value.toString();
                    List<Element> elmList = RuBeam.getStructure().readElements();
                    String elmName = elmList.get(rowIndex).getName();
                    String fire = "assignsection " 
                        + TextToken.stringToLiteral(newSection) + " " 
                        + TextToken.stringToLiteral(elmName);
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
            }
            else if (columnIndex == 4) { // set joint i
                try {
                    Boolean jointedI = Boolean.parseBoolean(value.toString());
                    List<Element> elmList = RuBeam.getStructure().readElements();
                    String elmName = elmList.get(rowIndex).getName();
                    String fire = "setelementhingei " 
                        + TextToken.stringToLiteral(elmName) + " " + jointedI;
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
            }
            else if (columnIndex == 5) { // set joint k
                try {
                    Boolean jointK = Boolean.parseBoolean(value.toString());
                    List<Element> elmList = RuBeam.getStructure().readElements();
                    String elmName = elmList.get(rowIndex).getName();
                    String fire = "setelementhingek " 
                        + TextToken.stringToLiteral(elmName) + " " + jointK;
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
            }
            else 
                throw new IllegalStateException();
        }
    }

    private void fireInput(String input) {
        if (inputDevice.isEnabled())
            inputDevice.fireInterruptingInput(input);
    }
}
