/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.gui;

import java.awt.Color;
import java.util.List;

import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

import rubeam.RuBeam;
import rubeam.model.Node;
import cib.cad.kernel.InputDeviceAdapter;
import cib.cad.kernel.TextToken;

public class NodeTablePanel extends JScrollPane {

    private static final long serialVersionUID = 1L;
    private static final NumberCellRenderer numberRenderer = new NumberCellRenderer();
    private static final String[] colNames = new String[] { "Name", "x", "y",
            "dx", "dy", "d\u03c6" };
    // GUI
    private JTable table = new JTable();
    private TableModel model = new TableModel();

    // Input device
    private InputDeviceAdapter inputDevice = new InputDeviceAdapter();

    public NodeTablePanel(InputDeviceAdapter inputDevice) {
        super();
        this.inputDevice = inputDevice;
        initialize();
    }

    private void initialize() {
        this.setViewportView(table);
        this.setVerticalScrollBarPolicy(VERTICAL_SCROLLBAR_AS_NEEDED);
        this.setHorizontalScrollBarPolicy(HORIZONTAL_SCROLLBAR_AS_NEEDED);
        table.setModel(model);
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        table.setGridColor(Color.LIGHT_GRAY);
        TableColumn tc = table.getColumnModel().getColumn(1);
        tc.setCellRenderer(numberRenderer);
        tc = table.getColumnModel().getColumn(2);
        tc.setCellRenderer(numberRenderer);
        tc = table.getColumnModel().getColumn(3);
        tc.setCellRenderer(numberRenderer);
        tc = table.getColumnModel().getColumn(4);
        tc.setCellRenderer(numberRenderer);
        tc = table.getColumnModel().getColumn(5);
        tc.setCellRenderer(numberRenderer);
    }

    public void updatePanel() {
        model.fireTableDataChanged();
    }

    public Node getSelectedNode() {
        int row = table.getSelectedRow();
        if (row != -1)
            return RuBeam.getStructure().readNodes().get(row);
        else
            return null;
    }

    class TableModel extends DefaultTableModel {

        private static final long serialVersionUID = 1L;

        public int getColumnCount() {
            return colNames.length;
        }

        public String getColumnName(int column) {
            return colNames[column];
        }

        public int getRowCount() {
            return RuBeam.getStructure().readNodes().size();
        }

        public Class<?> getColumnClass(int columnIndex) {
            switch (columnIndex) {
            case 0:
                return String.class;
            case 1:
                return Double.class;
            case 2:
                return Double.class;
            case 3:
                return Double.class;
            case 4:
                return Double.class;
            case 5:
                return Double.class;
            default:
                throw new IllegalStateException();
            }
        }

        public boolean isCellEditable(int row, int column) {
            return true;
        }

        public Object getValueAt(int rowIndex, int columnIndex) {
            List<Node> nodeList = RuBeam.getStructure().readNodes();
            if (rowIndex >= nodeList.size())
                return null;
            Node node = nodeList.get(rowIndex);
            switch (columnIndex) {
            case 0:
                return node.getName();
            case 1:
                return Double.valueOf(node.getPoint().getX());
            case 2:
                return Double.valueOf(node.getPoint().getY());
            case 3:
                double dx = node.getContraints()[2];
                return Double.valueOf((Math.abs(dx) != 0.0) ? -dx : Math
                        .abs(dx));
            case 4:
                return Double.valueOf(node.getContraints()[1]);
            case 5:
                return Double.valueOf(node.getContraints()[0]);
            default:
                throw new IllegalStateException();
            }
        }

        public void setValueAt(Object value, int rowIndex, int columnIndex) {
            switch (columnIndex) {
            case 0: // set name
                String newName = value.toString();
                List<Node> nodeList = RuBeam.getStructure().readNodes();
                String oldName = nodeList.get(rowIndex).getName();
                String fire = "setnodename " 
                    + TextToken.stringToLiteral(oldName) + " " 
                    + TextToken.stringToLiteral(newName);
                fireInput(fire);
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 1: // set x
                try {
                    double newX = Double.parseDouble(value.toString());
                    nodeList = RuBeam.getStructure().readNodes();
                    String nodeName = nodeList.get(rowIndex).getName();
                    fire = "setnodex " 
                        + TextToken.stringToLiteral(nodeName) + " " + newX;
                    fireInput(fire);
                } catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 2: // set y
                try {
                    double newY = Double.parseDouble(value.toString());
                    nodeList = RuBeam.getStructure().readNodes();
                    String nodeName = nodeList.get(rowIndex).getName();
                    fire = "setnodey " 
                        + TextToken.stringToLiteral(nodeName) + " " + newY;
                    fireInput(fire);
                } catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 3: // set dx
                String textDx;
                try {
                    Double.parseDouble(value.toString());
                    textDx = value.toString();
                } catch (Exception e) {
                    textDx = "-";
                }
                nodeList = RuBeam.getStructure().readNodes();
                String nodeName = nodeList.get(rowIndex).getName();
                fire = "setnodeconstx " 
                    + TextToken.stringToLiteral(nodeName) + " " + textDx;
                fireInput(fire);
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 4: // set dy
                String textDy;
                try {
                    Double.parseDouble(value.toString());
                    textDy = value.toString();
                } catch (Exception e) {
                    textDy = "-";
                }
                nodeList = RuBeam.getStructure().readNodes();
                nodeName = nodeList.get(rowIndex).getName();
                fire = "setnodeconsty " 
                    + TextToken.stringToLiteral(nodeName) + " " + textDy;
                fireInput(fire);
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 5: // set dPhi
                String textDphi;
                try {
                    Double.parseDouble(value.toString());
                    textDphi = value.toString();
                } catch (Exception e) {
                    textDphi = "-";
                }
                nodeList = RuBeam.getStructure().readNodes();
                nodeName = nodeList.get(rowIndex).getName();
                fire = "setnodeconstphi " 
                    + TextToken.stringToLiteral(nodeName) + " " + textDphi;
                fireInput(fire);
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            default:
                throw new IllegalStateException();
            }
        }
    }

    private void fireInput(String input) {
        if (inputDevice.isEnabled())
            inputDevice.fireInterruptingInput(input);

    }
}
