/* 
 * Copyright (C) 2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.geom.Path2D;
import java.awt.geom.PathIterator;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

/** 
 * Serializable wrapper of an AttributedShape object.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.1 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/03/08 19:01:12 $
 */
public class AttributedShapeSerializable extends AttributedShape implements 
Serializable {

    private static final long serialVersionUID = 1L;
    private static final int END_OF_COMPONENT = -1;

    // Constructors
    public AttributedShapeSerializable(Shape shape) {
        super(shape);
    }
    public AttributedShapeSerializable(AttributedShapeSerializable as) {
        super(as);
    }
    public AttributedShapeSerializable(Shape shape, BasicStroke stroke, 
            Color draw, Color fill) {
        super(shape, stroke, draw, fill);
    }

    // Object implementation
    public Object clone() {
        AttributedShapeSerializable clone = 
                (AttributedShapeSerializable)super.clone();
        return clone;
    }
    
    // Serialization
    private void writeObject(ObjectOutputStream out) throws IOException {

        // Serialize non-transient attributes
        out.defaultWriteObject();

        // Shape object
        int windingRule = m_shape instanceof Path2D ?
                ((Path2D)m_shape).getWindingRule() :
                Path2D.WIND_NON_ZERO;    
        out.writeInt(windingRule);

        double[] coord = new double[6];
        for (PathIterator it = m_shape.getPathIterator(null);
             ! it.isDone(); it.next()
        ) {
            // Get segment data
            int seg = it.currentSegment(coord);

            // Write segment type
            out.writeInt(seg);

            // Write correct number of coordinates
            int nCoord = 0;
            switch (seg) {
                case PathIterator.SEG_MOVETO :
                case PathIterator.SEG_LINETO :
                    nCoord = 2;
                    break;
                case PathIterator.SEG_QUADTO :
                    nCoord = 4;
                    break;
                case PathIterator.SEG_CUBICTO :
                    nCoord = 6;
                    break;
                case PathIterator.SEG_CLOSE :
                    nCoord = 0;
                    break;
                default :
                    throw new Error("Unexpected segment type");
            }
            for (int i = 0; i < nCoord; i++) {
                out.writeDouble(coord[i]);
            }
        }
        out.writeInt(END_OF_COMPONENT);
        
        // Stroke object
        Stroke stroke = m_stroke;
        if (stroke == null || stroke instanceof Serializable) {
            out.writeObject(stroke);
            out.writeBoolean(false);
        }
        else if (stroke instanceof BasicStroke) {
            out.writeObject(null);
            out.writeBoolean(true);     // BascStroke follows
            BasicStroke basicStroke = (BasicStroke)stroke;
            out.writeFloat(basicStroke.getLineWidth());
            out.writeInt(basicStroke.getEndCap()); 
            out.writeInt(basicStroke.getLineJoin()); 
            out.writeFloat(basicStroke.getMiterLimit()); 
            out.writeObject(basicStroke.getDashArray()); 
            out.writeFloat(basicStroke.getDashPhase()); 
        }
        else {
            out.writeObject(null);
            out.writeBoolean(false);
        }

        // Draw paint object
        Paint draw = m_draw;
        if (draw != null && ! (draw instanceof Serializable))
            draw = null;
        out.writeObject(draw);
        
        // Fill paint object
        Paint fill = m_fill;
        if (fill != null && ! (fill instanceof Serializable))
            fill = null;
        out.writeObject(fill);
    }
    private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {

        // Deserialize non-transient attributes
        in.defaultReadObject();

        // Shape object
        int windingRule = in.readInt();
        Path2D path = new Path2D.Double(windingRule);

        m_shape = null;
        while (m_shape == null) {
            int seg = in.readInt();
            switch (seg) {
                case PathIterator.SEG_MOVETO :
                    path.moveTo(in.readDouble(), in.readDouble());
                    break;
                case PathIterator.SEG_LINETO :
                    path.lineTo(in.readDouble(), in.readDouble());
                    break;
                case PathIterator.SEG_QUADTO :
                    path.quadTo(in.readDouble(), in.readDouble(), 
                            in.readDouble(), in.readDouble());
                    break;
                case PathIterator.SEG_CUBICTO :
                    path.curveTo(in.readDouble(), in.readDouble(),
                            in.readDouble(), in.readDouble(),
                            in.readDouble(), in.readDouble());
                    break;
                case PathIterator.SEG_CLOSE :
                    path.closePath();
                    break;
                case END_OF_COMPONENT :
                    m_shape = path;
                    break;
                default :
                    throw new Error("Unexpected segment type");
            }
        }
        
        // Stroke object
        m_stroke = (Stroke)in.readObject();
        if (in.readBoolean())   // BasicStroke follows
            m_stroke = new BasicStroke(
                    in.readFloat(),             // width
                    in.readInt(),               // cap 
                    in.readInt(),               // join 
                    in.readFloat(),             // miter limit
                    (float[])in.readObject(),   // dash array 
                    in.readFloat());            // dash phase 

        // Draw paint object
        m_draw = (Paint)in.readObject();
        
        // Fill paint object
        m_fill = (Paint)in.readObject();
    }

    // Attributes
    public void setShape(Shape shape) {
        if (shape == null)
            throw new IllegalArgumentException();
        super.setShape(shape);
    }
    public void setDrawPaint(Paint paint) {
        if (paint != null && ! (paint instanceof Serializable))
            throw new IllegalArgumentException();
        super.setDrawPaint(paint);
    }
    public void setFillPaint(Paint paint) {
        if (paint != null && ! (paint instanceof Serializable))
            throw new IllegalArgumentException();
        super.setFillPaint(paint);
    }
    public void setStroke(Stroke stroke) {
        if (stroke != null && ! (stroke instanceof Serializable) && 
                ! (stroke instanceof BasicStroke))
            throw new IllegalArgumentException();
        super.setStroke(stroke);
    }
}