/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.awt.Paint;
import java.awt.geom.AffineTransform;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.text.CharacterIterator;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

/** 
 * A placeable and serializable attributed string.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.11 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class AttributedText implements Cloneable, Serializable {

    private static final long serialVersionUID = 0L;

    // Constants
    public static final int JUST_LEFT_BASELINE = 0;
    public static final int JUST_LEFT_BOTTOM = 9;
    public static final int JUST_LEFT_CENTER = 1;
    public static final int JUST_LEFT_TOP = 2;
    public static final int JUST_CENTER_BASELINE = 3;
    public static final int JUST_CENTER_BOTTOM = 10;
    public static final int JUST_CENTER_CENTER = 4;
    public static final int JUST_CENTER_TOP = 5;
    public static final int JUST_RIGHT_BASELINE = 6;
    public static final int JUST_RIGHT_BOTTOM = 11;
    public static final int JUST_RIGHT_CENTER = 7;
    public static final int JUST_RIGHT_TOP = 8;
    
    // Attributes
    private AffineTransform m_trf;
    private int m_just = JUST_LEFT_BASELINE;
    private transient AttributedString m_as;
    private static final int END_OF_ATTRIBUTED_STRING = -1;
    private Paint m_draw = null;
    private Paint m_fill = null;
    
    // Constructors
    public AttributedText(AttributedString as, AffineTransform trf) {
        if (as == null || trf == null)
            throw new IllegalArgumentException();
        m_as = as;
        m_trf = trf;
    }
    public Object clone() {
        AttributedText cp = new AttributedText(new AttributedString(
            m_as.getIterator()), new AffineTransform(m_trf));
        cp.m_just = m_just;
        cp.m_draw = m_draw;
        cp.m_fill = m_fill;
        return cp;
    }

    // Serialization ....
    private void writeObject(ObjectOutputStream out) throws IOException {

        // Serialize non-transient attributes
        out.defaultWriteObject();

        // Write state of transient AttributedString ...
        AttributedCharacterIterator acit = m_as.getIterator();

        // ... text ...
        StringBuffer buf = new StringBuffer();
        for(char c = acit.first(); c != CharacterIterator.DONE; 
                c = acit.next()) {
            buf.append(c);
        }
        out.writeObject(buf);

        // ... and attributes
        int runLimit = 0;
        while (acit.setIndex(runLimit) != CharacterIterator.DONE) {

            Map<AttributedCharacterIterator.Attribute,Object> map =
                acit.getAttributes();
            int runStart = acit.getRunStart();
            runLimit = acit.getRunLimit();

            out.writeInt(runStart);
            out.writeInt(runLimit);

            Set<Map.Entry<AttributedCharacterIterator.Attribute,Object>> 
                    entrySet = map.entrySet();
            out.writeInt(entrySet.size());

            Iterator<Map.Entry<AttributedCharacterIterator.Attribute,Object>>
                    mit = entrySet.iterator();
            while (mit.hasNext()) {
                Map.Entry<AttributedCharacterIterator.Attribute,Object> entry =
                    mit.next();
                out.writeObject(entry.getKey());
                out.writeObject(entry.getValue());
            }
        }
        out.writeInt(END_OF_ATTRIBUTED_STRING);
    }

    private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {

        // Deserialize non-transient attributes
        in.defaultReadObject();

        // Read text and create the transient object ...
        StringBuffer buf = (StringBuffer)in.readObject();
        m_as = new AttributedString(buf.toString());

        // ... and add run data
        int runStart;
        while ((runStart = in.readInt()) != END_OF_ATTRIBUTED_STRING) {
            int runLimit = in.readInt();

            int nAttr = in.readInt();
            for (int iAttr = 0; iAttr < nAttr; iAttr++) {
                AttributedCharacterIterator.Attribute attr =
                    (AttributedCharacterIterator.Attribute)in.readObject();
                Object value = in.readObject();
                m_as.addAttribute(attr, value, runStart, runLimit);
            }
        }
    }

    // Attributes
    public AffineTransform getTransform() {
        return (AffineTransform)m_trf.clone();
    }
    public void setTransform(AffineTransform trf) {
        m_trf = trf;
    }
    public int getJustification() {
        return m_just;
    }
    public void setJustification(int just) {
        m_just = just;
    }
    public Paint getDrawPaint() {
        return m_draw;
    }
    public void setDrawPaint(Paint paint) {
        m_draw = paint;
    }
    public Paint getFillPaint() {
        return m_fill;
    }
    public void setFillPaint(Paint paint) {
        m_fill = paint;
    }
    public AttributedString getAttributedString() {
        return m_as;
    }
    public void setAttributedString(AttributedString as) {
        m_as = as;
    }

    // Manipulation
    public void transformBy(AffineTransform at) {
        m_trf.preConcatenate(at);
    }
}