/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.Set;

import cib.util.geo.Geo2D;

/**
 * An object of this class controls the coordinates of the database and their
 * transformation.</br>
 *
 * User coordinates are the coordinates that are presented to the user. It is
 * possible to define a coordinate system in the user coordinate space. A user
 * coordinate can be defined in terms of [mm]. It is possible to enter user
 * coordinates for instance in [m], [cm], [mm] or [km]. The definition of user
 * coordinates can be changed at any time.</br>
 *
 * User coordinates are mapped to and stored persistently in world coordinate
 * space. The world coordinates remain fixed when the user coordinates are
 * changed. The mapping between user and world coordinates allows to control the
 * accuracy of coordinates.</br>
 * 
 * To achieve the same accuracy at any point in coordinate space it is
 * recommended to map the user coordinate interval [minuc, maxuc] to the world
 * coordinate interval [-1, 1].</br>
 * 
 * The mapping between paper and user coordinates is specified by a scale
 * factor.</br>
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.19 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class CoordSpace implements Serializable {

    private static final long serialVersionUID = 0L;

    private static CoordSpace s_instance = null;

    private double m_scale = 100;

    private double m_wu = 1;    	// [m] per NU: rename to m_metersPerNu
    private double m_nu = 10000;    // NUs per WU: rename to m_nuPerWu

    private AffineTransform m_userCrdTransform = new AffineTransform();

    private AffineTransform m_userToWorldTransform = null;
    private AffineTransform m_worldToUserTransform = null;

    private String m_format = "0.0###########";
    private String m_shortFormat = "0.0###";
    
    private transient Set<Listener> m_listeners = new HashSet<Listener>();

    // Constructors
    private CoordSpace() {
        if (s_instance == null)
            s_instance = this;
        _invalidate();
    }

    /**
     * Returns an instance of the Coordinate Space.
     *
     * @return the Coordinate Space instance
     */
    public static CoordSpace getCoordSpace() {
        if (s_instance == null)
            new CoordSpace();
        return s_instance;
    }

    /**
     * Sets the scale between user and paper coordinates. Notifies all added
     * listeners about the event.
     *
     * @param scale the scale factor from paper to user units
     */
    public void setScale(double scale) {
        m_scale = scale;
        _notifyScaleChanged();
    }
    /**
     * Returns the scale factor
     *
     * @return the scale factor from paper to user units
     */
    public double getScale() {
        return m_scale;
    }

    /**
     * Assigns a number of natural units to one world unit.
     *
     * @param nuPerWu the number of natural units per world unit
     */
    public void setNaturalUnitsPerWorldUnit(double nuPerWu) {
        _invalidate();
        m_nu = nuPerWu;
    }
    /**
     * Returns the number of natural units assigned to one world unit.
     *
     * @return the number of natural units per world unit
     */
    public double getNaturalUnitsPerWorldUnit() {
        return m_nu;
    }

    /**
     * Defines the real length of a natural unit. Optionally, the mapping
     * between user and world coordinates can be adjusted automatically -- which
     * is strongly recommended. Listeners are notified by this event. The format
     * string is set to:
     * 
     * 
     *     [mm] :  "0.0"
     *     [cm] :  "0.0#"
     *     [dm] :  "0.0##"
     *     [m]  :  "0.0###"
     *     [km] :  "0.0###"
     *     else :  "0.0###"
     *
     * @param mm  the length of a natural unit in [mm]
     * @param format the format string
     * @param adjustNaturalUnitsPerWorldUnits
     *            true if the mapping between user and world coordinates should
     *            be adjusted.
     */
    public void setMillimetersPerNaturalUnit(double mm, String format,
            boolean adjustNaturalUnitsPerWorldUnits) {
        double scale = getMillimetersPerNaturalUnit() / mm;
        m_wu = mm / 1000.;

        if (format == null) {
            double eps = Geo2D.getEps();
            if (1 - eps <= mm && mm <= 1 + eps) {   // [mm]
                m_format = "0.0########";
                m_shortFormat = "0.0";
            }
            else if (10 - eps <= mm && mm <= 10 + eps) {    // [cm]
                m_format = "0.0#########";      
                m_shortFormat = "0.0#";      
            }
            else if (100 - eps <= mm && mm <= 100 + eps) {  // [dm]
                m_format = "0.0##########";     
                m_shortFormat = "0.0##";     
            }
            else if (1000 - eps <= mm && mm <= 1000 + eps) {    // [m]
                m_format = "0.0###########";    
                m_shortFormat = "0.0###";    
            }
            else if (1000000 - eps <= mm && mm <= 1000000 + eps) {  // [km]
                m_format = "0.000#";            
                m_shortFormat = "0.000#";            
            }
            else {
                m_format = "0.0###########";
                m_shortFormat = "0.0###";
            }
        }
        else {
            m_format = format;
        }
        
        _notifyNaturalUnitChanged();
        if (adjustNaturalUnitsPerWorldUnits)
            setNaturalUnitsPerWorldUnit(getNaturalUnitsPerWorldUnit() * scale);
    }
    /**
     * Returns the size of a natural unit in [mm].
     *
     * @return the size in [mm]
     */
    public double getMillimetersPerNaturalUnit() {
        return m_wu * 1000.;
    }

    /**
     * Returns the unit as a string.
     * 
     * @return the unit
     */
    public String getNaturalUnitString() {
        double mms = getMillimetersPerNaturalUnit();
        double eps = Geo2D.getEps();
        if (1 - eps <= mms && mms <= 1 + eps)
            return "mm";
        else if (10 - eps <= mms && mms <= 10 + eps)
            return "cm";
        else if (100 - eps <= mms && mms <= 100 + eps)
            return "dm";
        else if (1000 - eps <= mms && mms <= 1000 + eps)
            return "m";
        else if (1000000 - eps <= mms && mms <= 1000000 + eps)
            return "km";
        else 
            return mms + "*mm";
    }

    /**
     * Sets the format string, for instance "0.0###".
     * 
     * @param format
     */
    public void setFormat(String format) {
        m_format = format;
    }
    
    /**
     * Gets the format string, for instance "0.0###".
     * 
     * @return the format string
     */
    public String getFormat() {
        return m_format != null ? m_format : "0.0###########";
    }

    /**
     * Formats a coordinate.
     * 
     * @param x the coordinate to format
     * @return the formated coordinate
     */
    public String formatCoordinate(double x) {
        DecimalFormat nf = new DecimalFormat();
        nf.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.ENGLISH));
        nf.applyLocalizedPattern(getFormat());
        return nf.format(x);
    }

    
    /**
     * Sets the format string, for instance "0.0###".
     * 
     * @param format
     */
    public void setShortFormat(String format) {
        m_shortFormat = format;
    }
    
    /**
     * Gets the format string, for instance "0.0###".
     * 
     * @return the format string
     */
    public String getShortFormat() {
        return m_shortFormat != null ? m_shortFormat : "0.0###";
    }
    
    /**
     * Formats a coordinate.
     * 
     * @param x the coordinate to format
     * @return the formated coordinate
     */
    public String shortFormatCoordinate(double x) {
        DecimalFormat nf = new DecimalFormat();
        nf.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.ENGLISH));
        nf.applyLocalizedPattern(getShortFormat());
        return nf.format(x);
    }
    
    /**
     * Sets the user coordinate transformation.
     *
     * @param trf
     */
    public void setUserCrdTransform(AffineTransform trf) {
        _invalidate();
        m_userCrdTransform.setTransform(trf);
        _notifyUCSChanged();
    }
    /**
     * Returns the user coordinate transformation
     *
     * @return user coordinate transformation matrix
     */
    public AffineTransform getUserCrdTransform() {
        return (AffineTransform)m_userCrdTransform.clone();
    }

    /**
     * Gets the world to user transformation.
     *
     * @return the world to user transformation
     */
    public AffineTransform getWorldToUserTransform() {
        _validate();
        return (AffineTransform)m_worldToUserTransform.clone();
    }
    /**
     * Gets the user to world transformation.
     *
     * @return the user to world transformation
     */
    public AffineTransform getUserToWorldTransform() {
        _validate();
        return (AffineTransform)m_userToWorldTransform.clone();
    }

    private void _invalidate() {
        m_worldToUserTransform = null;
        m_userToWorldTransform = null;
    }
    private void _validate() {
        if (m_worldToUserTransform != null)
            return;

        // World to User transform
        try {
            double scale = m_nu;
            m_worldToUserTransform =
                AffineTransform.getScaleInstance(scale, scale);
            m_worldToUserTransform.preConcatenate(
                m_userCrdTransform.createInverse());

            // User to World transform
            m_userToWorldTransform = m_worldToUserTransform.createInverse();
        }
        catch (NoninvertibleTransformException e) {
            // Should never happen
            throw new IllegalStateException("Uncatchable error");
        }
    }

    // Serialization
    private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        // Deserialize non-transient attributes
        in.defaultReadObject();
        m_listeners = new HashSet<Listener>();
        s_instance = this;
    }

    // Notification
    /**
     * Listener for coordinate space events.
     *
     */
    public static interface Listener {
        /**
         * Called when the user coordinate system was changed.
         *
         * @param ucs the user coordinate transformation
         */
        public void ucsChanged(AffineTransform ucs);
        /**
         * Called when the scale was changed.
         *
         * @param scale the scale
         */
        public void scaleChanged(double scale);
        /**
         * Called when the size of the natural unit was changed.
         *
         * @param mm the size of a natural unit in [mm]
         */
        public void naturalUnitChanged(double mm);
    }
    /**
     * Adds a listener to the coordinate space.
     *
     * @param l the listener to be added
     * @return true if the listener was not already added
     */
    public boolean addListener(Listener l) {
        return m_listeners.add(l);
    }
    /**
     * Removes a listener from coordinate space.
     *
     * @param l the listener to be removed
     * @return true if the listener could be removed
     */
    public boolean removeListener(Listener l) {
        return m_listeners.remove(l);
    }
    private void _notifyUCSChanged() {
        Iterator<Listener> it = m_listeners.iterator();
        while (it.hasNext()) {
            it.next().ucsChanged(getUserCrdTransform());
        }
    }
    private void _notifyScaleChanged() {
        Iterator<Listener> it = m_listeners.iterator();
        while (it.hasNext()) {
            it.next().scaleChanged(m_scale);
        }
    }
    private void _notifyNaturalUnitChanged() {
        Iterator<Listener> it = m_listeners.iterator();
        while (it.hasNext()) {
            it.next().naturalUnitChanged(m_wu * 1000.);
        }
    }

}