/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.awt.geom.Point2D;
import java.io.Serializable;
import java.util.Properties;

/** 
 * Represents a rectangular grid.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.7 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class Grid implements Serializable, Cloneable {

    private static final long serialVersionUID = 0L;

    private double m_dx = 1;            // Spacing x
    private double m_dy = 1;            // Spacing y
    private double m_ox = 0;            // Origin x
    private double m_oy = 0;            // Origin y
    private double m_angle = 0;

    public static void main(String[] args) {
        Grid grd = new Grid(2, 1);
        System.out.println("All snaps should be (2, 1):");
        System.out.println("\t" + grd.snap(new Point2D.Double(1.1, 0.6), null));
        System.out.println("\t" + grd.snap(new Point2D.Double(1.1, 1.4), null));
        System.out.println("\t" + grd.snap(new Point2D.Double(2.9, 0.6), null));
        System.out.println("\t" + grd.snap(new Point2D.Double(2.9, 1.4), null));
    }

    // Constructors
    public Grid() {
    }
    public Grid(double dx, double dy) {
        m_dx = dx;
        m_dy = dy;
    }
    public Grid(Grid grid) {
        m_dx = grid.m_dx;
        m_dy = grid.m_dy;
        m_ox = grid.m_ox;
        m_oy = grid.m_oy;
        m_angle = grid.m_angle;
    }

    // Cloning
    public Object clone() {
        try {
            return super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();  // Should never happen
        }
    }
    
    // Manipulators
    public void setGrid(double dx, double dy) {
        m_dx = dx;
        m_dy = dy;
    }
    public void setOrigin(Point2D pnt) {
        m_ox = pnt.getX();
        m_oy = pnt.getY();
    }
    public void setOrigin(double x, double y) {
        m_ox = x;
        m_oy = y;
    }
    public void setAngle(double angle) {
        m_angle = angle;
    }

    // Accessors
    public double getDx() {
        return m_dx;
    }
    public double getDy() {
        return m_dy;
    }
    public Point2D getOrigin() {
        return getOrigin(null);
    }
    public Point2D getOrigin(Point2D dst) {
        if (dst == null)
            dst = new Point2D.Double();
        dst.setLocation(m_ox, m_oy);
        return dst;
    }
    public double getOriginX() {
        return m_ox;
    }
    public double getOriginY() {
        return m_oy;
    }

    public double getAngle() {
        return m_angle;
    }

    // Properties
    public void toProperties(Properties prop, String prefix) {
        double mPerNU = 1;
        CoordSpace cs = CoordSpace.getCoordSpace();
        if (cs != null)
            mPerNU = .001 * cs.getMillimetersPerNaturalUnit();
        prop.setProperty(prefix + "grid_dx", Double.toString(m_dx * mPerNU));
        prop.setProperty(prefix + "grid_dy", Double.toString(m_dy * mPerNU));
        prop.setProperty(prefix + "grid_ox", Double.toString(m_ox * mPerNU));
        prop.setProperty(prefix + "grid_oy", Double.toString(m_oy * mPerNU));
        prop.setProperty(prefix + "grid_angle", Double.toString(m_angle));
    }
    public void fromProperties(Properties prop, String prefix) {
        try {
            double mPerNU = 1;
            CoordSpace cs = CoordSpace.getCoordSpace();
            if (cs != null)
                mPerNU = .001 * cs.getMillimetersPerNaturalUnit();
            m_dx = Double.parseDouble(prop.getProperty(prefix + "grid_dx",
                    "1")) / mPerNU;
            m_dy = Double.parseDouble(prop.getProperty(prefix + "grid_dy",
                    "1")) / mPerNU;
            m_ox = Double.parseDouble(prop.getProperty(prefix + "grid_ox",
                    "0")) / mPerNU;
            m_oy = Double.parseDouble(prop.getProperty(prefix + "grid_oy",
                    "0")) / mPerNU;
            m_angle = Double.parseDouble(prop.getProperty(prefix + "grid_angle",
                    "0"));
        }
        catch (Throwable e) {
        }
    }
    
    // Calculations
    public Point2D snap(Point2D pntSrc, Point2D pntDst) {
        if (pntDst == null)
            pntDst = new Point2D.Double();
        double nx = Math.rint((pntSrc.getX() - m_ox) / m_dx);
        double ny = Math.rint((pntSrc.getY() - m_oy) / m_dy);
        pntDst.setLocation(m_ox + nx * m_dx, m_oy + ny * m_dy);
        return pntDst;
    }
}