/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;

import cib.util.coll.CollUtilities;
import cib.util.coll.ReverseAccessMap;

/**
 * Represents a name space by mapping objects (sometimes called "temporary IDs")
 * to names (sometimes called "persistent IDs").
 * Objects are represented by an Object reference, Names are represented by a
 * String reference.
 * Within a name space unique names consisting of a chosen prefix and a number
 * can be generated.
 * A name consists of a pure name and, optionally, a separator and an extension.
 * The extension can be set by the application and is stored as an attribute in
 * the NameSpace class (an empty extension is default). Names can be passed with
 * or without extension to all NameSpace methods. All methods return extended
 * names. Name instances are always stored as extended names.
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.12 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class NameSpace implements Serializable {

    private static final long serialVersionUID = 0L;

    public final char SEPARATOR = '@';

    private String m_nameExt = "";

    private ReverseAccessMap<Object,String> m_map =
        CollUtilities.bijectiveReverseAccessMap(new HashMap<Object, String>());

    private Map<String, Long> m_generatedNumbers = new HashMap<String, Long>();

    public void setNameExtension(String nameExt) {
        if (nameExt.indexOf(SEPARATOR) != -1)
            throw new IllegalArgumentException(
                    "Name extension must not contain a " + SEPARATOR);
        m_nameExt = nameExt;
    }

    public String getNameExtension() {
        return m_nameExt;
    }

    public String extendedName(String name) {
        // if (m_nameExt == "")
        if (m_nameExt.equals(""))
            return name;
        int pos = name.lastIndexOf(SEPARATOR);
        return pos == -1 ?
                name + SEPARATOR + m_nameExt :
                    name;
    }

    public String pureName(String name) {
        int pos = name.lastIndexOf(SEPARATOR);
        return pos == -1 ?
                name :
                    name.substring(0, pos);
    }

    public String generateName() {
        return generateName("obj");
    }

    public String generateName(String prefix) {
        String extPrefix = extendedName(prefix);
        if (! m_generatedNumbers.containsKey(extPrefix)) {
            m_generatedNumbers.put(extPrefix, new Long(0));
        }
        long number = m_generatedNumbers.get(extPrefix).longValue();
        m_generatedNumbers.put(extPrefix, new Long(number+1));
        return extendedName(prefix + Long.toString(number));
    }

    public void clear() {
        m_map.clear();
    }

    public void put(Object object, String name) {
        if (object == null || name == null)
            throw new NullPointerException();
        String extName = extendedName(name);
        if (m_map.containsKey(object) || m_map.containsValue(extName))
            throw new IllegalArgumentException("Can't change name");
        m_map.put(object, extName);
    }

    public String getName(Object object) {
        return m_map.get(object);
    }

    public Object getObject(String name) {
        return m_map.getKey(extendedName(name));
    }

    public boolean containsObject(Object object) {
        return m_map.containsKey(object);
    }

    public boolean containsName(String name) {
        return m_map.containsValue(extendedName(name));
    }

    public void remove(Object object) {
        m_map.remove(object);
    }

    public boolean isEmpty() {
        return m_map.isEmpty();
    }

    public int size() {
        return m_map.size();
    }

    // @added by beer
    // can set equal or greater number for a prefix
    public String setPrefixNumber(String prefix, long number) {
        String extPrefix = extendedName(prefix);
        if (m_generatedNumbers.containsKey(extPrefix)) {
            long _number = m_generatedNumbers.get(extPrefix).longValue();
            if (number < _number)
                throw new IllegalArgumentException(
                        "Current number for prefix '" + prefix + "' is " + number
                );
        }
        m_generatedNumbers.put(extPrefix, new Long(number));
        return extendedName(prefix + Long.toString(number));
    }
    public long getPrefixNumber(String prefix) {
        String extPrefix = extendedName(prefix);
        if (m_generatedNumbers.containsKey(extPrefix))
            return m_generatedNumbers.get(extPrefix).longValue();
        return -1;
    }

    public String toString() {
        return m_map.toString();
    }
}