/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.io.Serializable;
import java.util.Collection;
import java.util.Map;
import java.util.Set;

/**
 * An implementation for a consistent map.
 *
 * @author Berthold Firmenich, Torsten Richter
 * @version $Revision: 1.8 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
class _ConsistentHashMap<K,V> implements ReverseAccessMap<K,V>,
        Serializable {

    private static final long serialVersionUID = 0L;

    private ReverseAccessMap<K,V> m_map;
    private Set<K> m_keys;
    private Set<V> m_values;

    // Constructors
    protected _ConsistentHashMap(ReverseAccessMap<K,V> map,
                                 ObservableSet<K> keys, ObservableSet<V> values) {
        if (!map.isEmpty()) // Currently only empty maps supported!
            throw new IllegalArgumentException();
        m_map = map;

        // Attach listeners to the defining sets
        keys.addListener(new ObservableSet.Listener<K>() {
            private static final long serialVersionUID = 0L;

            public void wasRemoved(K key) {
                m_map.remove(key);
            }
        });

        values.addListener(new ObservableSet.Listener<V>() {
            private static final long serialVersionUID = 0L;

            public void wasRemoved(V value) {
                Set<K> keySet = m_map.keySet(value);
                for (K key : keySet)
                    m_map.remove(key);
            }
        });
    }

    // ReverseAccessMap implementation
    public K getKey(V value) {
        return m_map.getKey(value);
    }

    public Set<K> keySet(V value) {
        return m_map.keySet(value);
    }

    public void clear() {
        m_map.clear();
    }

    public boolean containsKey(Object key) {
        return m_map.containsKey(key);
    }

    public boolean containsValue(Object value) {
        return m_map.containsValue(value);
    }

    public Set<Entry<K,V>> entrySet() {
        return m_map.entrySet();
    }

    public V get(Object key) {
        return m_map.get(key);
    }

    public boolean isEmpty() {
        return m_map.isEmpty();
    }

    public Set<K> keySet() {
        return m_map.keySet();
    }

    public V put(K key, V value) {
        if (!m_keys.contains(key) || !m_values.contains(value))
            throw new IllegalArgumentException();
        return m_map.put(key, value);
    }

    public void putAll(Map<? extends K, ? extends V> t)
            throws UnsupportedOperationException {
        if (!m_keys.containsAll(t.keySet())
                || !m_values.containsAll(t.values()))
            throw new IllegalArgumentException();
        m_map.putAll(t);
    }

    public V remove(Object key) {
        return m_map.remove(key);
    }

    public int size() {
        return m_map.size();
    }

    public Collection<V> values() {
        return m_map.values();
    }

    public String toString() {
        return m_map.toString();
    }
}