/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.io.Serializable;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/**
 * A map implementation. Allows for efficient retrieval of the 1st and 2nd
 * component of the stored pairs.
 *
 * @author Berthold Firmenich, Torsten Richter
 * @version $Revision: 1.10 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
class _ReverseAccessHashMap<K,V> implements ReverseAccessMap<K,V>,
        Serializable {

    private static final long serialVersionUID = 0L;

    private Map<K,V> m_keyToValue;
    private Map<V, Set<K>> m_valueToKey;

    // Constructors
    protected _ReverseAccessHashMap(Map<K,V> map) {
        if (!map.isEmpty()) // Currently only empty maps supported!
            throw new IllegalArgumentException();
        m_keyToValue = map;
        m_valueToKey = new HashMap<V, Set<K>>();
    }

    // ReverseAccessMap implementation
    public K getKey(V value) {
        Set<K> keys = m_valueToKey.get(value);
        if (keys == null || keys.isEmpty())
            return null;
        return keys.iterator().next();
    }

    public Set<K> keySet(V value) {
        Set<K> keys = m_valueToKey.get(value);
        if (keys == null)
            keys = new HashSet<K>();
        return Collections.unmodifiableSet(keys); // Solves our problem!
    }

    // Map implementation
    public void clear() {
        m_keyToValue.clear();
        m_valueToKey.clear();
    }

    public boolean containsKey(Object key) {
        return m_keyToValue.containsKey(key);
    }

    public boolean containsValue(Object value) {
        return m_valueToKey.containsKey(value);
    }

    public Set<Entry<K,V>> entrySet() {
        return m_keyToValue.entrySet();
    }

    public V get(Object key) {
        return m_keyToValue.get(key);
    }

    public boolean isEmpty() {
        return m_keyToValue.isEmpty();
    }

    public Set<K> keySet() {
        return m_keyToValue.keySet();
    }

    public V put(K key, V value) {
        // Key already stored?
        V oldValue = m_keyToValue.get(key);
        if (oldValue != null) {
            if (oldValue == value)
                return value; // Pair already stored

            // Remove old value
            Set<K> keys = m_valueToKey.get(oldValue);
            if (!keys.remove(key))
                throw new Error("Unexpected state");
            if (keys.isEmpty())
                m_valueToKey.remove(oldValue);
        }

        // Put pair
        m_keyToValue.put(key, value);
        Set<K> keys = m_valueToKey.get(value);
        if (keys == null) {
            keys = new HashSet<K>();
            m_valueToKey.put(value, keys);
        }
        keys.add(key);

        return oldValue;
    }

    public void putAll(Map<? extends K, ? extends V> t)
            throws UnsupportedOperationException {
        throw new UnsupportedOperationException();
    }

    public V remove(Object key) {
        V value = m_keyToValue.remove(key);
        if (value == null)
            return null;
        Set<K> keys = m_valueToKey.get(value);
        if (!keys.remove(key))
            throw new Error("Unexpected state");
        if (keys.isEmpty())
            m_valueToKey.remove(value);
        return value;
    }

    public int size() {
        return m_keyToValue.size();
    }

    public Collection<V> values() {
        return m_keyToValue.values();
    }

    public String toString() {
        return m_keyToValue.toString();
    }
}