/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or write
 * to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.geo;

import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.Area;
import java.awt.geom.CubicCurve2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.IllegalPathStateException;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.QuadCurve2D;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;

/** 
 * A <code>Path2D</code> represents a two dimensional geometric path
 * constructed from gaps, lines, quadratic curves, cubic curves and arcs.
 * <p>
 *
 * @author Christian Koch, Berthold Firmenich
 * @version $Revision: 1.39 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/08/02 11:03:55 $
 */
public class Path2D implements Cloneable, Serializable {

    public static final int SEG_MOVETO = 0;
    public static final int SEG_LINETO = 1;
    public static final int SEG_QUADTO = 2;
    public static final int SEG_CUBICTO = 3;
    public static final int SEG_ARCTO = 4;
    public static final int SEG_CLOSE = 5;

    public static final int NULL = -1;

    /*
     * The following table defines the priority of segments if a shared vertex
     * is removed. For instance, if a vertex between a SEG_MOVETO (priority 5)
     * and a SEG_LINETO segment (priority 1) is removed then the SEG_LINETO
     * will be removed.
     */
    private final static int[] SEG_TYPE_PRIORITY = new int[] {
        5,      // SEG_MOVETO,
        1,      // SEG_LINETO,
        3,      // SEG_QUADTO,
        4,      // SEG_CUBICTO,
        2,      // SEG_ARCTO,
        0,      // SEG_CLOSE
    };
    
    private static final long serialVersionUID = 0L;
    
    private List<Primitive> m_path = new ArrayList<Primitive>();
        // (Vertex, Segment, ..., Vertex)
    
    private transient Map<Integer,Point2D> m_pnts =
        new HashMap<Integer,Point2D>();
    private transient Map<Integer,Primitive> m_name2prim =
        new HashMap<Integer,Primitive>();
    private transient Map<Primitive,Integer> m_prim2name =
        new HashMap<Primitive,Integer>();

    private long m_num = 0L;

    private boolean m_autoCleaning = true;
    private int m_subVersion = 1;
    private double m_eps = 0;
    
    // =========================================================================
    // Constructors

    /**
     * Creates an empty path.
     * 
     */
    public Path2D() {
    }

    /**
     * Creates a path that starts at a given location.
     * 
     * @param sp the start point
     */
    public Path2D(Point2D sp) {
        moveTo(sp);
    }

    /**
     * Creates a path that starts at a given location.
     * 
     * @param x the x coordinate of the start point
     * @param y the y coordinate of the start point
     */
    public Path2D(double x, double y) {
        moveTo(new Point2D.Double(x, y));
    }

    // =========================================================================
    // Resetting a path
    
    /**
     * Resets a Path2D by reducing it to one vertex only.
     * 
     * @param p the location of the vertex
     */
    public void reset(Point2D p) {
        reset(p.getX(), p.getY());
    }
    
    /**
     * Resets a Path2D by reducing it to one vertex only.
     * 
     * @param x the x location of the vertex
     * @param y the y location of the vertex
     */
    public void reset(double x, double y) {
        // Remove all primitives
        int n = m_path.size();
        for (int i = n - 1; i > 0; i--) {
            Primitive prm = m_path.get(i);
            m_path.remove(i);
            prm.destruct();
        }

        // Set location of first vertex
        Vertex vtx = (Vertex)m_path.get(0);
        vtx.setPoint(x, y);
    }

    // =========================================================================
    // Epsilon value for geometric arc calculations

    public void setEps(double eps) {
        if (eps <= Double.MIN_VALUE)
            throw new IllegalArgumentException();
        m_eps = eps;
    }
    public double getEps() {
        return m_eps <= Double.MIN_VALUE ? Geo2D.getEps() : m_eps;
    }
    
    private Ellipse2D _calculateEllipse2D(Point2D p1, Point2D p2, Point2D p3,
            Ellipse2D dst) {
        double svEps = Geo2D.getEps();
        try {
            Geo2D.setEps(getEps());
            Ellipse2D ell = Geo2D.calculateEllipse2D(p1, p2, p3, dst);
            return ell;
        }
        finally {
            Geo2D.setEps(svEps);
        }
    }

    private Arc2D _calculateArc2D(Point2D sp, Point2D p, Point2D ep, Arc2D dst) {
        double svEps = Geo2D.getEps();
        try {
            Geo2D.setEps(getEps());
            Arc2D arc = Geo2D.calculateArc2D(sp, p, ep, dst);
            return arc;
        }
        finally {
            Geo2D.setEps(svEps);
        }
    }
    
    // =========================================================================
    // Path generation

    /**
     * Moves the path to a given location.
     * 
     * @param p the location
     * @return the segment name or NULL
     */
    public int moveTo(Point2D p) {
        return moveTo(p.getX(), p.getY());
    }

    /**
     * Moves the path to a given location.
     * 
     * @param x the x coordinate of the location
     * @param y the y coordinate of the location
     * @return the segment name or NULL
     */
    public int moveTo(double x, double y) {
        if (m_path.size() == 1)
            return NULL;
        if (m_path.isEmpty()) {
            int idx = addPoint(new Point2D.Double(x, y));
            Vertex v = new Vertex(idx);
            m_path.add(v);
            return NULL;
        }
        int nPrims = m_path.size();
        Segment lastSeg = (Segment)m_path.get(nPrims-2);
        if (! (lastSeg instanceof MoveSegment)) {
            Segment seg = new MoveSegment();
            m_path.add(seg);
            int idx = addPoint(new Point2D.Double(x, y));
            Vertex v = new Vertex(idx);
            m_path.add(v);
            return m_prim2name.get(seg);
        }
        return NULL;
    }

    /**
     * Adds a line segment.
     * 
     * @param p the line's end point
     * @return the segment name
     */
    public int lineTo(Point2D p) {
        return lineTo(p.getX(), p.getY());
    }

    private void _checkPathStateForAdd() {
        if (m_path.size() == 0)
            throw new IllegalPathStateException(
                "Can't add segment, needs moveto");
        if (m_path.size() > 2) {
            int nPrims = m_path.size();
            Segment lastSeg = (Segment)m_path.get(nPrims-2);
            if (lastSeg instanceof CloseSegment)
                throw new IllegalPathStateException(
                    "Can't add segment, needs moveto");
        }
    }
    
    /**
     * Adds a line segment.
     * 
     * @param x the x coordinate of the line's end point
     * @param y the y coordinate of the line's end point
     * @return the segment name
     */
    public int lineTo(double x, double y) {
        _checkPathStateForAdd();
        int idx = addPoint(new Point2D.Double(x, y));
        Segment seg = new LineSegment();
        m_path.add(seg);
        m_path.add(new Vertex(idx));
        return m_prim2name.get(seg);
    }

    /**
     * Adds a quadratic segment.
     * 
     * @param cp the control point
     * @param p the end point
     * @return the segment name
     */
    public int quadTo(Point2D cp, Point2D p) {
        return quadTo(cp.getX(), cp.getY(), p.getX(), p.getY());
    }

    /**
     * Adds a quadratic segment.
     * 
     * @param cx the x coordinate of the control point
     * @param cy the y coordinate of the control point
     * @param x the x coordinate of the end point
     * @param y the y coordinate of the end point
     * @return the segment name
     */
    public int quadTo(double cx, double cy, double x, double y) {
        _checkPathStateForAdd();
        int c_idx = addPoint(new Point2D.Double(cx, cy));
        int v_idx = addPoint(new Point2D.Double(x, y));
        Segment seg = new QuadSegment(c_idx);
        m_path.add(seg);
        m_path.add(new Vertex(v_idx));
        return m_prim2name.get(seg);
    }

    /**
     * Adds a cubic segment.
     * 
     * @param cp1 the first control point
     * @param cp2 the second control point
     * @param p the end point
     * @return the segment name
     */
    public int cubicTo(Point2D cp1, Point2D cp2, Point2D p) {
        return cubicTo(cp1.getX(), cp1.getY(), cp2.getX(), cp2.getY(), p.getX(),
            p.getY());
    }

    /**
     * Adds a cubic segment.
     * 
     * @param cx1 the x coordinate of the first control point
     * @param cy1 the y coordinate of the first control point
     * @param cx2 the x coordinate of the second control point
     * @param cy2 the y coordinate of the second control point
     * @param x the x coordinate of the end point
     * @param y the y coordinate of the end point
     * @return the segment name
     */
    public int cubicTo(double cx1, double cy1, double cx2, double cy2,
    double x, double y) {
        _checkPathStateForAdd();
        int c1_idx = addPoint(new Point2D.Double(cx1, cy1));
        int c2_idx = addPoint(new Point2D.Double(cx2, cy2));
        int v_idx = addPoint(new Point2D.Double(x, y));
        Segment seg = new CubicSegment(c1_idx, c2_idx);
        m_path.add(seg);
        m_path.add(new Vertex(v_idx));
        return m_prim2name.get(seg);
    }

    /**
     * Adds an arc segment.
     * 
     * @param ap a point on the arc
     * @param p the end point of the arc
     * @return the segment name
     */
    public int arcTo(Point2D ap, Point2D p) {
        return arcTo(ap.getX(), ap.getY(), p.getX(), p.getY());
    }
    
    /**
     * Adds an arc segment.
     * 
     * @param px the x coordinate of a point on the arc
     * @param py the y coordinate of a point on the arc
     * @param x the x coordinate of the end point
     * @param y the y coordinate of the end point
     * @return the segment name
     */
    public int arcTo(double px, double py, double x, double y) {
        _checkPathStateForAdd();
        int p_idx = addPoint(new Point2D.Double(px, py));
        ArcSegment seg = new ArcSegment(p_idx);
        m_path.add(seg);
        int ep_idx = addPoint(new Point2D.Double(x, y));
        m_path.add(new Vertex(ep_idx));
        return m_prim2name.get(seg);
    }
    
    /**
     * Closes the last subpath.
     * 
     * @return the segment name or NULL if there was a request for a moveto
     *         segment
     */
    public int close() {
        if (m_path.isEmpty())
            return -1;
        Vertex v = new Vertex(-1);
        Segment seg = new CloseSegment();
        m_path.add(seg);
        m_path.add(v);
        return m_prim2name.get(seg);
    }

    /**
     * Puts automatic cleaning on or off.
     * 
     * @param on puts autocleaning on if true
     */
    public void setAutoCleaning(boolean on) {
        m_autoCleaning = on;
    }
    
    /**
     * Returns whether automatic cleaning is on or off. 
     * 
     * @return true if auto cleaning is on, false otherwise
     */
    public boolean isAutoCleaning() {
        return m_autoCleaning;
    }
    
    /**
     * Cleans the path:
     * - Removes moveto segments at the end of the path
     * - Reduces subsequent moveto segments to one single moveto segment
     * - Removes empty subpaths, that is a (moveto, close) sequence or a close
     *   segment at the start of the path
     * - Removes a moveto segment at the start of the path
     * 
     */
    public void clean() {
        // Remove (MoveSegemnt, CloseSegment) sequences
        for (int i = 2; i < m_path.size() - 2; i += 2) {
            Segment segi = (Segment)m_path.get(i - 1);
            Segment segj = (Segment)m_path.get(i + 1);
            if (segi instanceof MoveSegment && segj instanceof CloseSegment) {
                Vertex vtxj = (Vertex)m_path.get(i + 2);
                vtxj.destruct();
                segj.destruct();
                m_path.remove(i + 2);
                m_path.remove(i + 1);
                i -= 2;
            }
        }

        // Remove a CloseSegment at the start of a path
        if (m_path.size() > 1) {
            Segment seg = (Segment)m_path.get(1);
            if (seg instanceof CloseSegment) {
                m_path.remove(2);
                m_path.remove(1);
            }
        }

        // Reduce subsequent MoveSegments to one single MoveSegment
        for (int i = 2; i < m_path.size() - 2; i += 2) {
            Segment segi = (Segment)m_path.get(i - 1);
            Segment segj = (Segment)m_path.get(i + 1);
            if (segi instanceof MoveSegment && segj instanceof MoveSegment) {
                Vertex vtxi = (Vertex)m_path.get(i);
                vtxi.destruct();
                segj.destruct();
                m_path.remove(i + 1);
                m_path.remove(i);
                i -= 2;
            }
        }

        // Remove MoveSegments at the end of the path
        int n = m_path.size();
        for (int i = n - 2; i > 0; i -= 2) {
            Segment seg = (Segment)m_path.get(i);
            if (! (seg instanceof MoveSegment))
                break;
            Vertex vtx = (Vertex)m_path.get(i + 1);
            vtx.destruct();
            seg.destruct();
            m_path.remove(i + 1);
            m_path.remove(i);
        }

        // Remove MoveSegments at the start of the path
        for (int i = 1; i < m_path.size(); i += 2) {
            Segment seg = (Segment)m_path.get(i);
            if (! (seg instanceof MoveSegment))
                break;
            Vertex vtx = (Vertex)m_path.get(0);
            vtx.destruct();
            seg.destruct();
            m_path.remove(1);
            m_path.remove(0);
            i -= 2;
        }
    }
    
    /**
     * Finds the start of the subpath of a given Vertex.
     * 
     * @param closeVertex a Vertex of the subpath
     * @return
     */
    private Vertex _startOfLoop(Vertex closeVertex) {
        int pos = m_path.indexOf(closeVertex);
        for (int i = pos - 1; i > 0; i -= 2) {
            Segment seg = (Segment)m_path.get(i);
            if (seg instanceof MoveSegment)
                return (Vertex)m_path.get(i + 1);
        }
        return (Vertex)m_path.get(0);
    }
    
    // =========================================================================
    // Shape representation

    /**
     * Returns a Shape representation of a given segment.
     * 
     * @param name the segment's name
     * @return the Shape
     */
    public Shape getSegmentShape(int name) {
        Object segment = m_prim2name.get(name);
        int segIndex = m_path.indexOf(segment);
        if (segIndex == -1)
            throw new IllegalArgumentException("Illegal segment name");
        Object start = m_path.get(segIndex - 1);
        Object end = m_path.get(segIndex + 1);
        if (! (segment instanceof Segment && start instanceof Vertex
            && end instanceof Vertex))
            throw new IllegalPathStateException();
        return ((Segment) segment).getShape((Vertex)start, (Vertex)end);
    }

    /**
     * Returns a Shape representation of the path.
     * 
     * @return the Shape
     */
    public Shape getShape() {
        java.awt.geom.Path2D path = new java.awt.geom.Path2D.Double();
        path.setWindingRule(java.awt.geom.Path2D.WIND_EVEN_ODD);
        if (m_path.size() >= 1) {
            Vertex vtx = (Vertex)m_path.get(0);
            Point2D pnt = vtx.getPoint();
            path.moveTo(pnt.getX(), pnt.getY());
        }
        for (int i = 1; i < m_path.size(); i += 2) {
            Object el = m_path.get(i);
            if (el instanceof Segment) {
                Segment s = (Segment)el;
                Vertex prev = (Vertex)m_path.get(i - 1);
                Vertex next = (Vertex)m_path.get(i + 1);
                switch (s.getType()) {
                case SEG_MOVETO:
                    path.moveTo(next.getPoint().getX(), next.getPoint().getY());
                    break;
                case SEG_CLOSE:
                    path.closePath();
                    break;
                default:
                    Shape shape = s.getShape(prev, next);
                    
                    double[] crds = new double[6];
                    for (PathIterator pit = shape.getPathIterator(null);
                            ! pit.isDone(); pit.next()) {
                        int seg = pit.currentSegment(crds);
                        switch (seg) {
                        case PathIterator.SEG_MOVETO :
                            // Nothing to do
                            break;
                        case PathIterator.SEG_LINETO :
                            path.lineTo(crds[0], crds[1]);
                            break;
                        case PathIterator.SEG_QUADTO :
                            path.quadTo(crds[0], crds[1], crds[2], crds[3]);
                            break;
                        case PathIterator.SEG_CUBICTO :
                            path.curveTo(crds[0], crds[1], crds[2], crds[3], 
                                    crds[4], crds[5]);
                            break;
                        default:
                            throw new InternalError("Unexpected segment type");
                        }
                    }
                }
            }
            else
                throw new IllegalPathStateException("Can't create shape.");
        }
        return path;
    }

    // =========================================================================
    // Path properties
    
    /**
     * Calculates the length of the path.
     * 
     * @return the path length
     */
    public double length() {
        // Set world unit epsilon
        double svEps = Geo2D.getEps();
        Geo2D.setEps(m_eps);

        // For each segment
        double len = 0;
        for (int i = 1; i < m_path.size(); i += 2) {
            Segment s = (Segment)m_path.get(i);
            Vertex prev = (Vertex)m_path.get(i - 1);
            Vertex next = (Vertex)m_path.get(i + 1);
            len += s.length(prev, next);
        }

        Geo2D.setEps(svEps);
        return len;
    }
    
    /**
     * Calculates the area of the path.
     * 
     * @return the path area
     */
    public double area() {
        // Is it closed?
        if (! isClosed())
            return Double.NaN;

        // Use CAG for the calculation
        Shape shape = getShape();
        if (shape == null)
            return Double.NaN;
        
        double area = 0;
        Area areaShape = new Area(shape);
        double[] subCurveStart = new double[2];
        double[] lastCrds = new double[2];
        double[] crds = new double[2];
        for (PathIterator pit = areaShape.getPathIterator(null, 
                .1 * Geo2D.getEps()); ! pit.isDone(); pit.next()) {
            int seg = pit.currentSegment(crds);
            switch (seg) {
            case PathIterator.SEG_MOVETO:
                subCurveStart[0] = crds[0];
                subCurveStart[1] = crds[1];
                lastCrds[0] = crds[0];
                lastCrds[1] = crds[1];
                break;
            case PathIterator.SEG_LINETO:
                // Line from lastCrds to crds
                area += (crds[0] - lastCrds[0]) * .5 * (lastCrds[1] + crds[1]);
                lastCrds[0] = crds[0];
                lastCrds[1] = crds[1];
                break;
            case PathIterator.SEG_CLOSE:
                // Line from lastCrds to subCurveStart
                area += (subCurveStart[0] - lastCrds[0]) * .5 * (lastCrds[1] +
                        subCurveStart[1]);
                break;
            default:
                throw new InternalError("Unexpected segment type");
            }
        }
        return area;
    }
    
    public boolean isClosed() {
        // Set world unit epsilon
        double svEps = Geo2D.getEps();
        Geo2D.setEps(m_eps);
        
        // Search gaps in the path
        int startPosLoop = 0;
        for (int i = 1; i < m_path.size(); i += 2) {
            Segment seg = (Segment)m_path.get(i);
            switch (seg.getType()) {
            case SEG_MOVETO:
                // Was the last loop closed?
                Vertex sVtx = (Vertex)m_path.get(startPosLoop);
                Vertex eVtx = (Vertex)m_path.get(i - 1);
                if (! Geo2D.equality(sVtx.getPoint(), eVtx.getPoint())) {
                    Geo2D.setEps(svEps);
                    return false;
                }
                
                // Save new loop start position
                startPosLoop = i + 1;
                break;
            case SEG_CLOSE:
                startPosLoop = i + 1;
                break;
            default:
            }
        }
        
        // Check the last point
        Vertex sVtx = (Vertex)m_path.get(startPosLoop);
        Vertex eVtx = (Vertex)m_path.get(m_path.size() - 1);
        boolean status = Geo2D.equality(sVtx.getPoint(), eVtx.getPoint());

        Geo2D.setEps(svEps);
        return status;
    }
    
    // =========================================================================
    // Points (for internal use, only)

    /**
     * Adds a point and assigns a unique name.
     * 
     * @param loc the location
     * @return the point's name
     */
    private int addPoint(Point2D loc) {
        int name = (int)m_num++;
        m_pnts.put(name, loc);
        return name;
    }
    
    // =========================================================================
    // Vertices
    
    /**
     * Returns the number of vertices.
     * 
     * @return the number of vertices
     */
    public int getVertexCount() {
        int n = m_path.isEmpty() ?
            0 : (m_path.size() + 1) / 2;
        return n;
    }
    
    /**
     * Gets the name of the vertex at a given index.
     * 
     * @param index the index of the vertex
     * @return the name of the vertex
     */
    public int getVertexName(int index) {
        if (index < 0 || index >= getVertexCount())
            throw new IndexOutOfBoundsException("Invalid vertex index " +
                index);
        int pathIndex = index + index;
        Vertex vtx = (Vertex)m_path.get(pathIndex);
        int name = m_prim2name.get(vtx);
        return name;
    }

    /**
     * Tests if a vertex of a given name exists.
     * 
     * @param name the name of the vertex
     * @return true if the vertex exists, false otherwise
     */
    public boolean containsVertex(int name) {
        Object obj = m_name2prim.get(name);
        return obj instanceof Vertex;
    }
    
    /**
     * Gets the location of a vertex.
     * 
     * @param name the name of the vertex
     * @param dst the destination object where the location is stored
     * @return the location
     */
    public Point2D getVertexLocation(int name, Point2D dst) {
        if (dst == null)
            dst = new Point2D.Double();
        Vertex vtx = (Vertex)m_name2prim.get(name);
        dst.setLocation(vtx.getPoint());
        return dst;
    }
    
    /**
     * Sets the location of a given vertex.
     * 
     * @param name the name of the vertex
     * @param location the location
     */
    public void setVertexLocation(int name, Point2D location) {
        Vertex vtx = (Vertex)m_name2prim.get(name);
        vtx.setPoint(location);
    }

    /**
     * Removes a vertex.
     * 
     * @param name the name of the vertex
     */
    public void removeVertex(int name) {
        
        // Don't remove the last Vertex
        if (m_path.size() <= 1)
            return;

        // Remove a Vertex and a Segment
        Vertex vtxi = (Vertex)m_name2prim.get(name);
        int i = m_path.indexOf(vtxi);
        if (i == -1)
            return;
        
        if (i == 0) { // First Vertex
            vtxi.destruct();
            Segment seg = (Segment)m_path.get(1);
            seg.destruct();
            m_path.remove(1);
            m_path.remove(0);
        }
        else if (i == m_path.size() - 1) { // Last Vertex
            vtxi.destruct();
            Segment seg = (Segment)m_path.get(i - 1);
            seg.destruct();
            m_path.remove(i);
            m_path.remove(i - 1);
        }
        else { // Interior Vertex
            Segment segi = (Segment)m_path.get(i - 1);
            Segment segj = (Segment)m_path.get(i + 1);
            int prioi = SEG_TYPE_PRIORITY[segi.getType()];
            int prioj = SEG_TYPE_PRIORITY[segj.getType()];
            
            // Keep the segment with the higher priority
            if (prioi >= prioj) {
                vtxi.destruct();
                Segment seg = (Segment)m_path.get(i + 1);
                seg.destruct();
                
                Vertex vtxj = (Vertex)m_path.get(i + 2);
                if (vtxj.isCloseVertex())
                    vtxj.toCloseVertex(/*close=*/ false);
                
                m_path.remove(i + 1);
                m_path.remove(i);
            }
            else {
                vtxi.destruct();
                Segment seg = (Segment)m_path.get(i - 1);
                seg.destruct();
                
                m_path.remove(i);
                m_path.remove(i - 1);
            }
        }

        // Clean the structure
        if (m_autoCleaning)
            clean();
    }

    // =========================================================================
    // Segments
    
    /**
     * Returns the number of segments.
     * 
     * @return the number of segments
     */
    public int getSegmentCount() {
        int n = m_path.isEmpty() ?
            0 : (m_path.size() - 1) / 2;
        return n;
    }

    /**
     * Gets the name of a segment at a given index.
     * 
     * @param index the segment index
     * @return the segment name
     */
    public int getSegmentName(int index) {
        if (index < 0 || index >= getSegmentCount())
            throw new IndexOutOfBoundsException("Invalid segment index " +
                index);
        int pathIndex = index + index + 1;
        Segment seg = (Segment)m_path.get(pathIndex);
        int name = m_prim2name.get(seg);
        return name;
    }

    /**
     * Returns the segment previous to a given vertex.
     * 
     * @param vertexName the vertex name
     * @return the previous segment name
     */
    public int getPreviousSegmentName(int vertexName) {
        Vertex vtx = (Vertex)m_name2prim.get(vertexName);
        int index = m_path.indexOf(vtx);
        if (index == 0)
            return -1;
        Segment seg = (Segment)m_path.get(--index);
        int name = m_prim2name.get(seg);
        return name;
    }
    
    /**
     * Returns the segment next to a given vertex.
     * 
     * @param vertexName the vertex name
     * @return the subsequent segment name
     */
    public int getNextSegmentName(int vertexName) {
        Vertex vtx = (Vertex)m_name2prim.get(vertexName);
        int index = m_path.indexOf(vtx);
        if (index >= m_path.size() - 1)
            return -1;
        Segment seg = (Segment)m_path.get(++index);
        int name = m_prim2name.get(seg);
        return name;
    }
    
    /**
     * Tests if a segment of a given name exists.
     * 
     * @param name the name of the segment
     * @return true if the segment exists, false otherwise
     */
    public boolean containsSegment(int name) {
        Object obj = m_name2prim.get(name);
        return obj instanceof Segment;
    }

    /**
     * Sets a segment of a given name. Please note that the segment type can be
     * changed.
     * 
     * @param name the name of the Segment in the Path
     * @param type the Segment type. One of SEG_MOVETO, SEG_LINETO, SEG_QUADTO,
     *      SEG_CUBICTO, SEG_ARCTO or SEG_CLOSE
     * @param crds the Segment coordinates
     *      SEG_MOVETO: start point, end point
     *      SEG_LINETO: start point, end point
     *      SEG_QUADTO: start point, control point, end point
     *      SEG_CUBICTO: start point, control point 1, control point 2, end point
     *      SEG_ARCTO: start point, point on arc, end point
     *      SEG_CLOSE: no coordinates
     */
    public void setSegment(int name, int type, double[] crds) {
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex previousVtx = (Vertex)m_path.get(pathIndex - 1);
        Vertex nextVtx = (Vertex)m_path.get(pathIndex + 1);
        switch (type) {
        case SEG_MOVETO:
            if (seg.getType() != SEG_MOVETO) {
                seg = new MoveSegment(name);
                m_path.set(pathIndex, seg);
            }
            seg.setSegment(previousVtx, nextVtx, crds);
            break;
        case SEG_LINETO:
            if (seg.getType() != SEG_LINETO) {
                seg = new LineSegment(name);
                m_path.set(pathIndex, seg);
            }
            seg.setSegment(previousVtx, nextVtx, crds);
            break;
        case SEG_QUADTO:
            if (seg.getType() != SEG_QUADTO) {
                int ctrl = addPoint(new Point2D.Double(crds[2], crds[3]));
                seg = new QuadSegment(name, ctrl);
                m_path.set(pathIndex, seg);
            }
            seg.setSegment(previousVtx, nextVtx, crds);
            break;
        case SEG_CUBICTO:
            if (seg.getType() != SEG_CUBICTO) {
                int ctrl1 = addPoint(new Point2D.Double(crds[2], crds[3]));
                int ctrl2 = addPoint(new Point2D.Double(crds[4], crds[5]));
                seg = new CubicSegment(name, ctrl1, ctrl2);
                m_path.set(pathIndex, seg);
            }
            seg.setSegment(previousVtx, nextVtx, crds);
            break;
        case SEG_ARCTO:
            if (seg.getType() != SEG_ARCTO) {
                int pnt = addPoint(new Point2D.Double(crds[2], crds[3]));
                seg = new ArcSegment(name, pnt);
                m_path.set(pathIndex, seg);
            }
            seg.setSegment(previousVtx, nextVtx, crds);
            break;
        case SEG_CLOSE:
            if (seg.getType() != SEG_CLOSE) {
                seg = new CloseSegment(name);
                m_path.set(pathIndex, seg);
            }
            seg.setSegment(previousVtx, nextVtx, crds);
            break;
        default:
            throw new IllegalPathStateException();
        }
    }
    
    /**
     * Gets a segment.
     * 
     * @param name the name of the Segment in the Path
     * @param crds the Segment coordinates
     *      SEG_MOVETO: start point, end point
     *      SEG_LINETO: start point, end point
     *      SEG_QUADTO: start point, control point, end point
     *      SEG_CUBICTO: start point, control point 1, control point 2, end point
     *      SEG_ARCTO: start point, point on arc, end point
     *      SEG_CLOSE: no coordinates
     * @return the Segment type. One of SEG_MOVETO, SEG_LINETO, SEG_QUADTO,
     *      SEG_CUBICTO, SEG_ARCTO or SEG_CLOSE
     */
    public int getSegment(int name, double[] crds) {
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex startVtx = (Vertex)m_path.get(pathIndex - 1);
        Vertex endVtx = (Vertex)m_path.get(pathIndex + 1);
        return seg.getSegment(startVtx, endVtx, crds);
    }
    
    /**
     * Calculates a point on the segment. This point will usually be the center
     * point. In the case of quadratic and cubic curves De Casteljau subdivision
     * is performed.
     * 
     * @param name the name of the segment
     * @param pnt the point to store the coordinates or null
     * @return the point object describing the location
     */
    public Point2D getPointOnSegment(int name, Point2D pnt) {
        if (pnt == null)
            pnt = new Point2D.Double();
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex startVtx = (Vertex)m_path.get(pathIndex - 1);
        Vertex endVtx = (Vertex)m_path.get(pathIndex + 1);
        return seg.getPointOnSegment(startVtx, endVtx, pnt);
    }

    /**
     * Calculates the location of the point representing the segment. The point
     * must not necessarily be located on the segment curve.
     * 
     * @param name the name of the segment
     * @param pnt the point to store the coordinates or null
     * @return the point object describing the location
     */
    public Point2D getSegmentRepresentativeLocation(int name, Point2D pnt) {
        if (pnt == null)
            pnt = new Point2D.Double();
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex startVtx = (Vertex)m_path.get(pathIndex - 1);
        Vertex endVtx = (Vertex)m_path.get(pathIndex + 1);
        return seg.getRepresentativeLocation(startVtx, endVtx, pnt);
    }

    /**
     * Gets the name of a segment's start vertex.
     * 
     * @param name the segment name
     * @return the start vertex name
     */
    public int getSegmentStartVertex(int name) {
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex vtx = (Vertex)m_path.get(pathIndex - 1);
        return m_prim2name.get(vtx);
    }
    
    /**
     * Gets the name of a segment's end vertex.
     * 
     * @param name the segment name
     * @return the end vertex name
     */
    public int getSegmentEndVertex(int name) {
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex vtx = (Vertex)m_path.get(pathIndex + 1);
        return m_prim2name.get(vtx);
    }
    
    /**
     * Splits a segment at the center point. Moveto and Close segments cannot be
     * splitted. The original segment remains and the name of the newly inserted
     * segment is returned.
     * 
     * @return the name of the newly created segment.
     */
    public int splitSegment(int name) {
        Segment seg = (Segment)m_name2prim.get(name);
        int pathIndex = m_path.indexOf(seg);
        Vertex startVtx = (Vertex)m_path.get(pathIndex - 1);
        Vertex endVtx = (Vertex)m_path.get(pathIndex + 1);
        double[] crds = new double[8];
        int type = seg.getSegment(startVtx, endVtx, crds);
        
        Point2D center = getPointOnSegment(name, null);
        Vertex newVtx = new Vertex(addPoint(center));

        switch (type) {
        case SEG_LINETO: {
            Segment newSeg = new LineSegment();
            m_path.add(pathIndex+1, newVtx);
            m_path.add(pathIndex+2, newSeg);

            crds[2] = center.getX();
            crds[3] = center.getY();
            seg.setSegment(startVtx, newVtx, crds);
            
            crds[0] = center.getX();
            crds[1] = center.getY();
            crds[2] = endVtx.getPoint().getX();
            crds[3] = endVtx.getPoint().getY();
            newSeg.setSegment(newVtx, endVtx, crds);
            
            return m_prim2name.get(newSeg);
        }
        case SEG_QUADTO: {
            Segment newSeg = new QuadSegment(addPoint(new Point2D.Double()));
            m_path.add(pathIndex+1, newVtx);
            m_path.add(pathIndex+2, newSeg);
            
            QuadCurve2D quad = new QuadCurve2D.Double(crds[0], crds[1],
                crds[2], crds[3], crds[4], crds[5]);
            QuadCurve2D left = new QuadCurve2D.Double();
            QuadCurve2D right = new QuadCurve2D.Double();
            quad.subdivide(left, right);

            crds[0] = left.getX1();
            crds[1] = left.getY1();
            crds[2] = left.getCtrlX();
            crds[3] = left.getCtrlY();
            crds[4] = left.getX2();
            crds[5] = left.getY2();
            seg.setSegment(startVtx, newVtx, crds);
            
            crds[0] = right.getX1();
            crds[1] = right.getY1();
            crds[2] = right.getCtrlX();
            crds[3] = right.getCtrlY();
            crds[4] = right.getX2();
            crds[5] = right.getY2();
            newSeg.setSegment(newVtx, endVtx, crds);
            
            return m_prim2name.get(newSeg);
        }
        case SEG_CUBICTO: {
            Segment newSeg = new CubicSegment(addPoint(new Point2D.Double()),
                addPoint(new Point2D.Double()));
            m_path.add(pathIndex+1, newVtx);
            m_path.add(pathIndex+2, newSeg);
            
            CubicCurve2D cubic = new CubicCurve2D.Double(crds[0], crds[1],
                crds[2], crds[3], crds[4], crds[5], crds[6], crds[7]);
            CubicCurve2D left = new CubicCurve2D.Double();
            CubicCurve2D right = new CubicCurve2D.Double();
            cubic.subdivide(left, right);

            crds[0] = left.getX1();
            crds[1] = left.getY1();
            crds[2] = left.getCtrlX1();
            crds[3] = left.getCtrlY1();
            crds[4] = left.getCtrlX2();
            crds[5] = left.getCtrlY2();
            crds[6] = left.getX2();
            crds[7] = left.getY2();
            seg.setSegment(startVtx, newVtx, crds);
            
            crds[0] = right.getX1();
            crds[1] = right.getY1();
            crds[2] = right.getCtrlX1();
            crds[3] = right.getCtrlY1();
            crds[4] = right.getCtrlX2();
            crds[5] = right.getCtrlY2();
            crds[6] = right.getX2();
            crds[7] = right.getY2();
            newSeg.setSegment(newVtx, endVtx, crds);
            
            return m_prim2name.get(newSeg);
        }
        case SEG_ARCTO: {
            Segment newSeg = new ArcSegment(addPoint(new Point2D.Double()));
            m_path.add(pathIndex+1, newVtx);
            m_path.add(pathIndex+2, newSeg);

            // Setup transformation
            Point2D start = new Point2D.Double(crds[0], crds[1]);
            Point2D end = new Point2D.Double(crds[4], crds[5]);
            Arc2D arc = _calculateArc2D(start, center, end, new Arc2D.Double());
            double sweepAngle = - Math.toRadians(arc.getAngleExtent());
            AffineTransform trf = AffineTransform.getRotateInstance(
                .25 * sweepAngle, arc.getCenterX(), arc.getCenterY());
            
            // Left part
            Point2D arcPnt = new Point2D.Double(crds[0], crds[1]);
            trf.transform(arcPnt, arcPnt);
            crds[2] = arcPnt.getX();
            crds[3] = arcPnt.getY();
            crds[4] = center.getX();
            crds[5] = center.getY();
            seg.setSegment(startVtx, newVtx, crds);

            // Right part
            arcPnt.setLocation(center);
            trf.transform(arcPnt, arcPnt);
            crds[0] = center.getX();
            crds[1] = center.getY();
            crds[2] = arcPnt.getX();
            crds[3] = arcPnt.getY();
            crds[4] = endVtx.getPoint().getX();
            crds[5] = endVtx.getPoint().getY();
            newSeg.setSegment(newVtx, endVtx, crds);

            return m_prim2name.get(newSeg);
        }
        default:
            throw new IllegalArgumentException();
        }
    }

    /**
     * Removes a segment leaving a hole. The segment is replaced by a Moveto
     * segment.
     * 
     * @param name the name of the segment
     */
    public void removeSegment(int name) {
        Segment seg = (Segment)m_name2prim.get(name);

        int pathIndex = m_path.indexOf(seg);
        Vertex startVtx = (Vertex)m_path.get(pathIndex - 1);
        Vertex endVtx = (Vertex)m_path.get(pathIndex + 1);
        double[] crds = new double[4];
        crds[0] = startVtx.getPoint().getX();
        crds[1] = startVtx.getPoint().getY();
        crds[2] = endVtx.getPoint().getX();
        crds[3] = endVtx.getPoint().getY();
        setSegment(name, SEG_MOVETO, crds);

        // Clean the path
        if (m_autoCleaning)
            clean();
    }
    
    // =========================================================================
    // Segment points

    /**
     * Returns the number of segment points (excluding the vertices)
     * 
     * @param segName
     *            the segment name
     * @return the number of points
     */
    public int getSegmentPointCount(int segName) {
        Segment seg = (Segment) m_name2prim.get(segName);
        return seg.getPointCount();
    }

    /**
     * Returns the name of a segment point at a given index.
     * 
     * @param segName
     *            the segment name
     * @param index
     *            the index of the point
     * @return the name of the point
     */
    public int getSegmentPointName(int segName, int index) {
        Segment seg = (Segment) m_name2prim.get(segName);
        return seg.getPointName(index);
    }

    /**
     * Gets the location of a segment point.
     * 
     * @param pntName
     *            the name of the point
     * @param dst
     *            the object to store the location or null
     * @return the location
     */
    public Point2D getSegmentPointLocation(int pntName, Point2D dst) {
        if (dst == null)
            dst = new Point2D.Double();
        dst.setLocation(m_pnts.get(pntName));
        return dst;
    }

    /**
     * Sets the location of a segment point.
     * 
     * @param pntName
     *            the name of the point
     * @param p
     *            the location
     */
    public void setSegmentPointLocation(int pntName, Point2D p) {
        Point2D pnt = m_pnts.get(pntName);
        pnt.setLocation(p);
    }

    /**
     * Checks whether a point of a given name is used by any segment.
     * 
     * @param pntName the name of the point to be searched for
     * @return true if the point is used by any of the segments, false otherwise 
     */
    public boolean containsSegmentPoint(int pntName) {
        int nSegs = getSegmentCount();
        for (int iSeg = 0; iSeg < nSegs; iSeg++) {
            int segName = getSegmentName(iSeg);
            int nPnts = getSegmentPointCount(segName);
            for (int iPnt = 0; iPnt < nPnts; iPnt++) {
                if (pntName == getSegmentPointName(segName, iPnt))
                    return true;
            }
        }
        return false;
    }
    
    /**
     * Gets the segment of a given segment point.
     * 
     * @param pntName the name of the segment point
     * @return the name of the segment or -1 if not found
     */
    public int getSegmentOfSegmentPoint(int pntName) {
        int nSegs = getSegmentCount();
        for (int iSeg = 0; iSeg < nSegs; iSeg++) {
            int segName = getSegmentName(iSeg);
            int nPnts = getSegmentPointCount(segName);
            for (int iPnt = 0; iPnt < nPnts; iPnt++) {
                if (pntName == getSegmentPointName(segName, iPnt))
                    return segName;
            }
        }
        return -1;
    }
    
    // =========================================================================
    // Standard protocol

    public void transformBy(AffineTransform at) {
        Set<Integer> keySet = m_pnts.keySet();
        Iterator<Integer> it = keySet.iterator();
        while (it.hasNext()) {
            int i = it.next();
            Point2D p = m_pnts.get(i);
            Point2D pt = new Point2D.Double();
            at.transform(p, pt);
            m_pnts.put(i, pt);
        }
    }

    // Serialization
    private void writeObject(ObjectOutputStream out) throws IOException {
        // Write non-transient attributes
        out.defaultWriteObject();

        // Write points
        int size = m_pnts.size();
        out.writeInt(size);
        Set<Integer> keySet = m_pnts.keySet();
        Iterator<Integer> it = keySet.iterator();
        while (it.hasNext()) {
            int name = it.next();
            out.writeInt(name);
            Point2D p = m_pnts.get(name);
            out.writeDouble(p.getX());
            out.writeDouble(p.getY());
        }
    }

    private void readObject(ObjectInputStream in) throws IOException,
    ClassNotFoundException {
        // Read non-transient attributes
        in.defaultReadObject();
                
        // Read points
        m_pnts = new HashMap<Integer, Point2D>();
        int size = in.readInt();
        for (int i = 0; i < size; i++) {
            int name = in.readInt();
            Point2D p = new Point2D.Double(in.readDouble(), in.readDouble());
            m_pnts.put(name, p);
        }
        
        // Setup primitive name maps
        m_name2prim = new HashMap<Integer, Primitive>();
        m_prim2name = new HashMap<Primitive, Integer>();
        for (int i = 0; i < m_path.size(); i++) {
            Primitive obj = m_path.get(i);
            int name = obj.getPrimitiveName();
            if (m_subVersion == 0) {
               name = (int)m_num++;
               obj.setPrimitiveName(name);
            }
            m_name2prim.put(name, obj);
            m_prim2name.put(obj, name);
        }

        // Schema evolution
        int n = m_path.size();
        for (int i = 1; i < n; i += 2) {
            Segment seg = (Segment)m_path.get(i);
            seg.evolve((Vertex)m_path.get(i - 1), (Vertex)m_path.get(i + 1));
        }
        m_subVersion = 1;
    }

    protected void _dump() {
        System.out.println("m_path");
        Iterator<Primitive> it = m_path.iterator();
        while (it.hasNext()) {
            System.out.println("  " + it.next());
        }
        
        System.out.println("m_pnts");
        Set<Integer> set = m_pnts.keySet();
        Iterator<Integer> iit = set.iterator();
        while (iit.hasNext()) {
            int key = iit.next();
            Point2D pnt = m_pnts.get(key);
            System.out.println("  " + key + " -> " + pnt);
        }
        
        System.out.println("m_name2prim");
        set = m_name2prim.keySet();
        iit = set.iterator();
        while (iit.hasNext()) {
            int key = iit.next();
            Object obj = m_name2prim.get(key);
            System.out.println("  " + key + " -> " + obj);
        }

        System.out.println("m_prim2name");
        Set<Primitive> objSet = m_prim2name.keySet();
        Iterator<Primitive> oit = objSet.iterator();
        while (oit.hasNext()) {
            Primitive key = oit.next();
            int val = m_prim2name.get(key);
            System.out.println("  " + val + " <- " + key);
        }

        System.out.println("m_num = " + m_num);
    }
    
    // Cloning
    public Object clone() {
        
        Path2D clone = new Path2D();

        // Clone points
        clone.m_pnts = new HashMap<Integer, Point2D>();
        Iterator<Integer> kit = m_pnts.keySet().iterator();
        while (kit.hasNext()) {
            int key = kit.next();
            clone.m_pnts.put(key, (Point2D) m_pnts.get(key).clone());
        }

        // Clone path
        clone.m_path = new ArrayList<Primitive>();
        for (int i = 0; i < m_path.size(); i++) {
            Object el = m_path.get(i);
            if (el instanceof Vertex) {
                Vertex vtx = (Vertex)el;
                clone.m_path.add(clone.new Vertex((int)vtx.m_name));
            }
            else if (el instanceof Segment) {
                Segment s_org = (Segment) el;
                switch (s_org.getType()) {
                case SEG_MOVETO:
                    clone.m_path.add(clone.new MoveSegment());
                    break;
                case SEG_LINETO:
                    clone.m_path.add(clone.new LineSegment());
                    break;
                case SEG_QUADTO:
                    clone.m_path.add(clone.new QuadSegment(
                        ((QuadSegment)el).getCtrlPtName()));
                    break;
                case SEG_CUBICTO:
                    clone.m_path.add(clone.new CubicSegment(
                        ((CubicSegment)el).getCtrlP1Name(),
                        ((CubicSegment)el).getCtrlP2Name()));
                    break;
                case SEG_ARCTO:
                    clone.m_path.add(clone.new ArcSegment(
                        ((ArcSegment)el).getArcPtName()));
                    break;
                case SEG_CLOSE:
                    clone.m_path.add(clone.new CloseSegment());
                    break;
                default:
                    throw new IllegalPathStateException();
                }
            }
            else throw new IllegalStateException();
        }
        
        // Clone num
        clone.m_num = m_num;
        
        // Setup primitive name maps
        clone.m_name2prim.clear();
        clone.m_prim2name.clear();
        for (int i = 0; i < clone.m_path.size(); i++) {
            Object src = m_path.get(i);
            int name = m_prim2name.get(src);
            Primitive dst = clone.m_path.get(i);
            clone.m_name2prim.put(name, dst);
            clone.m_prim2name.put(dst, name);
        }

        return clone;
    }
    
    public Path2DIterator getIterator() {
        return this.new Path2DIterator();
    }

    public String toString() {
        String string = new String("Path2D:\n  " +
            new MoveSegment().toString()+"\n");
        ListIterator<Primitive> lit = m_path.listIterator();
        while (lit.hasNext()) {
            string += "  " + lit.next().toString() + "\n";
        }
        return string;
    }

    // Helpers
    private interface Primitive extends Serializable {
        /**
         * Destructs this object and frees resources.
         * 
         */
        public void destruct();
        
        /**
         * Sets the persistent Primitive name.
         */
        public void setPrimitiveName(int name);
        
        /**
         * Returns the persistent Primitive name.
         */
        public int getPrimitiveName();
    }
    private class Vertex implements Primitive {
        private static final long serialVersionUID = 0L;
        private long m_name;    // Point name or -1 for end vertex of close seg
        private int m_vertexName;
        public Vertex(int name) {
            if (name < -1)
                throw new IllegalArgumentException();
            m_name = name;
            m_vertexName = (int)m_num;
            m_name2prim.put((int)m_num, this);
            m_prim2name.put(this, (int)m_num++);
        }
        public Point2D getPoint() {
            Point2D pnt = m_pnts.get(getPointName());
            return pnt;
        }
        public void setPoint(Point2D pnt) {
            setPoint(pnt.getX(), pnt.getY());
        }
        public void setPoint(double x, double y) {
            if (! isCloseVertex()) {
                Point2D p = m_pnts.get(getPointName());
                p.setLocation(x, y);
            }
        }
        public void setPointName(int name) {
            if (name < -1)
                throw new IllegalArgumentException();
            m_name = name;
        }
        public int getPointName() {
            if (isCloseVertex()) {
                Vertex vtx = _startOfLoop(this);
                return (int)vtx.m_name;
            }
            return (int)m_name;
        }
        public void setPrimitiveName(int name) {
            m_vertexName = name;
        }
        public int getPrimitiveName() {
            return m_vertexName;
        }
        public void destruct() {
            if (! isCloseVertex())
                m_pnts.remove(getPointName());
            m_name = 0;
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        public boolean isCloseVertex() {
            return m_name == -1;
        }
        public void toCloseVertex(boolean close) {
            if (close) {
                if (m_name != -1) {
                    m_pnts.remove(m_name);
                    m_name = -1;
                }
            }
            else {
                if (m_name == -1) {
                    Point2D p = getPoint();
                    if (p == null)
                        p = new Point2D.Double();
                    m_name = addPoint(p);
                }
            }
        }
        public String toString() {
            Point2D pnt = getPoint();
            if (pnt == null) {
                return (isCloseVertex() ? "Close Vertex:[" : "Vertex:[") +
                    "null]";
            }
            return (isCloseVertex() ? "Close:[" : "Vertex:[") +
                    Geo2D.formatCoordinate(pnt.getX()) + ", " +
                    Geo2D.formatCoordinate(pnt.getY()) + "]";
        }
    }

    public interface Segment extends Primitive {
    
        /**
         * Returns the Segment type.
         * 
         * @return the Segment type, e.g. one of MOVETO, LINETO, QUADTO,
         *      CUBICTO, ARCTO or CLOSE
         */
        public int getType();

        /**
         * Returns the Segment type and its cordinates.
         * 
         * @param start the start vertex
         * @param end the end vertex
         * @param crds a double array holding the coordinates. The array must
         *      be capable to hold 8 values 
         * @return the Segment type, e.g. one of MOVETO, LINETO, QUADTO,
         *      CUBICTO, ARCTO or CLOSE
         */
        public int getSegment(Vertex start, Vertex end, double[] crds);

        /**
         * Sets the Segment coordinates.
         * 
         * @param start the start vertex
         * @param end the end vertex
         * @param crds a double array holding the coordinates. The array must
         *      be capable to hold 8 values 
         */
        public void setSegment(Vertex start, Vertex end, double[] crds);

        /**
         * Gets the number of internal points.
         * 
         * @return the number of points
         */
        public int getPointCount();

        /**
         * Gets the name of the internal point at a given index.
         * 
         * @param index the index of the internal point
         * @return the name f the internal point
         */
        public int getPointName(int index);

        /**
         * Calculates a point on the segment. Note that the point must be 
         * located on the segment curve, but not necessarily at the center.
         * 
         * @param start the start Vertex
         * @param end the end Vertex
         * @param dst the point coordinates to store the point or null
         * @return the coordinates of the point on the segment
         */
        public Point2D getPointOnSegment(Vertex start, Vertex end, Point2D dst);
        
        /**
         * Calculates the location to represent the segment. Note that the
         * location must not necessarily be on the segment curve.
         * 
         * @param start the start Vertex
         * @param end the end Vertex
         * @param dst the point coordinates to store the point or null
         * @return the coordinates of the point representing the segment
         */
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst);
        
        /**
         * Calculates the Segment's Shape.
         * 
         * @param start the start Vertex
         * @param end the end Vertex
         * @return the Shape object
         */
        public Shape getShape(Vertex start, Vertex end);

        /**
         * Calculates the Segment's length.
         * 
         * @param start the start Vertex
         * @param end the end Vertex
         * @return the segment length
         */
        public double length(Vertex start, Vertex end);

        /**
         * Calculates the area by integrating the segment curve.
         * 
         * @param start the start Vertex
         * @param end the end Vertex
         * @return the signed segment area
         */
        public double area(Vertex start, Vertex end);

        /**
         * Evolves the schema
         * 
         */
        public void evolve(Vertex sp, Vertex ep);
    }

    private class MoveSegment implements Segment {
        private static final long serialVersionUID = 0L;
        private int m_segmentName;
        
        public MoveSegment() {
            this((int)m_num++);
        }
        public MoveSegment(int segName) {
            m_segmentName = segName;
            m_name2prim.put(segName, this);
            m_prim2name.put(this, segName);
        }
        public int getType() {
            return SEG_MOVETO;
        }
        public int getSegment(Vertex start, Vertex end, double[] crds) {
            Point2D p = start.getPoint();
            crds[0] = p.getX();
            crds[1] = p.getY();
            p = end.getPoint();
            crds[2] = p.getX();
            crds[3] = p.getY();
            return SEG_MOVETO;
        }
        public void setSegment(Vertex start, Vertex end, double[] crds) {
            start.toCloseVertex(false);
            end.toCloseVertex(false);
            start.setPoint(crds[0], crds[1]);
            end.setPoint(crds[2], crds[3]);
        }
        public int getPointCount() {
            return 0;
        }
        public int getPointName(int index) {
            throw new IllegalArgumentException();
        }
        public Point2D getPointOnSegment(Vertex start, Vertex end,
        Point2D dst) {
            return null;
        }
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst) {
            return getPointOnSegment(start, end, dst);
        }
        public Shape getShape(Vertex start, Vertex end) {
            return null;
        }
        public double length(Vertex start, Vertex end) {
            return 0;
        }
        public double area(Vertex start, Vertex end) {
            return 0;
        }
        public void destruct() {
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        public void setPrimitiveName(int name) {
            m_segmentName = name;
        }
        public int getPrimitiveName() {
            return m_segmentName;
        }
        public void evolve(Vertex sp, Vertex ep) {
        }
        public String toString() {
            return "Segment.MOVETO";
        }
    }

    private class LineSegment implements Segment {
        private static final long serialVersionUID = 0L;
        private int m_segmentName;
        
        public LineSegment() {
            this((int)m_num++);
        }
        public LineSegment(int segName) {
            m_segmentName = segName;
            m_name2prim.put(segName, this);
            m_prim2name.put(this, segName);
        }
        public int getType() {
            return SEG_LINETO;
        }
        public int getSegment(Vertex start, Vertex end, double[] crds) {
            Point2D p = start.getPoint();
            crds[0] = p.getX();
            crds[1] = p.getY();
            p = end.getPoint();
            crds[2] = p.getX();
            crds[3] = p.getY();
            return SEG_LINETO;
        }
        public void setSegment(Vertex start, Vertex end, double[] crds) {
            start.toCloseVertex(false);
            end.toCloseVertex(false);
            start.setPoint(crds[0], crds[1]);
            end.setPoint(crds[2], crds[3]);
        }
        public int getPointCount() {
            return 0;
        }
        public int getPointName(int index) {
            throw new IllegalArgumentException();
        }
        public Point2D getPointOnSegment(Vertex start, Vertex end,
        Point2D dst) {
            if (dst == null)
                dst = new Point2D.Double();
            Point2D sp = start.getPoint();
            Point2D ep = end.getPoint();
            dst.setLocation(0.5 * (sp.getX() + ep.getX()),
                0.5 * (sp.getY() + ep.getY()));
            return dst;
        }
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst) {
            return getPointOnSegment(start, end, dst);
        }
        public Shape getShape(Vertex start, Vertex end) {
            return new Line2D.Double(start.getPoint(), end.getPoint());
        }
        public double length(Vertex start, Vertex end) {
            return Geo2D.distance(start.getPoint(), end.getPoint());
        }
        public double area(Vertex start, Vertex end) {
            return Geo2D.area(new Line2D.Double(start.getPoint(), 
                    end.getPoint()));
        }
        public void setPrimitiveName(int name) {
            m_segmentName = name;
        }
        public int getPrimitiveName() {
            return m_segmentName;
        }
        public void destruct() {
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        public void evolve(Vertex sp, Vertex ep) {
        }
        public String toString() {
            return "Segment.LINETO";
        }
    }

    private class QuadSegment implements Segment {
        private static final long serialVersionUID = 0L;
        private long m_cp;
        private int m_segmentName;
        
        public QuadSegment(int name) {
            this((int)m_num++, name);
        }
        public QuadSegment(int segName, int name) {
            m_cp = name;
            m_segmentName = segName;
            m_name2prim.put(segName, this);
            m_prim2name.put(this, segName);
        }
        public int getType() {
            return SEG_QUADTO;
        }
        public int getCtrlPtName() {
            return (int)m_cp;
        }
        public int getSegment(Vertex start, Vertex end, double[] crds) {
            Point2D p = start.getPoint();
            crds[0] = p.getX();
            crds[1] = p.getY();
            p = m_pnts.get(getCtrlPtName());
            crds[2] = p.getX();
            crds[3] = p.getY();
            p = end.getPoint();
            crds[4] = p.getX();
            crds[5] = p.getY();
            return SEG_QUADTO;
        }
        public void setSegment(Vertex start, Vertex end, double[] crds) {
            start.toCloseVertex(false);
            end.toCloseVertex(false);
            start.setPoint(crds[0], crds[1]);
            Point2D p = m_pnts.get(getCtrlPtName());
            p.setLocation(crds[2], crds[3]);
            end.setPoint(crds[4], crds[5]);
        }
        public int getPointCount() {
            return 1;
        }
        public int getPointName(int index) {
            return getCtrlPtName();
        }
        public Point2D getPointOnSegment(Vertex start, Vertex end,
        Point2D dst) {
            if (dst == null)
                dst = new Point2D.Double();

            double[] crds = new double[16];
            getSegment(start, end, crds);
            QuadCurve2D.subdivide(crds, 0, crds, 6, crds, 10);

            dst.setLocation(crds[10], crds[11]);
            return dst;
        }
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst) {
            return getPointOnSegment(start, end, dst);
        }
        public Shape getShape(Vertex start, Vertex end) {
            Point2D p = m_pnts.get(getCtrlPtName());
            return new QuadCurve2D.Double(start.getPoint().getX(), 
                    start.getPoint().getY(), p.getX(), p.getY(), 
                    end.getPoint().getX(), end.getPoint().getY());
        }
        public double length(Vertex start, Vertex end) {
            return Geo2D.length((QuadCurve2D)getShape(start, end));
        }
        public double area(Vertex start, Vertex end) {
            return Geo2D.area((QuadCurve2D)getShape(start, end));
        }
        public void setPrimitiveName(int name) {
            m_segmentName = name;
        }
        public int getPrimitiveName() {
            return m_segmentName;
        }
        public void destruct() {
            m_pnts.remove(getCtrlPtName());
            m_cp = 0;
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        public void evolve(Vertex sp, Vertex ep) {
        }
        public String toString() {
            Point2D p = m_pnts.get(getCtrlPtName());
            return "Segment.QUADTO:[" + Geo2D.formatCoordinate(p.getX()) +
                    ", " + Geo2D.formatCoordinate(p.getY()) + "]";
        }
    }

    private class CubicSegment implements Segment {
        private static final long serialVersionUID = 0L;
        private int m_segmentName;
        private long m_cp1;
        private long m_cp2;

        public CubicSegment(int name1, int name2) {
            this((int)m_num++, name1, name2);
        }
        public CubicSegment(int segName, int name1, int name2) {
            m_cp1 = name1;
            m_cp2 = name2;
            m_segmentName = segName;
            m_name2prim.put(segName, this);
            m_prim2name.put(this, segName);
        }
        public int getType() {
            return SEG_CUBICTO;
        }
        public int getCtrlP1Name() {
            return (int)m_cp1;
        }
        public int getCtrlP2Name() {
            return (int)m_cp2;
        }
        public int getSegment(Vertex start, Vertex end, double[] crds) {
            Point2D p = start.getPoint();
            crds[0] = p.getX();
            crds[1] = p.getY();
            p = m_pnts.get(getCtrlP1Name());
            crds[2] = p.getX();
            crds[3] = p.getY();
            p = m_pnts.get(getCtrlP2Name());
            crds[4] = p.getX();
            crds[5] = p.getY();
            p = end.getPoint();
            crds[6] = p.getX();
            crds[7] = p.getY();
            return SEG_CUBICTO;
        }
        public void setSegment(Vertex start, Vertex end, double[] crds) {
            start.toCloseVertex(false);
            end.toCloseVertex(false);
            start.setPoint(crds[0], crds[1]);
            Point2D p = m_pnts.get(getCtrlP1Name());
            p.setLocation(crds[2], crds[3]);
            p = m_pnts.get(getCtrlP2Name());
            p.setLocation(crds[4], crds[5]);
            end.setPoint(crds[6], crds[7]);
        }
        public int getPointCount() {
            return 2;
        }
        public int getPointName(int index) {
            switch (index) {
            case 0:
                return getCtrlP1Name();
            case 1:
                return getCtrlP2Name();
            }
            throw new IllegalArgumentException();
        }
        public Point2D getPointOnSegment(Vertex start, Vertex end,
        Point2D dst) {
            if (dst == null)
                dst = new Point2D.Double();

            double[] crds = new double[24];
            getSegment(start, end, crds);
            CubicCurve2D.subdivide(crds, 0, crds, 8, crds, 16);

            dst.setLocation(crds[14], crds[15]);
            return dst;
        }
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst) {
            return getPointOnSegment(start, end, dst);
        }
        public Shape getShape(Vertex start, Vertex end) {
            Point2D cp1 = m_pnts.get(getCtrlP1Name());
            Point2D cp2 = m_pnts.get(getCtrlP2Name());
            return new CubicCurve2D.Double(start.getPoint().getX(), start
                .getPoint().getY(), cp1.getX(), cp1.getY(), cp2.getX(), cp2
                .getY(), end.getPoint().getX(), end.getPoint().getY());
        }
        public double length(Vertex start, Vertex end) {
            return Geo2D.length((CubicCurve2D)getShape(start, end));
        }
        public double area(Vertex start, Vertex end) {
            return Geo2D.area((CubicCurve2D)getShape(start, end));
        }
        public void setPrimitiveName(int name) {
            m_segmentName = name;
        }
        public int getPrimitiveName() {
            return m_segmentName;
        }
        public void destruct() {
            m_pnts.remove(getCtrlP1Name());
            m_cp1 = 0;
            m_pnts.remove(getCtrlP2Name());
            m_cp2 = 0;
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        public void evolve(Vertex sp, Vertex ep) {
        }
        public String toString() {
            Point2D cp1 = m_pnts.get(getCtrlP1Name());
            Point2D cp2 = m_pnts.get(getCtrlP2Name());
            return "Segment.CURVETO:[" + Geo2D.formatCoordinate(cp1.getX()) +
                    ", " + Geo2D.formatCoordinate(cp1.getY()) + "] [" +
                    Geo2D.formatCoordinate(cp2.getX()) + ", " +
                    Geo2D.formatCoordinate(cp2.getY()) + "]";
        }
    }

    private class ArcSegment implements Segment {
        private static final long serialVersionUID = 0L;
        private int m_subVersion = 1;
        private int m_segmentName;
        private long m_cp;      // Point on the arc (was Center point)
        
        public ArcSegment(int arcPoint) {
            this((int)m_num++, arcPoint);
        }
        public ArcSegment(int segName, int arcPoint) {
            m_cp = arcPoint;
            m_segmentName = segName;
            m_name2prim.put(segName, this);
            m_prim2name.put(this, segName);
        }
        public int getType() {
            return SEG_ARCTO;
        }
        public int getArcPtName() {
            return (int)m_cp;
        }
        public int getSegment(Vertex start, Vertex end, double[] crds) {
            Point2D p = start.getPoint();
            crds[0] = p.getX();
            crds[1] = p.getY();
            p = m_pnts.get(getArcPtName());
            crds[2] = p.getX();
            crds[3] = p.getY();
            p = end.getPoint();
            crds[4] = p.getX();
            crds[5] = p.getY();
            return SEG_ARCTO;
        }
        public void setSegment(Vertex start, Vertex end, double[] crds) {
            start.toCloseVertex(false);
            end.toCloseVertex(false);
            start.setPoint(crds[0], crds[1]);
            Point2D pArc = m_pnts.get(getArcPtName());
            pArc.setLocation(crds[2], crds[3]);
            end.setPoint(crds[4], crds[5]);
        }
        public int getPointCount() {
            return 1;
        }
        public int getPointName(int index) {
            if (index == 0)
                return getArcPtName();
            throw new IllegalArgumentException();
        }
        public Point2D getPointOnSegment(Vertex start, Vertex end,
        Point2D dst) {
            if (dst == null)
                dst = new Point2D.Double();

            Point2D sp = start.getPoint();
            Point2D ep = end.getPoint();
            Point2D p = m_pnts.get(getArcPtName());
            Arc2D arc = _calculateArc2D(sp, p, ep, new Arc2D.Double());
            if (arc == null) {
                dst.setLocation(
                   .5 * (sp.getX() + ep.getX()),
                   .5 * (sp.getY() + ep.getY()));
                return dst;
            }
            
            double alfaStart = -arc.getAngleStart();
            double alfaSweep = -arc.getAngleExtent();
            double angle = Math.toRadians(alfaStart + 0.5 * alfaSweep);
            double centerX = arc.getCenterX();
            double centerY = arc.getCenterY();
            double r = Geo2D.distance(sp, new Point2D.Double(centerX, centerY));
            double x = centerX + r * Math.cos(angle);
            double y = centerY + r * Math.sin(angle);
            dst.setLocation(x, y);
            
            return dst;
        }
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst) {
            if (dst == null)
                dst = new Point2D.Double();

            Point2D sp = start.getPoint();
            Point2D ep = end.getPoint();
            Ellipse2D ellipse = _calculateEllipse2D(sp, 
                    m_pnts.get(getArcPtName()), ep, new Ellipse2D.Double());
            if (ellipse == null)
                dst.setLocation(.5 * (sp.getX() + ep.getX()),
                        .5 * (sp.getY() + ep.getY()));
            else
                dst.setLocation(ellipse.getCenterX(), ellipse.getCenterY());
            return dst;
        }
        public Shape getShape(Vertex start, Vertex end) {
            Point2D sp = start.getPoint();
            Point2D p = m_pnts.get(getArcPtName());
            Point2D ep = end.getPoint();

            // Calculate arc
            Arc2D arc = _calculateArc2D(sp, p, ep, new Arc2D.Double());

            // Return arc or line if degenerated
            return arc == null ? new Line2D.Double(sp, ep) : arc;
        }
        public double length(Vertex start, Vertex end) {
            Shape shape = getShape(start, end);
            if (shape instanceof Arc2D) {
                Arc2D arc = (Arc2D)shape;
                double alfaSweep = Math.abs(arc.getAngleExtent());
                double r = Geo2D.distance(start.getPoint(), 
                        new Point2D.Double(arc.getCenterX(), arc.getCenterY()));
                return Math.toRadians(alfaSweep) * r;
            }
            else {
                Line2D line = (Line2D)shape;
                return Geo2D.length(line);
            }
        }
        public double area(Vertex start, Vertex end) {
            Shape shape = getShape(start, end);
            return shape instanceof Arc2D ? Geo2D.area((Arc2D)shape) :
                Geo2D.area((Line2D)shape);
        }
        public void setPrimitiveName(int name) {
            m_segmentName = name;
        }
        public int getPrimitiveName() {
            return m_segmentName;
        }
        public void destruct() {
            m_pnts.remove(getArcPtName());
            m_cp = 0;
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        // Former attributes: version 0
        // private long m_sp;       // Start point name (dummy)
        private double m_alpha;     // Sweep angle
        public void evolve(Vertex start, Vertex end) {
            if (m_subVersion == 0) {
                m_subVersion = 1;

                // Calculate a point on the arc
                Point2D startPoint = start.getPoint();
                Point2D center = m_pnts.get((int)m_cp);
                double radius = Geo2D.distance(startPoint, center);
                Vector2D vcs = new Vector2D(center, startPoint);
                double angle = Vector2D.X_UNIT.getAngleCCW(vcs);
                angle += .5 * m_alpha;
                Vector2D vArcPoint = new Vector2D(angle).scaleBy(radius);

                // Change the coordinates of the former center point
                center.setLocation(center.getX() + vArcPoint.x,
                    center.getY() + vArcPoint.y);
            }
        }
        public String toString() {
            Point2D cp = m_pnts.get(getArcPtName());
            return "Segment.ARCTO:[" + Geo2D.formatCoordinate(cp.getX()) +
                    ", " + Geo2D.formatCoordinate(cp.getY()) + "] ";
        }
    }
    
    private class CloseSegment implements Segment {
        private static final long serialVersionUID = 0L;
        private int m_segmentName;
        
        public CloseSegment() {
            this((int)m_num++);
        }
        public CloseSegment(int segName) {
            m_segmentName = segName;
            m_name2prim.put(segName, this);
            m_prim2name.put(this, segName);
        }
        public int getType() {
            return SEG_CLOSE;
        }
        public int getSegment(Vertex start, Vertex end, double[] crds) {
            return SEG_CLOSE;
        }
        public void setSegment(Vertex start, Vertex end, double[] crds) {
            start.toCloseVertex(false);
            end.toCloseVertex(true);
        }
        public int getPointCount() {
            return 0;
        }
        public int getPointName(int index) {
            throw new IllegalArgumentException();
        }
        public Point2D getPointOnSegment(Vertex start, Vertex end,
        Point2D dst) {
            if (dst == null)
                dst = new Point2D.Double();
            Point2D sp = start.getPoint();
            Point2D ep = end.getPoint();
            dst.setLocation(0.5 * (sp.getX() + ep.getX()),
                0.5 * (sp.getY() + ep.getY()));
            return dst;
        }
        public Point2D getRepresentativeLocation(Vertex start, Vertex end,
                Point2D dst) {
            return getPointOnSegment(start, end, dst);
        }
        public Shape getShape(Vertex start, Vertex end) {
            return new Line2D.Double(start.getPoint(), end.getPoint());
        }
        public double length(Vertex start, Vertex end) {
            return Geo2D.distance(start.getPoint(), end.getPoint());
        }
        public double area(Vertex start, Vertex end) {
            return Geo2D.area(new Line2D.Double(start.getPoint(), 
                    end.getPoint()));
        }
        public void setPrimitiveName(int name) {
            m_segmentName = name;
        }
        public int getPrimitiveName() {
            return m_segmentName;
        }
        public void destruct() {
            int name = m_prim2name.get(this);
            m_name2prim.remove(name);
            m_prim2name.remove(this);
        }
        public void evolve(Vertex sp, Vertex ep) {
        }
        public String toString() {
            return "Segment.CLOSE";
        }
    }
    
    public class Path2DIterator {
        private int pos = -1;
        private boolean done = false;
        public void next() {
            pos += 2;
            if (m_path.size() <= pos) done = true;
        }
        public boolean isDone() {
            return done;
        }
        public int currentSegment(double[] crds) {
            if (crds == null)
                crds = new double[6];
            if (pos == -1) {
                Point2D pt = ((Vertex)m_path.get(0)).getPoint();
                crds[0] = pt.getX();
                crds[1] = pt.getY();
                return SEG_MOVETO;
            }
            Segment cs = ((Segment)m_path.get(pos));
            Point2D pt = ((Vertex)m_path.get(pos+1)).getPoint();
            switch (cs.getType()) {
            case SEG_MOVETO:
                crds[0] = pt.getX();
                crds[1] = pt.getY();
                return SEG_MOVETO;
            case SEG_LINETO:
                crds[0] = pt.getX();
                crds[1] = pt.getY();
                return SEG_LINETO;
            case SEG_QUADTO:
                int name = ((QuadSegment)cs).getCtrlPtName();
                Point2D cp = m_pnts.get(name);
                crds[0] = cp.getX();
                crds[1] = cp.getY();
                crds[2] = pt.getX();
                crds[3] = pt.getY();
                return SEG_QUADTO;
            case SEG_CUBICTO:
                int name1 = ((CubicSegment)cs).getCtrlP1Name();
                int name2 = ((CubicSegment)cs).getCtrlP2Name();
                Point2D cp1 = m_pnts.get(name1);
                Point2D cp2 = m_pnts.get(name2);
                crds[0] = cp1.getX();
                crds[1] = cp1.getY();
                crds[2] = cp2.getX();
                crds[3] = cp2.getY();
                crds[4] = pt.getX();
                crds[5] = pt.getY();
                return SEG_CUBICTO;
            case SEG_ARCTO:
                int name3 = ((ArcSegment)cs).getArcPtName();
                Point2D cp3 = m_pnts.get(name3);
                crds[0] = cp3.getX();
                crds[1] = cp3.getY();
                crds[2] = pt.getX();
                crds[3] = pt.getY();
                return SEG_ARCTO;
            case SEG_CLOSE:
                return SEG_CLOSE;
            default:
                throw new IllegalPathStateException();
            }
        }
        public Point2D getSegmentPoint(Point2D dst) {
            if (pos == -1)
                return null;
            if (dst == null)
                dst = new Point2D.Double();
            double[] crds = new double[6];
            int segType = currentSegment(crds);
            Point2D sp = ((Vertex)m_path.get(pos-1)).getPoint();
            switch (segType) {
            case SEG_MOVETO:
                return null;
            case SEG_LINETO:
                dst.setLocation(0.5 * (sp.getX()+crds[0]),
                    0.5 * (sp.getY()+crds[1]));
                return dst;
            case SEG_QUADTO:
                dst.setLocation(0.5 * (sp.getX()+crds[2]),
                    0.5 * (sp.getY()+crds[3]));
                return dst;
            case SEG_CUBICTO:
                dst.setLocation(0.5 * (sp.getX()+crds[4]),
                    0.5 * (sp.getY()+crds[5]));
                return dst;
            case SEG_ARCTO:
                dst.setLocation(0.5 * (sp.getX()+crds[0]),
                    0.5 * (sp.getY()+crds[1]));
                return dst;
            case SEG_CLOSE:
                return null;
            default:
                throw new IllegalPathStateException();
            }
        }
    }
}