/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or write
 * to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.geo;

import java.awt.geom.PathIterator;
import java.util.ArrayList;
import java.util.List;

/** 
 * A <code>RewindablePathIterator</code> extends the <code>PathIterator</code> 
 * by the possibility of multiple iterations.
 * <p>
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.5 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class RewindablePathIterator implements PathIterator {

    private static class _Segment {
        private int m_seg;
        private double[] m_crds;
        private _Segment(int seg, double[] crds) {
            m_seg = seg;
            m_crds = crds;
        }
    }
    private List<_Segment> m_segments = new ArrayList<_Segment>(); // List of _Seg objects
    private int m_currentIndex = 0;
    private int m_windingRule;
    
    // Constructor
    public RewindablePathIterator(PathIterator pit) {

        // Store the winding rule
        m_windingRule = pit.getWindingRule();
        
        // Store the sequence of segments
        double[] crds = new double[6];
        for (; ! pit.isDone(); pit.next()) {
            int type = pit.currentSegment(crds);
            switch (type) {
            case PathIterator.SEG_MOVETO :
            case PathIterator.SEG_LINETO : {
                double[] _crds = new double[2];
                _crds[0] = crds[0];
                _crds[1] = crds[1];
                m_segments.add(new _Segment(type, _crds));
            }
            break;
            case PathIterator.SEG_QUADTO : {
                double[] _crds = new double[4];
                for (int i = 0; i < 4; i++)
                    _crds[i] = crds[i];
                m_segments.add(new _Segment(type, _crds));
            }
            break;
            case PathIterator.SEG_CUBICTO : {
                double[] _crds = new double[6];
                for (int i = 0; i < 6; i++)
                    _crds[i] = crds[i];
                m_segments.add(new _Segment(type, _crds));
            }
            break;
            case PathIterator.SEG_CLOSE :
                m_segments.add(new _Segment(type, new double[0]));
            break;
            default :
                throw new Error("Unexpected segment type");
            }
        }
    }
    
    // Rewind
    public void rewind() {
        m_currentIndex = 0;
    }

    // PathIterator methods
    public int currentSegment(double[] coords) {
        _Segment _seg = m_segments.get(m_currentIndex);
        for (int i = 0; i < _seg.m_crds.length; i++)
            coords[i] = _seg.m_crds[i];
        return _seg.m_seg;
    }
    public int currentSegment(float[] coords) {
        _Segment _seg = m_segments.get(m_currentIndex);
        for (int i = 0; i < _seg.m_crds.length; i++)
            coords[i] = (float)_seg.m_crds[i];
        return _seg.m_seg;
    }
    public int getWindingRule() {
        return m_windingRule;
    }
    public boolean isDone() {
        return m_currentIndex >= m_segments.size();
    }
    public void next() {
        if (! isDone())
            m_currentIndex++;
    }
}