/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.msg;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Iterator;
import java.util.Set;

import cib.util.Dbg;

/**
 * The ServerMsgHandler class provides - based upon the functionality of
 * the {@link MsgHandler MsgHandler} class - a high level protocol for
 * communication in TCP/IP networks (e.g. the Internet).
 * <p>
 * The ServerMsgHandler class supports an arbitrary number of Client
 * Connections. Each Client should be represented by a {@link
 * ClientMsgHandler ClientMsgHandler} object.
 * <p>
 * Connections are actually created by the class function
 * {@link #getClientConnection(int port) getClientConnection()} which returns
 * a ServerMsgHandler object on success.
 * <p>
 * In addition to the MsgHandler class unique client IDs and (non-unique)
 * client names (currently the DNS name) can be obtained for the connected
 * ClientMsgHandler object.
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.5 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 *
 * @see Msg
 * @see MsgReactor
 * @see ClientMsgHandler
 * @see MsgHandler
 * @see MsgBrokenConnection
 */
public class ServerMsgHandler extends MsgHandler {

    private static ServerSocket s_serverSocket = null;
    private Socket m_clientSocket = null;

    private static int s_clientId = 0;
    private int m_clientId = 0;  // Not set

   /**
    * Allows a ClientMsgHandler object to connect to a ServerMsgHandler
    * objects. This method blocks until a connection is established.
    * The message listener thread is started automatically.
    * @param port the port number
    * @return the ServerMsgHandler object or null
    */
    public static synchronized ServerMsgHandler getClientConnection(int port) {
        try {
            if (s_serverSocket == null)
                s_serverSocket = new ServerSocket(port);

            Socket clientSocket = s_serverSocket.accept(); // Blocks
            OutputStream out = clientSocket.getOutputStream();
            InputStream in = clientSocket.getInputStream();
            ServerMsgHandler smh = new ServerMsgHandler(clientSocket, in, out);

            // Tell the client its Id
            DataOutputStream dout = new DataOutputStream(out);
            dout.writeInt(smh.getClientId());

            // Start the message handler thread
            smh.start();
            return smh;
        }
        catch (java.io.IOException e) {
            Dbg.println("ServerMsgHandler.getClientConnection(): " + e);
        }
        return null;
    }

    // Construction and destruction
    private ServerMsgHandler(Socket cs, InputStream in,
        OutputStream out)
    {
        super(in, out);
        m_clientSocket = cs;
        m_clientId = ++s_clientId;    // Starts with 1
    }

   /**
    * Closes this mesage handler. Normally there is no need to close a
    * Server Message Handler.
    */
    public void close() {
        super.close();
        if (m_clientSocket != null) {
            try {
                m_clientSocket.close();
                m_clientSocket = null;
            }
            catch (IOException e) {
            }
        }
    }

   /**
    * Returns the client ID that was automatically generated by the server.
    * @return the unique client ID
    */
    public int getClientId() {
        return m_clientId;
    }

   /**
    * Returns the client name (currently as DNS name)
    * @return the non-unique client name
    */
    public String getClientName() {
        InetAddress addr = m_clientSocket.getInetAddress();
        return addr.getHostName();
    }

   /**
    * Convenience method to get the names of all ClientMsgHandlers
    * connected to the ServerMsgHandler objects in this application.
    * @return an array of possibly non-unique client name
    */
    public static String[] getAllClientNames() {
        Set<MsgHandler> clients = getAllMsgHandlers();
//        synchronized (clients) {
        int nClients = clients.size();
        String[] clientNames = new String[nClients];
        Iterator<MsgHandler> it = clients.iterator();
        int iClient = 0;
        while (it.hasNext()) {
            ServerMsgHandler smh = (ServerMsgHandler) it.next();
            clientNames[iClient++] = smh.getClientName();
        }
        return clientNames;
//        }
    }
}