/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.update;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Set;

import cib.util.coll.CollUtilities;
import cib.util.coll.ObservableSet;

/**
 * This class represents an updatable, directed graph.
 *
 * @author Christian Koch
 * @version $Revision: 1.15 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 * @see cib.util.update.UpdatableGraph
 * @see cib.util.update.Graph
 * @see cib.util.update.Update
 * @see cib.util.update.Node
 * @see cib.util.update.UpdateInterface
 */
@Deprecated
public class _UpdatableGraph implements UpdatableGraph, Serializable {

    private static final long serialVersionUID = 0L;

    /**
     * The link graph.
     */
    private Graph m_linkGraph = null;

    /**
     * The adjacency matrix for the link graph.
     */
    private transient boolean[][] m_link = null;

    /**
     * The adjacency matrix for the link graph's hull.
     */
    private transient boolean[][] m_hull = null;

    /**
     * The update interface.
     */
    private UpdateInterface m_updateInterface = null;

    /**
     * The set of link graph's nodes.
     */
    private ObservableSet<Node> m_nodeSet = null;

    /**
     * The set of changed nodes.
     */
    private Set<Node> m_changedSet = null;

    /**
     * The update.
     */
    private transient Update m_update = null;

    /**
     * Constructs an updatable graph with a specified update interface.
     * @param updateInterface the specified update interface
     */
    public _UpdatableGraph(UpdateInterface updateInterface) {
        this.m_updateInterface = updateInterface;
        this.m_linkGraph = new _Graph();
        this.m_nodeSet = CollUtilities.observableSet(new HashSet<Node>());
        this.m_changedSet = CollUtilities.subSet(new HashSet<Node>(), this.m_nodeSet);
    }

    /**
     * Access method for the m_linkGraph property.
     * @return the current value of the m_linkGraph property
     */
    public Graph getLinkGraph() {
        return this.m_linkGraph;
    }

    /**
     * Access method for the m_updateInterface property.
     * @return the current value of the m_updateInterface property
     */
    public UpdateInterface getUpdateInterface() {
        return this.m_updateInterface;
    }

    /**
     * Access method for the m_nodeSet property.
     * @return the current value of the m_nodeSet property
     */
    public Set<Node> getNodeSet() {
        return this.m_nodeSet;
    }

    /**
     * Access method for the m_changedSet property.
     * @return the current value of the m_changedSet property
     */
    public Set<Node> getChangedNodeSet() {
        return this.m_changedSet;
    }

    /**
     * Access method for the m_link property.
     * @return the current value of the m_link property
     */
    public boolean[][] getLink() {
        computeLink();
        return this.m_link;
    }

    /**
     * Access method for the m_hull property.
     * @return the current value of the m_hull property
     */
    public boolean[][] getHull() {
        computeLink();
        computeHull();
        return this.m_hull;
    }

    /**
     * Sets the update object for this upatable graph.
     * @param update the update object
     */
    public void setUpdate(Update update) {
        this.m_update = update;
    }
    /**
     * Starts the update associated with this updatable graph.
     */
    public void update(Object context) {
        if (this.m_update != null) {
            this.computeHull();
            if (!this.m_changedSet.isEmpty()) {
                this.m_update.update(context);
            }
        }
        else System.err.println("Update does not exist.");
    }

    /**
     * Adds a specified node to the set of changed nodes.
     */
    public void wasChanged(Node node) {
        if (node != null) {
            this.m_changedSet.add(node);
        }
    }

    /**
     * Updates a specified node.
     * @param node the specified node
     */
    public void updateNode(Node node, Object context) {
        Updatable u = this.m_updateInterface.getUpdatable(node);
        //System.out.println("update: " + u);
        u.update(context);
        this.m_changedSet.remove(node);
    }

    /**
     * Access method for the link graph's nodes.
     * @return an array of the link graph's nodes
     */
    public Node[] getNodes() {
        return this.m_linkGraph.getNodes();
    }

    /**
     * Access method for the successor nodes of a specified node.
     * @param node the specified node
     * @return an array of the successor nodes
     * @throws GraphException a graph exception
     */
    public Node[] getEdges(Node node) {
        return this.m_linkGraph.getEdges(node);
    }

    /**
     * Adds a specified node to the link graph and updates link graph's hull.
     * @param node the specified node
     * @return <code>true</code> if the node was added
     */
    public boolean addNode(Node node) {
        if (! this.m_linkGraph.addNode(node)) {
            if (this.m_nodeSet.contains(node))
                throw new IllegalArgumentException(
                    "Inconsistent data structure.");
            return false;
        }
        if (! m_nodeSet.add(node))
            throw new IllegalArgumentException("Inconsistent data structure.");
        //this.computeHull();
        return true;
    }

    /**
     * Removes a specified node from the link graph and updates link graph's
     * hull.
     * @param node the specified node
     * @return <code>true</code> if the node was removed
     */
    public boolean removeNode(Node node) {
        if (! this.m_nodeSet.remove(node)) {
            if (this.m_linkGraph.containsNode(node))
                throw new IllegalArgumentException(
                    "Inconsistent data structure.");
            return false;
        }
        if (! this.m_linkGraph.removeNode(node))
            throw new IllegalArgumentException("Inconsistent data structure.");
        //this.computeHull();
        return true;
    }

    /**
     * Adds an edge to the link graph and updates link graph's hull.
     * @param src the specified source node
     * @param dest the specified destination node
     * @return <code>true</code> if the edge was added
     * @throws GraphException a graph exception
     */
    public boolean addEdge(Node src, Node dest) throws GraphException {
        return this.m_linkGraph.addEdge(src, dest);
        //this.computeHull();
    }

    /**
     * Removes an edge from the link graph and updates link graph's hull.
     * @param src the specified source node
     * @param dest the specified destination node
     * @return <code>true</code> if the edge was removed
     */
    public boolean removeEdge(Node src, Node dest) {
        return this.m_linkGraph.removeEdge(src, dest);
        //this.computeHull();
    }

    /**
     * Determines if the link graph contains a specified node.
     * @param node the specified node
     * @return <code>true</code> if the link graph contains the specified node
     */
    public boolean containsNode(Node node) {
        return this.m_linkGraph.containsNode(node);
    }

    /**
     * Determines if the link graph is cyclic.
     * @return <code>true</code> if the link graph is cyclic
     */
    public boolean hasCycle() {
        return this.m_linkGraph.hasCycle();
    }

    /**
     * Computes the adjacency matrix for the link graph.
     */
    private void computeLink() {
        ArrayList nodes = ((_Graph)this.m_linkGraph).getNodeList();
        int dim = nodes.size();
        this.m_link = new boolean[dim][dim];
        //initiate with false
        for (int i=0; i < dim; i++) {
            for (int j=0; j < dim; j++) {
                this.m_link[i][j] = false;
            }
        }
        //set values
        ArrayList edges;
        for (int i=0; i < dim; i++) {
            edges = ((Node)nodes.get(i)).getEdges();
            for (int e=0; e < edges.size(); e++) {
                int index = nodes.indexOf(edges.get(e));
                this.m_link[i][index] = true;
            }
        }
    }

    /**
     * Computes the adjacency matrix for the link graph's hull.
     */
    private void computeHull() {
        this.computeLink();
        int dim = this.m_link.length;
        this.m_hull = new boolean[dim][dim];

        /*//copy values
         for (int i=0; i<dim; i++) {
         for (int j=0; j<dim; j++) {
         this.m_hull[i][j] = this.m_link[i][j];
         }
         }
         //set values
          for (int k=0; k<dim; k++) {
          for (int i=0; i<dim; i++) {
          for (int j=0; j<dim; j++) {
          if ( (this.m_hull[i][k] == true) && (this.m_hull[k][j] == true) )
          this.m_hull[i][j] = true;
          }
          }
          }
          */


        boolean[] mark = new boolean[dim];
        boolean hit;
        int m, row, col;
        for (row = 0; row < dim; row++) {
            System.arraycopy(this.m_link[row], 0, this.m_hull[row], 0, dim);
            for (m = 0; m < dim; m++) mark[m] = false;
            mark[row] = true;
            do {
                hit = false;
                for (col = 0; col < dim; col++) {
                    if (hit = (this.m_hull[row][col] == true) && !mark[col])
                        break;
                }
                if (hit) {
                    mark[col] = true;
                    for (m = 0; m < dim; m++) {
                        this.m_hull[row][m] |= this.m_link[col][m];
                    }
                }
            } while (hit);
        }
    }

    /*
     public void printLink() {
     this.computeLink();
     String s = new String("---Links:--- \n");
     for (int i=0; i < this.m_link.length; i++) {
     s += "( ";
     for (int j=0; j < this.m_link.length; j++) {
     s += this.m_link[i][j]+" ";
     }
     s += ")\n";
     }
     System.out.println(s);
     }

     public void printHull() {
     this.computeHull();
     String s = new String("---Hull:--- \n");
     for (int i=0; i < this.m_hull.length; i++) {
     s += "( ";
     for (int j=0; j < this.m_hull.length; j++) {
     s += this.m_hull[i][j]+" ";
     }
     s += ")\n";
     }
     System.out.println(s);
     }

     public void printChangedSet() {
     String s = new String("---ChangedSet:---\n( ");
     Iterator iter = this.m_changedSet.iterator();
     while (iter.hasNext())
     s += ((TestObject)((Vertex)iter.next()).getKey()).getID()+", ";
     s += ")";
     System.out.println(s);
     }

     public void printNodeSet() {
     String s = new String("---VertexSet:---\n( ");
     Iterator iter = this.m_nodeSet.iterator();
     while (iter.hasNext())
     s += ((TestObject)((Vertex)iter.next()).getKey()).getID()+", ";
     s += ")";
     System.out.println(s);
     }
     */

}//end of class