/* 
 * Copyright (C) 2009-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.Stroke;
import java.awt.font.FontRenderContext;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.text.CharacterIterator;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import cib.util.AttributedImage;
import cib.util.AttributedShape;
import cib.util.AttributedText;
import cib.util.CoordSpace;
import cib.util.geo.Geo2D;
import cib.util.view.FillPaint;
import cib.util.view.Projection2D;

/** 
 * Draws a rectangular area in a given scale.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.3 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/03/11 19:17:05 $
 */
public class DeviceRenderer {

    private static final AffineTransform TEXT_TRF = new AffineTransform();
    private static final Stroke DEFAULT_STROKE = new BasicStroke(
        /* float width      */  0,
        /* int cap          */  BasicStroke.CAP_ROUND,
        /* int join         */  BasicStroke.JOIN_ROUND,
        /* float miterlimit */  0,
        /* float[] dash     */  null,
        /* float dash_phase */  0
    );

    private List<AttributedShape> m_shapes = new ArrayList<AttributedShape>();
    private List<AttributedText> m_texts = new ArrayList<AttributedText>();
    private List<AttributedImage> m_images = new ArrayList<AttributedImage>();
    private Projection2D m_projection = null;
    private double m_scale = 1;

    public static final double PT_PER_IN = 72; // Java units [Pt]
    public static final double IN_PER_M = 39.37;
    
    private static final AffineTransform FLIP_X = new AffineTransform(
            /*m00*/  1, /*m10*/  0,
            /*m01*/  0, /*m11*/ -1,
            /*m02*/  0, /*m12*/  0
        );
    static {
        final double SCALE = 1. / (PT_PER_IN * IN_PER_M);
        TEXT_TRF.preConcatenate(AffineTransform.getScaleInstance(SCALE, SCALE));
        TEXT_TRF.preConcatenate(FLIP_X);
    }

    /**
     * Constructs a DeviceRenderer object to be extended by specific Renderers.
     * 
     * @param projection the projection model
     * @param scale the scale factor
     */
    public DeviceRenderer(Projection2D projection, double scale) {
        if (projection == null)
            throw new IllegalArgumentException();
        m_projection = projection;
        m_scale = scale;
    }
    
    /**
     * Returns the scale factor.
     * 
     * @return the scale factor
     */
    public double getScale() {
        return m_scale;
    }
    
    /**
     * Returns the projection object. Changes of the projection object change
     * the state of this object.
     * 
     * @return the reference to the internal projection object
     */
    public Projection2D getProjection() {
        return m_projection;
    }

    /**
     * Returns the number of inches per natural units.
     * 
     * @return the number of inches
     */
    public double inPerNaturalUnit() {
        CoordSpace cs = CoordSpace.getCoordSpace();
        return IN_PER_M * cs.getMillimetersPerNaturalUnit() / 1000.;
    }

    /**
     * Adds an AttributedShape object to be rendered.
     * 
     * @param shape the object to be rendered
     */
    public void addShape(AttributedShape shape) {
        m_shapes.add(shape);
    }
    
    /**
     * Adds an AttributedText object to be rendered.
     * 
     * @param text the object to be rendered
     */
    public void addText(AttributedText text) {
        m_texts.add(text);
    }
    
    /**
     * Adds an AttributedImage object to be rendered.
     * 
     * @param image the object to be rendered
     */
    public void addImage(AttributedImage image) {
        m_images.add(image);
    }
    
    /**
     * The render method to be used by deriving classes.
     * 
     * @param g the Graphics to render with
     */
    final protected void _render(Graphics2D g) {
        // Draw shapes
        Iterator<AttributedShape> its = m_shapes.iterator();
        while (its.hasNext()) {
            AttributedShape as = its.next();
            draw(g, as, false);
        }

        // Draw Texts
        FontRenderContext frc = g.getFontRenderContext();
        Iterator<AttributedText> textIt = m_texts.iterator();
        while (textIt.hasNext()) {
            AttributedText _at = textIt.next();
            AttributedText at = (AttributedText)_at.clone();
            AttributedString as = at.getAttributedString();

            // Empty text?
            AttributedCharacterIterator aci = as.getIterator();
            boolean emptyText = aci.getRunLimit() == 0;

            // Text layout in [pt]
            TextLayout[] layouts = null;
            double boundingHeight = 0;
            double boundingWidth = 0;
            double dx = 0;
            double dy = 0;
            double textScale = 1;
            if (! emptyText) {
                // Scale for Windows and Linux Graphics systems
                double maxHeight = 0;
                int runLimit = 0;
                while (aci.setIndex(runLimit) != CharacterIterator.DONE) {
                    runLimit = aci.getRunLimit();
                    Map<AttributedCharacterIterator.Attribute,Object> map =
                        aci.getAttributes();
                    Object value = map.get(TextAttribute.SIZE);
                    if (value != null && value instanceof Number) {
                        double height = ((Number)value).doubleValue();
                        if (height > maxHeight)
                            maxHeight = height;
                    }
                }
                
                final double MAX_TEXT_SIZE = 1000;
                final double OPT_TEXT_SIZE = 1000;
                final double MIN_TEXT_SIZE = 5;
                if (maxHeight > MAX_TEXT_SIZE || maxHeight < MIN_TEXT_SIZE) {
                    textScale = OPT_TEXT_SIZE / maxHeight;
                    runLimit = 0;
                    while (aci.setIndex(runLimit) !=
                        CharacterIterator.DONE) {
                        int runStart = aci.getIndex();
                        runLimit = aci.getRunLimit();
                        
                        Map<AttributedCharacterIterator.Attribute,Object>
                            map = aci.getAttributes();
                        Object value = map.get(TextAttribute.SIZE);
                        if (value != null && value instanceof Number) {
                            double height = ((Number)value).doubleValue();
                            as.addAttribute(TextAttribute.SIZE,
                                new Float(height*textScale), runStart,
                                runLimit);
                        }
                    }
                }
                
                // Find the list of line break positions
                List<Integer> lineBreakPositions = new ArrayList<Integer>();
                CharacterIterator cit = as.getIterator();
                for (char chr = cit.first(); chr != CharacterIterator.DONE;
                chr = cit.next()
                ) {
                    if (chr == '\n') {
                        int index = cit.getIndex();
                        lineBreakPositions.add(new Integer(index));
                    }
                }
                
                // Create an array of Textlayouts
                layouts = new TextLayout[lineBreakPositions.size() + 1];
                for (int i = 0; i < layouts.length; i++) {
                    int beginIndex = i == 0 ?
                        0 : lineBreakPositions.get(i - 1) + 1;
                    int endIndex = i == layouts.length - 1 ?
                        cit.getEndIndex() : lineBreakPositions.get(i);
                    layouts[i] = null;
                    if (beginIndex != endIndex) {
                        AttributedCharacterIterator acit = as.getIterator(
                                null, beginIndex, endIndex);
                        layouts[i] = new TextLayout(acit, frc);
                    }
                }
                
                // Justification is based on text layout. The standard 
                // placement is the left edge of the baseline. The text
                // layout consists of 
                //  - ascent:  distance from baseline to top position
                //  - descent: distance from baseline to bottom position
                //  - advance: length of layout
                // The lineSpacing must be currently a constant value

                double ascent = layouts[0] == null ?
                    0 : layouts[0].getAscent();
                boundingHeight = 0;
                boundingWidth = 0;
                int nEmptyLines = 0;
                for (int i = 0; i < layouts.length; i++) {
                    if (layouts[i] == null) {
                        nEmptyLines++;
                    }
                    else {
                        boundingHeight += layouts[i].getAscent() +
                            layouts[i].getDescent();
                        if (layouts[i].getAdvance() > boundingWidth)
                            boundingWidth = layouts[i].getAdvance();
                    }
                }

                // Handle empty layouts
                int nNotEmptyLines = layouts.length - nEmptyLines;
                int avgLineHeight = nNotEmptyLines == 0 ?
                    0 : (int)Math.round(boundingHeight / nNotEmptyLines);
                
                for (int i = 0; i < layouts.length; i++) {
                    if (layouts[i] == null)
                        layouts[i] = new TextLayout(" ",
                            new Font("SansSerif", Font.PLAIN,
                                    avgLineHeight), frc);
                }

                // Justification
                switch (at.getJustification()) {
                case AttributedText.JUST_LEFT_BASELINE:
                    dx = 0;
                    dy = 0;
                    break;
                case AttributedText.JUST_LEFT_BOTTOM:
                    dx = 0;
                    dy = ascent - boundingHeight;
                    break;
                case AttributedText.JUST_LEFT_CENTER:
                    dx = 0;
                    dy = lineBreakPositions.isEmpty() ?
                        0.5 * ascent : ascent - 0.5 * boundingHeight;
                    break;
                case AttributedText.JUST_LEFT_TOP:
                    dx = 0;
                    dy = ascent;
                    break;

                case AttributedText.JUST_CENTER_BASELINE:
                    dx = - 0.5 * boundingWidth;
                    dy = 0;
                    break;
                case AttributedText.JUST_CENTER_BOTTOM:
                    dx = - 0.5 * boundingWidth;
                    dy = ascent - boundingHeight;
                    break;
                case AttributedText.JUST_CENTER_CENTER:
                    dx = - 0.5 * boundingWidth;
                    dy = lineBreakPositions.isEmpty() ?
                        0.5 * ascent : ascent - 0.5 * boundingHeight;
                    break;
                case AttributedText.JUST_CENTER_TOP:
                    dx = - 0.5 * boundingWidth;
                    dy = ascent;
                    break;

                case AttributedText.JUST_RIGHT_BASELINE:
                    dx = - boundingWidth;
                    dy = 0;
                    break;
                case AttributedText.JUST_RIGHT_BOTTOM:
                    dx = - boundingWidth;
                    dy = ascent - boundingHeight;
                    break;
                case AttributedText.JUST_RIGHT_CENTER:
                    dx = - boundingWidth;
                    dy = lineBreakPositions.isEmpty() ?
                        0.5 * ascent : ascent - 0.5 * boundingHeight;
                    break;
                case AttributedText.JUST_RIGHT_TOP:
                    dx = - boundingWidth;
                    dy = ascent;
                    break;

                default:
                    dx = 0;
                    dy = 0;
                }
            }

            // Text to be drawn
            // Setup transformation: justified text in [pt]
            AffineTransform trf = AffineTransform.getTranslateInstance(
                    dx, dy);
            // Unscale (16 bit stuff)
            trf.preConcatenate(AffineTransform.getScaleInstance(
                1. / textScale, 1. / textScale));
            // Transform to [m]
            trf.preConcatenate(TEXT_TRF);
            // Unscale so that the text height remains unchanged
            AffineTransform atTrf = at.getTransform();
            int flags = atTrf.getType() &
                AffineTransform.TYPE_MASK_SCALE;
            if (flags == AffineTransform.TYPE_UNIFORM_SCALE ||
                flags == AffineTransform.TYPE_GENERAL_SCALE) {
                Point2D o = new Point2D.Double();
                Point2D x = new Point2D.Double(1, 0);
                Point2D y = new Point2D.Double(0, 1);
                Point2D o_ = atTrf.transform(o, null);
                Point2D x_ = atTrf.transform(x, null);
                Point2D y_ = atTrf.transform(y, null);
                double scx = 1. / Geo2D.distance(o_, x_);
                double scy = 1. / Geo2D.distance(o_, y_);
                trf.preConcatenate(AffineTransform.getScaleInstance(scx,
                        scy));
            }
            // Apply the text transformation
            trf.preConcatenate(atTrf);

            // Normalize because the specific layouting methods behave
            // sensitive on large translation values
            double trfNrmX = trf.getTranslateX();
            double trfNrmY = trf.getTranslateY();
            trf.preConcatenate(AffineTransform.getTranslateInstance(
                    -trfNrmX, -trfNrmY));
            
            // Load the text as a shape
            Shape shape = null;
            if (layouts == null || layouts.length == 0) {
                shape = trf.createTransformedShape(
                    new Ellipse2D.Double(-200, -200, 400, 400));
            }
            else {
                Shape _shape = layouts[0].getOutline(trf);
                Path2D path = new Path2D.Double(_shape);
                double tx = 0;
                double ty = 0;
                for (int i = 1; i < layouts.length; i++) {
                    ty += layouts[i].getAscent() +
                        layouts[i].getDescent();
                    AffineTransform trfText =
                        AffineTransform.getTranslateInstance(tx, ty);
                    trfText.preConcatenate(trf);    
                    _shape = layouts[i].getOutline(trfText);
                    path.append(_shape, false);
                }
                shape = path;
            }

            // Undo the normalizing translation
            AffineTransform trfBack = AffineTransform.getTranslateInstance(
                    trfNrmX, trfNrmY);
            shape = trfBack.createTransformedShape(shape);

            // Draw text as shape
            AttributedShape attributedShape = new AttributedShape(shape, null,
                    at.getDrawPaint(), at.getDrawPaint());

            draw(g, attributedShape, true);
        }

        // Draw images
        AffineTransform modelToView = m_projection.getModelToView();
        Iterator<AttributedImage> iti = m_images.iterator();
        while (iti.hasNext()) {
            AttributedImage ai = iti.next();

            // Transformation in view units
            AffineTransform trf = ai.getTransform();
            trf.preConcatenate(modelToView);

            // Transform and draw the image
            BufferedImage image = ai.getImage();
            double scaleX = ai.getWidth() / image.getWidth();
            double scaleY = ai.getHeight() / image.getHeight();
            AffineTransform _trf = new AffineTransform(trf);
            _trf.scale(scaleX, scaleY);
            _trf.translate(0, image.getHeight());
            _trf.concatenate(FLIP_X);
            g.drawImage(image, _trf, null);
        }
    }

    private void draw(Graphics2D g, AttributedShape shape, boolean isText) {
        // Is it to be drawn?
        AffineTransform modelToView = m_projection.getModelToView();
        Rectangle2D r2D = shape.getBounds2D();
        Shape r2DT = modelToView.createTransformedShape(r2D);
        Rectangle r = r2DT.getBounds();
        if (! g.hitClip(r.x, r.y, r.width, r.height))
            return;
        
        // Transform the Shape to view coordinates
        Shape shapeT = modelToView.createTransformedShape(shape);

        // Fill the interior of the Shape
        Paint fill = shape.getFillPaint();
        if (fill != null) {
            Shape shapeFill = shapeT;

            // There is a strange phenomenon on Mac OS X: the coordinate
            // difference between draw() and fill() is approx. 4.2 pt in x and
            // 4.2 pt in y direction?!
            AffineTransform compensateMacOSbug = 
                AffineTransform.getTranslateInstance(_getFillOffsetX(), 
                        _getFillOffsetY());
            shapeFill = compensateMacOSbug.createTransformedShape(shapeT);
            
            if (fill instanceof FillPaint)
                ((FillPaint)fill).setModelToView(modelToView);
            else if (fill instanceof Color && fill.equals(Color.WHITE))
                fill = Color.black;
            g.setPaint(fill);
            g.fill(shapeFill);
        }

        // Set boundary color
        Paint drawPaint = shape.getDrawPaint();
        if (drawPaint == null)
            drawPaint = Color.black;
        else if (drawPaint instanceof Color && drawPaint.equals(Color.WHITE))
            drawPaint = Color.black;
        g.setPaint(drawPaint);

        // Set boundary stroke 
        Stroke stroke = shape.getStroke();
        if (stroke == null)
            stroke = DEFAULT_STROKE;
        if (stroke instanceof BasicStroke) {
            float m2vScale = (float)modelToView.getScaleX();

            BasicStroke basicStroke = (BasicStroke)stroke;
            float width = basicStroke.getLineWidth() * m2vScale;
            int cap = basicStroke.getEndCap();
            int join = basicStroke.getLineJoin();
            float miterlimit = basicStroke.getMiterLimit() * m2vScale;
            float[] dash = basicStroke.getDashArray();
            if (dash != null) {
                for (int i = 0; i < dash.length; i++)
                    dash[i] *= m2vScale;
            }
            float dash_phase = basicStroke.getDashPhase() * m2vScale;
            
            g.setStroke(new BasicStroke(width, cap, join, miterlimit, dash,
                dash_phase));
        }

        // Draw
        g.draw(shapeT);
    }

    /**
     * Allows to define an offset for filled structures. Is necessary to fix a
     * Mac OS X bug resulting in drawn and filled structures to be rendered at 
     * different locations.
     * 
     * @return the offset in x direction [pt]
     */
    protected double _getFillOffsetX() {
        return 0;
    }
    /**
     * Allows to define an offset for filled structures. Is necessary to fix a
     * Mac OS X bug resulting in drawn and filled structures to be rendered at 
     * different locations.
     * 
     * @return the offset in y direction [pt]
     */
    protected double _getFillOffsetY() {
        return 0;
    }
}